if (!('contains' in String.prototype)) {
	  String.prototype.contains = function(str, startIndex) {
	    return ''.indexOf.call(this, str, startIndex) !== -1;
	  };
	}

if (typeof String.prototype.startsWith != 'function') {
    String.prototype.startsWith = function(prefix) {
        return this.slice(0, prefix.length) == prefix;
    };
}
		 
if (typeof String.prototype.endsWith != 'function') {
    String.prototype.endsWith = function(suffix) {
        return this.slice(-suffix.length) == suffix;
    };
}

var ngp = require('./ecxngp');   

var funcSteps = [                                                                                  
	{                                                                                              
		continue: 'next',                                                                               
		name: 'adhoc_run',   
		verb: 'POST',
		path: null,                                                   
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
    }
];

const adhocRunStepFuncs = {
	adhoc_run: function(step, funcStream, ecxSessionId, oFromNg, cb){                              
		adhocRun(step, funcStream, ecxSessionId, oFromNg, cb);		
	}
};

exports.init = init;

function init(req, res, providerType)                                    
{
	var ngState = ngp.getSessionState(req, res);

	if ( ngState.stale )                                                                           
		return;                                                                                    

	var ecxSid = ngState.ecxSid;
    res.ecxSid = ecxSid;

	var stepsStream = JSON.parse(JSON.stringify(funcSteps));        

	stepsStream.forEach(function(step){
		step.ip = ngp.localIp;
		step.port = ngp.ecxExternal;

		step.func = adhocRunStepFuncs[step.name];                                                  
		
		step.cb = function(ecxSid, data){ 
			ngp.logMsg('info', "Finished step '" + ngp.getCurrentFunc(step.reqId, ecxSid) + '(' + this.name + ")' in session ID '" + ecxSid + "'");

			if ( data )
			{
				ngp.logMsg('debug', "Data returned by step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'=" + JSON.stringify(data));
			}
		};
	});

	ngp.createRequestStore(stepsStream, ecxSid, req.reqId, 'adhocRun_' + providerType);  

	ngState.currentFunc = 'adhocRun_' + providerType;                                 
	return stepsStream;
}

function adhocRun(step, funcStream, ecxSessionId, oFromNg, cb)
{
	const urlParse = require('url-parse');
	
	var path = '/api/endeavour/job?filter=';
	var filter = '[{"property":"name","value":"' + oFromNg.subtype + '_' + oFromNg.slaPolicyName + '"}]';

	var escFilter = ngp.escape(filter);

	var finalPath = path + escFilter;

	ngp.sendRequestToEcx('GET', ngp.ecxExternal, finalPath, ecxSessionId, null, function(jobO) {
		var jobStatus = jobO.rO['jobs'];
		var slaPolicyName = (jobStatus[0].name).substring((jobStatus[0].name).indexOf('_')+1)

		if ( jobO.e || jobO.response.statusCode != 200 ) {
			var failResult = ngp.getGenericEcxErrorResponse(jobO, ecxSessionId, step); 
			ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);    
		}
		else {
			var sprofs = jobO.rO['jobs'];

			if(sprofs[0].status === 'HELD') {
				var failResult = {
						statusCode: 405,
						response: { rc: ('Cannot initiate run of job \'' + sprofs[0].name + '\' whose schedule is currently held. Navigate to \'Jobs and Operations\' screen, under \'Schedule\' tab, release schedue for job \'' + sprofs[0].name) + '\'' }
					};
					ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
			} else {

				for ( jobLinkName in sprofs[0].links ) {                                                          
					var jobLink = sprofs[0].links[jobLinkName];
					
					if ( jobLink.rel && jobLink.rel == 'action' && jobLinkName == 'start')
					{ 	var path = urlParse(jobLink.href).pathname +  urlParse(jobLink.href).query ;
						break;
					}
				}

				var postBody = { 
					source: oFromNg.resource
				};
	
				var scrubbedNgInput = ngp.scrubPostBody(oFromNg, true);
				ngp.logMsg('warn', "Input from client for provider options in session ID='" + ecxSessionId + "'=" + JSON.stringify(scrubbedNgInput));
	
				ngp.sendRequestToEcx('POST', ngp.ecxExternal, path, ecxSessionId, postBody, function(runAdhocJob) {
					if ( runAdhocJob.e || runAdhocJob.response.statusCode != 201 ) {
						var failResult = ngp.getGenericEcxErrorResponse(runAdhocJob, ecxSessionId, step); 
						ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);    
					}
					else {				
						var successResult = {
							statusCode: 201,
							response: runAdhocJob.rO,
						};
			
						ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);	
					}         
				});
			} 

			
		}
	});
}