var request = require('request')
process.env.NODE_TLS_REJECT_UNAUTHORIZED = "0";
var host = null;
var storage_path = require('./config').storage_path
var applicationdatabase_path = require('./config').applicationdatabase_path
var database_path = require('./config').database_path
var setSessionandHeader = require('./helpfunction').setSessionandHeader
var ngp = require('../../ecxngp');
var logMsg = require('../../ecxngp').logMsg;
const json2csv = require('json2csv').parse;
const promise_limit = 100;

var option = {
    method:'GET',
    url:'',
    headers:
        {
            'x-endeavour-sessionid':'',
            'x-endeavour-auditrequest':'yes',
            'Content-Type':'application/json',
            'Accept':'application/json'
        },
};

var returnres ={
    "links":{
        "self" :{
            "rel" : "self",
            "href" : ""
        }  
    }
}
var dbinfo_list= new Array();
var dbcounts = {};
var storagename = {};


function dbparsefilter(query,next)
{
    try{
        if(typeof(query['filter']) == undefined || query['filter'] ==null ) return {}
        var arr = JSON.parse(query['filter'])
        var filters = {};
        for ( var i in arr )
        {
            if( isNaN(arr[i].value))
            {

                var params = new Set(arr[i].value);
                if( params.has('All') || arr[i].value == "All")
                {
                    continue;
                }
                filters[arr[i].property] = params;
            } 
            else
            {
                filters[arr[i].property] = arr[i].value;
            }
        }
        return filters;
    }
    catch(error)
    {
        logMsg('error', error)
        const err = new Error("Incorrect filtering Parameters")
        err.status = 400
        next(err)
    }
}
function getstorageName( starturl , resolve , reject)
{
    option.url = starturl
    function loop ()
    {
        request( option , function(error , response , body ){
            if(!error && response.statusCode === 200 )
            {
                var storages = JSON.parse(body).storages
                var links = JSON.parse(body).links
                for( var i in storages )
                {
                    storagename[ storages[i].storageId ] = storages[i].name
                }
                if('nextPage' in links)
                {
                    option.url = links.nextPage.href
                    return loop()
                }
                else
                {            
                    logMsg('info', storagename)
                    return resolve()
                }
            }
            else
            {
                logMsg('error', error)
                const err = new Error("Error while getting storage names")
                err.status = 403
                return reject(err);
            }

        })
    }
    return loop()
}

function Issatisfy(uiItem,dbinfo,dbItem,filters)
{
    return !(uiItem in filters) ||  filters[uiItem].has(dbinfo[dbItem])
}
function Istimesatisfy(ProtectionTime,filters)
{
    if(!('Backup History' in filters)) return true;
    var currenttime = new Date().getTime();
    var timelimit = currenttime - filters['Backup History']*24*60*60*1000;
    if( ProtectionTime > timelimit ) return true;
    return false;
}
function parseinfo( ori,dbinfo,sucess,filters)
{
    dbinfo['pk'] = ori['pk'];
    dbinfo['parentPk'] = ori['parentPk'];
    dbinfo['DB'] = ori['name'];
    if(!Issatisfy('DBs',dbinfo,'DB',filters)) return false;
    dbinfo['SLAPolicy'] = ori.protectionInfo.storageProfileName;
    if(!Issatisfy( 'SLA Policy',dbinfo,'SLAPolicy',filters)) return false;
    dbinfo['ProtectionTime'] = ori.protectionInfo.protectionTime;
    if(!Istimesatisfy(dbinfo['ProtectionTime'],filters)) return false;
    if(sucess)
    {
        dbinfo['Status'] = 'Success';
        dbinfo['BackupStorage'] = storagename[ori.protectionInfo.storageSnapshots[0].storageId]
        var transferSize = ori.protectionInfo.transferSize;
        var transferDuration = ori.protectionInfo.transferDuration;
        dbinfo['BackupDetails'] ={ 'BackupData' : transferSize , 'BackupSpeed' : transferDuration}
    }
    else{
        dbinfo['Status'] = 'Failed';
        dbinfo['BackupStorage'] = '';
        dbinfo['BackupDetails'] = { 'FailureReason' : ori.protectionInfo.lastErrorMessage}
    }
    return true
}

function getlinks(starturl,resolve,reject)
{
    linklists = new Array();
    option.url = starturl
    function loop()
    {
        request(option,function(error,response,body){
            if(!error && response.statusCode === 200 )
            {
                var child_lists = JSON.parse(body).children;
                var links = JSON.parse(body).links;
                for( var index in child_lists)
                {
                   if(child_lists[index].links && child_lists[index].links.versions) {
                       linklists.push(child_lists[index].links.versions.href)
                    };
                }
                if('nextPage' in links)
                {
                    option.url = links.nextPage.href
                    return loop()
                }
                else
                {
                    return resolve(linklists)
                } 
            }
            else if( error )
            {
                logMsg('error', error)
                const err = new Error("Error when getting page links")
                err.status = 403
                reject(err);
            }
        });        
    }
    return loop();
}

function getrecoverydb(linklists, filters, resolve, reject) {
    logMsg('info', 'getrecoverydb: start');
    if ('Status' in filters && filters['Status'].has('Success')) {
        return resolve(); 
    } 
    var counter = 0;
    let linklists_index=0;
    const linklists_length = linklists.length;
    logMsg('info', 'getrecoverydb: num version links found: ' + linklists_length);
    var promise_array = new Array();
    for ( let i = 0; i < linklists_length; i++) {
        var promise = new Promise( function(resolve,reject) {
            option.url = linklists[i];
            function loop() {
                request(option, function(error,response,body) {
                    if (!error && response.statusCode === 200 ) {
                        var contents = JSON.parse(body).contents;
                        var links = JSON.parse(body).links;
                        if (!(!('nextPage' in links) && 'firstPage' in links && counter >= 1)) {                        
                            const contents_length = contents.length;
                            for (let i = 0; i < contents_length; i ++) {
                                var db = contents[i];
                                var dbinfo = {};
                                if(!parseinfo(db.properties.errObj, dbinfo, false, filters)) {
                                    continue;
                                }
                                dbinfo_list.push(dbinfo);
                                dbcounts[dbinfo['DB']] = dbinfo['DB'] in dbcounts ? dbcounts[dbinfo['DB']] + 1 : 1;
                            }
                        }
                        if ('nextPage' in links) {
                            option.url = links.nextPage.href;
                            loop();
                        } else {
                            if ('firstPage' in links && counter < 1) {
                                option.url = links.firstPage.href;
                                counter++;
                                loop();
                            } else {
                                resolve();
                            }
                        }  
                    } else if ( error ) {
                        logMsg('error', error)
                        const err = new Error("getrecoverydb: Error when getting failed backup history")
                        err.status = 403
                        reject(err);
                    }

                    resolve();
                });
            }
            loop();
        });

        promise_array.push(promise);
    }

    logMsg('info', 'getrecoverydb: promise.all called. count = ' + promise_array.length);
    Promise.all(promise_array)
        .then(() => resolve())
        .catch((error) => {
           logMsg('error', error)
           const err = new Error("Error while getting sucessful backup history")
           err.status = 403
           reject(err);
        }) 
}

function gethypervisordb(linklists, filters, resolve, reject) {
    logMsg('info', 'gethypervisordb: start');
    if ('Status' in filters && filters['Status'].has('Failed')) {
        return resolve(); 
    } 
    var counter = 0;
    let linklists_index=0;
    const linklists_length = linklists.length;
    logMsg('info', 'gethypervisordb: num version links found: ' + linklists_length);
    var promise_array = new Array();

    for (let i = 0; i < linklists_length; i++) {
        var promise = new Promise( function(resolve,reject) {
            option.url = linklists[i];
            function loop() {
                request(option, function(error,response,body) {
                    if (!error && response.statusCode === 200 ) {
                        var dblists = JSON.parse(body).contents;
                        var links = JSON.parse(body).links;
                        if (!(!('nextPage' in links) && 'firstPage' in links && counter >= 1)) {
                            const dblists_length = dblists.length;
                            for (let i = 0; i < dblists_length; i++) {
                                var vm = dblists[i]
                                var dbinfo = {};
                                if(!parseinfo(vm.properties, dbinfo, true, filters)) {
                                    continue;
                                }
                                dbinfo_list.push(dbinfo);
                                dbcounts[dbinfo['DB']] = dbinfo['DB'] in dbcounts ? dbcounts[dbinfo['DB']] + 1 : 1;
                            }
                        }
                        if ('nextPage' in links) {
                            option.url = links.nextPage.href;
                            loop();
                        } else {
                            if ('firstPage' in links && counter < 1) {
                                option.url = links.firstPage.href;
                                counter++;
                                loop();
                            } else {
                                resolve();
                            }
                        } 
                    } else if ( error ) {
                        logMsg('error', error)
                        const err = new Error("gethypervisordb: Error while getting sucessful backup history")
                        err.status = 403
                        reject(err);
                    }

                    resolve();
                });
            }
            loop();
        });

        promise_array.push(promise)
    }

    logMsg('info', 'gethypervisordb: promise.all called. count = ' + promise_array.length);
    Promise.all(promise_array)
        .then(() => resolve())
        .catch((error) => {
           logMsg('error', error)
           const err = new Error("Error while getting sucessful backup history")
           err.status = 403
           reject(err);
        })
}

function compare( a,b )
{
    if((a.DB === b.DB && a.ProtectionTime < b.ProtectionTime) || (a.DB > b.DB))
    {
        return 1;
    }
    else return -1;
}
function countdbs(res,req){
    dbinfo_list.sort(compare)
    for (var index in dbinfo_list) {
        var db = dbinfo_list[index];
        db['TotalRuns'] = dbcounts[db.DB];
    }
    returnres['instances'] = dbinfo_list;
    if('download' in req.query && req.query['download'] == 'true')
    {
        if('type' in req.query && req.query['type'] == 'csv')
        {
            res.setHeader('Content-Disposition', 'attachment;filename = "report.csv"')
            if(dbinfo_list.length == 0)
            {
                const fields = ['DB', 'SLAPolicy', 'ProjectTime','Status','BackupStorage','BackupDetails'];
                const opts = { fields };    
                var csv = json2csv(dbinfo_list,opts);
                res.write(csv)
                res.end()                       
            }
            else
            {
                var csv = json2csv(dbinfo_list);
                res.write(csv)
                res.end()   
            }

        }
        else
        {
            res.setHeader('Content-Disposition', 'attachment')
            res.json(dbinfo_list)
        }
    }
    else
    {
        res.json(returnres)
    }
}
function getdbbackuphistory(filters,res,req,next)
{
    setSessionandHeader(res,req)
    .then(function(value){
        option.headers['x-endeavour-sessionid'] = value[0]
        host = value[1]
        returnres["links"]["self"]["href"] = host + "/ngp/analytics/dbbackuphistory";
        dbinfo_list = [];
        dbcounts = {};
        storagename = {};
    })
    .then(function(){
        return new Promise( function(resolve,reject) {
            getstorageName(host+storage_path,resolve,reject)
        }) 
    })
    .then(function(){
        return new Promise(function(resolve,reject){
            //getlinks(host+hypervisordb_path,resolve,reject)
              getlinks(host+applicationdatabase_path,resolve,reject)
        })
    })
    .then( function(linklists){
        return new Promise( function(resolve,reject){
            gethypervisordb(linklists,filters,resolve,reject)
        })
    })
    .then( function()
    {
        return new Promise( function(resolve,reject){
            getlinks(host+database_path,resolve,reject)
        })
    })
    .then(function(linklists){
        return new Promise( function( resolve,reject){
            getrecoverydb(linklists,filters,resolve,reject)
        })
    })
    .then(function()
    {
        countdbs(res,req)
    })
    .catch(function(error){
        logMsg('error', error)
        next(error)
    })  
}

module.exports =
    {
        getdbbackuphistory : getdbbackuphistory,
        dbparsefilter : dbparsefilter
    };
