if (!('contains' in String.prototype)) {
	  String.prototype.contains = function(str, startIndex) {
	    return ''.indexOf.call(this, str, startIndex) !== -1;
	  };
	}

if (typeof String.prototype.startsWith != 'function') {
    String.prototype.startsWith = function(prefix) {
        return this.slice(0, prefix.length) == prefix;
    };
}
		 
if (typeof String.prototype.endsWith != 'function') {
    String.prototype.endsWith = function(suffix) {
        return this.slice(-suffix.length) == suffix;
    };
}

var ngp = require('./ecxngp');   

var applyPolicyOptionsFunctionSteps = [                                                                                  
	{                                                                                             
		continue: 'next',                                                                         
		name: 'update_policy',
		verb: 'PUT',                                                                               
		path: '/api/endeavour/policy',                    
		ip: null,                                                                                                                                      
		port: null, 
		type: null,
		func: null,                                                                                 
		cb: function(res, data){ logMsg('info', 'Step=' + this.name + ' finished');}
	}, 		    
];

const applyScriptFunctions = {
    update_policy: function(step, funcStream, ecxSessionId, oFromNg, cb) {
		updatePolicyScript(step, funcStream, ecxSessionId, oFromNg, cb); 		
    }
};

exports.init = init;

function init(req, res, slaPolicyName, slaSubType)                                    
{
	ngp.logMsg('info', 'init()  ' + JSON.stringify(req.query));
	var ngState = ngp.getSessionState(req, res);

	if ( ngState.stale ) 
		return;

	var ecxSid = ngState.ecxSid;
	res.ecxSid = ecxSid;
	var stepsStream = JSON.parse(JSON.stringify(applyPolicyOptionsFunctionSteps));
    
	stepsStream.forEach(function(step){
		step.ip = ngp.localIp;
        step.port = ngp.ecxExternal;
        step.queryParameters = ngp.ngpUtils.formatRequestQueryObject(req.query);
        step.userId = req.query['userId'];
        step.func = applyScriptFunctions[step.name];
		
		step.cb = function(ecxSid, data){ 
			ngp.logMsg('info', "Finished step '" + ngp.getCurrentFunc(step.reqId, ecxSid) + '(' + this.name + ")' in session ID '" + ecxSid + "'");
			if ( data )
			{
				ngp.logMsg('debug', "Data returned by step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'=" + JSON.stringify(data));
			}
		};
	});

	ngp.createRequestStore(stepsStream, ecxSid, req.reqId, "applyPolicyOptions_" + slaSubType  + "_" + slaPolicyName);
	ngp.setRequestState(req.reqId, ecxSid, 'subType', slaSubType);
	ngp.setRequestState(req.reqId, ecxSid, 'slaPolicyName', slaPolicyName);
    ngState.currentFunc = "applyPolicyOptions_" + slaSubType  + "_" + slaPolicyName;

	return stepsStream;
}

function updatePolicyScript(step, funcStream, ecxSessionId, oFromNg, cb) { 
	var ngState = ngp.ecxSidHash[ecxSessionId];   		
	var policyId = null;

	getPolicyId (step, ecxSessionId, policyId, function(callback) {
		var updateSprofInfo = ngp.getRequestState(step.reqId, ecxSessionId, 'updateSprofInfo');   	
		var sprofTplClone = JSON.parse(JSON.stringify(ngp.sprofTpl));
				
		ngp.sendRequestToEcx('GET', ngp.ecxExternal, step.path, ecxSessionId, null, function(getPolicySpecObj){ 	
			if ( getPolicySpecObj.e || getPolicySpecObj.response.statusCode != 200 ) {
	     		var failResult = ngp.getGenericEcxErrorResponse(getPolicySpecObj, ecxSessionId, step);			
				ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult);	
			}				
			else {   
				var policySpecObject  = JSON.parse(JSON.stringify(getPolicySpecObj.rO.spec));

				var postBodyPolicyOptions = getPostBody(ecxSessionId, step, ngState, oFromNg).policyOption;
				
				for (var key in postBodyPolicyOptions) {
					policySpecObject.option[key] = postBodyPolicyOptions[key];
				}

				var postBody = {
					script: getPostBody(ecxSessionId, step, ngState, oFromNg).script,					
					spec:  policySpecObject                                                                          					
				};
								
				ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path, ecxSessionId, postBody, function(sprofObj) {			
					if ( sprofObj.e || sprofObj.response.statusCode != 200 ) {
						var failResult = ngp.getGenericEcxErrorResponse(sprofObj, ecxSessionId, step);			
						ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult);	
					}	
					else {
						ngp.setRequestState(step.reqId, ecxSessionId, 'sprofName', sprofObj.rO.name);        // save for next step
						var successResult = {                       
							   statusCode: sprofObj.response.statusCode,
							  response: sprofObj.rO,
						 };
						ngp.setSerialStreamResult(ecxSessionId, successResult);                              // perhaps replaced by next step
						step.cb(ecxSessionId, sprofObj.rO);                                                  
						ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	
					}					
				});
			}		
	    });
	});
};

function getPolicyId(step, ecxSid, policyId, callback)
{
	ngp.ngpCoreHelper.getPolicies(ecxSid, function(getPoliciesResObj) {
		var policiesArray= getPoliciesResObj.rO['policies'];		
		var subType = ngp.getRequestState(step.reqId, ecxSid, 'subType');
		var slaPolicyName = ngp.getRequestState(step.reqId, ecxSid, 'slaPolicyName');	
		var policyName = subType + '_' + slaPolicyName;
		
		policiesArray.forEach(function(policy){
			if ( policyId )
				return;
			if ( policyName == policy.name )
			{
				policyId = policy.id;
			}
		});

		if ( policyId )
		{			
			step.path += '/' + policyId;
			callback();
		}
		else
		{
			callback();
		}
	});
}

function getPostBody(ecxSessionId, step, ngState, oFromNg) {
	if ( ! ngp.getRequestState(step.reqId, ecxSessionId, 'userHref'))                                                 // FTI we skipped any user logic
		return oFromNg;
	oFromNg.options.user = {
		href: ngp.getRequestState(step.reqId, ecxSessionId, 'userHref'),
	};
	return oFromNg;
}