if (!('contains' in String.prototype)) {
    String.prototype.contains = function(str, startIndex) {
      return ''.indexOf.call(this, str, startIndex) !== -1;
    };
  }

if (typeof String.prototype.startsWith != 'function') {
  String.prototype.startsWith = function(prefix) {
      return this.slice(0, prefix.length) == prefix;
  };
}

if (typeof String.prototype.endsWith != 'function') {
  String.prototype.endsWith = function(suffix) {
      return this.slice(-suffix.length) == suffix;
  };
}

var ngp = require('../ecxngp');
var catalogApplyslasHelper = require('./catalog-applysla-helper');
var catalogRestoreHelper = require('./catalog-restore-helper');

var url = require('url');

/**
 * Init function thats sets the steps and its values
 *
 * @param req
 * @param res
 * @param functionType
 */
function init(req, res, functionType){

    ngp.logMsg('info', 'Catalog provider init()  ' + JSON.stringify(req.query));
	var ngState = ngp.getSessionState(req, res);

	if ( ngState.stale )
		return;

	var ecxSid = ngState.ecxSid;
	res.ecxSid = ecxSid;

    switch (functionType){
        case 'applySLAPolicies':
            var stepsStream = JSON.parse(JSON.stringify(catalogApplyslasHelper.catalogApplySlasFuncSteps));
            ngp.createRequestStore(stepsStream, ecxSid, req.reqId, 'catalog_applySLAPolicies');
            ngState.currentFunc = 'catalog_applySLAPolicies';
			break;

		case 'restore':
            var stepsStream = JSON.parse(JSON.stringify(catalogRestoreHelper.catalogRestoreFuncSteps));
            ngp.createRequestStore(stepsStream, ecxSid, req.reqId, 'catalog_recovery');
            ngState.currentFunc = 'catalog_recovery';
			break;
    }

	stepsStream.forEach(function(step){
		step.ip = ngp.localIp;
        step.port = ngp.ecxExternal;

        switch (functionType){
            case 'applySLAPolicies':
                step.func = catalogApplyslasHelper.catalogApplySlasFuncs[step.name];
                break;

			case 'restore':
                step.func = catalogRestoreHelper.catalogRestoreFuncs[step.name];
				break;
        }

		step.cb = function(ecxSid, data){
			ngp.logMsg('info', "Finished step '" + ngp.getCurrentFunc(step.reqId, ecxSid) + '(' + this.name + ")' in session ID '" + ecxSid + "'");

			if ( data ){
				ngp.logMsg('debug', "Data returned by step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'=" + JSON.stringify(data));
			}
		};
	});

	return stepsStream;
}

exports.init = init;