
if (!('contains' in String.prototype)) {
	  String.prototype.contains = function(str, startIndex) {
	    return ''.indexOf.call(this, str, startIndex) !== -1;
	  };
	}

if (typeof String.prototype.startsWith != 'function') {
    String.prototype.startsWith = function(prefix) {
        return this.slice(0, prefix.length) == prefix;
    };
}
		 
if (typeof String.prototype.endsWith != 'function') {
    String.prototype.endsWith = function(suffix) {
        return this.slice(-suffix.length) == suffix;
    };
}
 

const SECUREPORT=8083;
var ecxInternal=8082;
var ecxExternal=8443;    

var HashMap = require('hashmap');
var getEcxLocaleHeader = new HashMap();
const ONEMEG=1024*1024;
const LOGFN = 'ecxngp.log';
const PUBLICDIR = __dirname + '/public/';
const ECXSESSIONHEADER = 'x-endeavour-sessionid';
const ECXLOCALEHEADER = 'x-endeavour-locale';
const TIMERINTERVAL = 10*1000;  
const bodyParser = require('body-parser');

const express = require('express');
const app = express();

app.set('x-powered-by', false);
app.set('runningAppsString', '');
app.set('timerStarted', false);

const os = require('os');
const url = require('url');
const fs = require('fs');
const crypto = require('crypto');

// Via NPM

const yaml = require('js-yaml');
const traverse = require('traverse');
const modresp = require('express-mung');
const winston = require('winston');
const si = require('systeminformation');

winston.level = 'info';

var Rotate = require('winston-logrotate').Rotate;

if ( process.env.LOG_LEVEL ) 
	winston.level = process.env.LOG_LEVEL;

if ( process.env.NODE_LOG_DIR ) 
var logDir = '/' + process.env.NODE_LOG_DIR;
else
var logDir = '/data/log/node-cdm-service';

if ( process.env.LOG_DIR ) 
	logDir = process.env.LOG_DIR;	

const logger = new (winston.Logger)({
	transports: [
	   new (winston.transports.Console)({colorize : true, timestamp:true, level: winston.level}),
	   new Rotate({
					//file: '/data/log/node-cdm-service/' + LOGFN,
					file: logDir + '/' + LOGFN,
        	colorize: false,
        	timestamp: true,
        	json: false,
        	size: '20m',
        	keep: 5,
        	compress: true
	   }),
	 ]
});

service_header();

if( process.env.CORE_IP )
{
	var localIp = process.env.CORE_IP;
}
else
{
	var localIp = getLocalIp();  
}
                                                                        
const apiFirstStep = '/ngp/';

exports.logMsg = logMsg;

exports.logLevel = winston.level;


process.on('uncaughtException', function(err) {                                                             // made it to the event loop!
	logger.clear();
	logger.close();
	appendErrorToLog(err);
	throw new Error('Restarting Node-cdm-service after ncaught exception');                                                   // let Node bring matters to a halt, Exception within exception handler
});

//var jive = noProp;                                                                                        // cause ReferenceError if you want to test the handler above
 
exports.express = express;
exports.fs = fs;

exports.PUBLICDIR = PUBLICDIR;
exports.app = app;                                                                                          // !! maybe put all these into a single object, then export that object
exports.apiFirstStep = apiFirstStep;
exports.getSessionState = getSessionState;
exports.traverse = traverse;
exports.localIp = localIp;
exports.ecxExternal = ecxExternal;
exports.ecxInternal = ecxInternal;

exports.setStepSuccessResult = setStepSuccessResult;
exports.setStepFailResult = setStepFailResult;
exports.generateNameForUserIdentity = generateNameForUserIdentity;
exports.runSerialFuncs = runSerialFuncs;                                                                    
exports.setSerialStreamResult = setSerialStreamResult;
exports.convertReplSubpolicies = convertReplSubpolicies;
exports.arrayToHash = arrayToHash;
exports.getGenericEcxErrorResponse = getGenericEcxErrorResponse;
exports.sendRequestToEcx = sendRequestToEcx;
exports.validateQs = validateQs;
exports.returnBadRequestError = returnBadRequestError;
exports.startFunc = startFunc;
exports.addDoubleSpace = addDoubleSpace;


//start of my modules

const crUser = require('./cruser.js');
const regProvider = require('./regprovider');  
const getcompliance = require('./analytics/models/vmbackupslacom').getcompliance;  
const getdbcompliance = require('./analytics/models/dbbackupslacom').getdbcompliance;                                                          // register provider endpoint stuff 
const getbackuphistory = require('./analytics/models/vmbackuphistory').getbackuphistory;
const parsefilter = require('./analytics/models/vmbackuphistory').parsefilter;
const getdbbackuphistory = require('./analytics/models/dbbackuphistory').getdbbackuphistory;
const dbparsefilter = require('./analytics/models/dbbackuphistory').dbparsefilter;
const getBillinfo = require('./analytics/models/vmbackupbill').getBillinfo;
const regCloudProvider = require('./regCloudProvider');
const regSprof = require('./regstorageprofile');   
const updSprof = require('./updatestorageprofile');   
const getSla = require('./getslapolicy');
const getBackupSla = require('./getBackupSLAPolicy');
const getInstancesStatus = require('./getInstancesStatus');
const crHyProtPol = require('./crhyppolicy');
const crApProtPol = require('./createApplicationBackupPolicy');
const createAppRestorePol = require('./createApplicationRestorePolicy');
const crStProtPol = require('./crstppolicy');
const getJobSess = require('./getjobsessions');
const crHyRecovPol = require('./crhyrpolicy');
const regSmtpLdap = require('./regsmtpldap');
const updSmtpLdap = require('./updatesmtpldap');
const regVADP = require('./vadp-provider');
const updateVADP = require('./vadp-update-provider');
const applyOptions = require('./applyoptions');
const adhocRun = require('./adhocRun');
const applyPolicyOptions = require('./applyPolicyOptions');
const updProv = require('./updateproviderVE');                                                            //!! rename to lose the VE                                              
const updCloudProv = require('./updateCloudProvider'); 
const secPerms = require('./secperms');
const hyrJob = require('./hyrjob');
const appRestoreJob = require('./appRestoreJobs.js');
const crRfPol = require('./crrfpolicy.js');
const metrics = require('./metrics');
const pingURL = require('./pingURL');
const delAppServer  = require('./deleteAppServer')
const delScriptServer  = require('./deleteScriptServer')
const getAppServers  = require('./getAppServers')
const getScriptServers  = require('./getScriptServers')
const schedule = require('./hold-release-schedule');
//const steptest = require('./steptest');

//Export Constants
const ngpConstants = require('./constants');
exports.ngpConstants = ngpConstants;

//Export Utils
const ngpUtils = require('./commons/ngputils');
exports.ngpUtils = ngpUtils;

const ngpCoreHelper = require('./commons/ngp-core-helper');
exports.ngpCoreHelper = ngpCoreHelper;

/*
   ===================================
   LOAD ROUTERS FOR OTHER MODULES HERE
   ===================================
*/
//REPORTS
var report = require('./report/report-routes');
var reportProvider = require('./report/report-provider');

//CATALOG
var catalog = require('./catalog/catalog-routes');
var catalogProvider = require('./catalog/catalog-provider');

var cors = false;
var useHttps = false;                                                                                       // see ecxngp.yml

var startupArgs = {
	"--cfgfile": function(arg){
		cfgFile = arg;
	},
};

var cfgFile = __dirname + '/ecxngp.yml'                                                                     // override via --cfgfile
const sprofTplFile = PUBLICDIR + '/sproftpl.json';
var sprofTpl = null;                                                                                        // storage profile template 

const hyppolTplFile = PUBLICDIR + '/hyppoltpl.json';
var hyppolTpl = null;                                                                                       // hypervisor protection policy 

const applicationProtectionPolicyTemplateFile = PUBLICDIR + '/applicationProtectionPolicyTemplate.json';
var applicationProtectionPolicyTemplate = null;                                                                                       // hypervisor protection policy 

const applicationRestorePolicyTemplateFile = PUBLICDIR + '/applicationRestorePolicyTemplate.json';
var applicationRestorePolicyTemplate = null;                                                                                       // hypervisor protection policy 

const jobTplFile = PUBLICDIR + '/jobtpl.json';
var jobTpl = null;                                                                                          // prot policy job 

const stppolTplFile = PUBLICDIR + '/stppoltpl.json';
var stppolTpl = null;                                                                                       // storage protection policy 

const hyrpolTplFile = PUBLICDIR + '/hyrpoltpl.json';
var hyrpolTpl = null;

const triggerTplFile = PUBLICDIR + '/triggertpl.json';
var triggerTpl = null; 

const rfpolTplFile = PUBLICDIR + '/rfpoltpl.json';
var rfpolTpl = null; 

const rptpolTplFile = PUBLICDIR + '/rptpoltpl.json';
var rptpolTpl = null;

const versionFile = '/opt/ECX/release.json';

process.argv.shift();
process.argv.shift();

if ( process.argv.length > 0 )                                                                          // allow override of default configuration file
	getStartupArgs(process.argv);

var effectivePort = null;

getConfiguration();

exports.sprofTpl = sprofTpl;
exports.hyppolTpl = hyppolTpl;
exports.hyrpolTpl = hyrpolTpl;
exports.applicationProtectionPolicyTemplate = applicationProtectionPolicyTemplate;
exports.applicationRestorePolicyTemplate = applicationRestorePolicyTemplate;
exports.jobTpl = jobTpl;
exports.triggerTpl = triggerTpl;
exports.stppolTpl = stppolTpl;
exports.rfpolTpl = rfpolTpl;
exports.rptpolTpl = rptpolTpl;

const https = require('https');                                                                         // NB: even if configured for http only, need this to act as HTTPS client to ECX
const http = require('http');                                                                           // for ecxngp as ECX client during development; perhaps as server to NG and test.js

if ( process.env.HTTPS )                                                                                // already set from yml (but not via cmdline); so if ENV override
{
	if ( process.env.HTTPS == 'true' )
		useHttps = true;
	else                                                                                                  // if present but not true, then set false
		useHttps = false;		
}

var productionMode = false;                                                                             // true --> https + zuul routing

if ( process.env.PRODUCTION )
{
		if ( process.env.PRODUCTION == 'true' )
				productionMode = true;
}

if ( ! effectivePort )                                                                                  // FTI default port not overridden by 
	effectivePort = SECUREPORT;                                                                           // NB: always listen here, regardless of protocol

exports.ngpPort = effectivePort;

//logMsg('info', 'Listen port=' + effectivePort + ', HTTPS=' + useHttps);

if ( useHttps )
{
	const sslOpts = {
			key: fs.readFileSync('./key.pem'),                                                                // using self-signed for now; assumed to be in CWD
			cert: fs.readFileSync('./cert.pem')
		};

	const httpsServer = https.createServer(sslOpts, app);
	httpsServer.timeout = 600000;

	httpsServer.listen(effectivePort, function() {
		logger.log('info', 'Node-cdm-service started and listening on: https://' + getLocalIp() + ':%s', effectivePort);
		if( process.env.CORE_IP )
		{
		    logMsg('info', "Core(Endeavour) Hostname/IP Address: " + localIp);
		}
	});
}
else
{
	const httpServer = http.createServer(app);
	httpServer.timeout = 600000;

	httpServer.listen(effectivePort, function() {
		logger.log('info', 'Node-cdm-service started and listening on: http://' + getLocalIp() + ':%s', effectivePort);
		if( process.env.CORE_IP )
		{
		    logMsg('info', "Core(Endeavour) Hostname/IP Address: " + localIp);
		}
	});
}

var ecxSidHash = {};                                                                                    // key is ECX session ID established by NG2, value is current state for that NG2 
var ecxSitesHash = null;

exports.ecxSidHash = ecxSidHash;

app.options('*', function(req, res, next) {
	setCorsHeaders(res);
    next();
});


// Swagger endpoint configuration
const swaggerUi = require('swagger-ui-express');
const swaggerJSDoc = require('swagger-jsdoc');

var productInfo = null;

try {
	productInfo = JSON.parse(fs.readFileSync(versionFile, 'utf8'));
}
catch(e) {
	logMsg('warn', "Failed to read version file'" + versionFile + "' because of error=" + e.toString());
	productInfo = {'product': 'SpectrumProtect Plus', 'version': 'unknown'};
}	

const options = {
 	definition: {
		info: {
			title: productInfo.product,
			version: productInfo.version,
			description: `${productInfo.product} NGP endpoint documentation for version ${productInfo.version} build ${productInfo.build}`
		},
		securityDefinitions: {
			auth: {
				type: 'apiKey',
				name: 'x-endeavour-sessionid',
				in: 'header'
			}
		},
		security: [
			{ auth: [] }
		],
	},
	basePath: '/ngp',
	apis: ['./report/report-routes.js', '*.js'],
};

// End of Swagger configuration


// this app.use waypoint (Express middleware) is called before any of the real endpoints. I am using it to enforce the presence (where required) of the x-endeavour-sessionid header

app.use(function(req, res, next)
{
	addDoubleSpace();
	
	var clientIp = (req.headers['x-forwarded-for'] || req.connection.remoteAddress).match('[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$');  // ipv4 only & handles case where we're fronted by proxy

	logMsg('info', 'REQUEST received from client at IP=' + clientIp + ' ==> METHOD(%s) URL(%s) SESSION ID(%s)', req.method, req.url, req.get(ECXSESSIONHEADER));

	if ( req.method == 'OPTIONS' )                                                                        // OPTIONS is browser-generated part of CORS pre-flight; will contain no required headers
	{
		return next();
	}

/* 
	With arrival of Gozer's advance man, er...dog, no static content served by ecxngp	
	if ( ! req.url.contains(apiFirstStep)  )                                                              // FTI request for static content; no required header
	{
		return next();                                                                                      // let express.static waypoint handle it; we've logged it here
	} 
*/

	if ( req.url.contains(apiFirstStep + 'version') )                                                  // require no headers for this
	{
		return next();                                                                                      // no requirement for a session ID header
	} 

	if ( ! req.get(ECXSESSIONHEADER) )                                                                   // FTI non-static request lacks required header	
	{
		logMsg('info', req.method + ' request for ' + req.url + ' lacks required header');
		returnMissingSidError(res);
		return;
	} 

	var ecxSid = req.get(ECXSESSIONHEADER);

	var ecxLocaleHeader = 'en-us';

	if(req.get(ECXLOCALEHEADER))
	ecxLocaleHeader = req.get(ECXLOCALEHEADER);
	
	getEcxLocaleHeader.set('ecxLocale',ecxLocaleHeader);
	
	if ( ! ecxSidHash[ecxSid] )                                                                           // FTI new NG2 with new ECX sid
	{
		app.set(ecxSid, initNgState(ecxSid));                                                               // point sid to virgin session state; available via app.get and...
		ecxSidHash[ecxSid] = app.get(ecxSid);                                                               // ...via ref to hash
		ecxSidHash[ecxSid].reqProtocol = req.protocol;

		if(req.headers['x-forwarded-host'])
		{
			ecxSidHash[ecxSid].host = req.headers['x-forwarded-host'];
			
		}
		else {
			var hostVal = req.headers['host'].split(':');
			ecxSidHash[ecxSid].host = hostVal[0];
			
		}
		
		ecxSidHash[ecxSid].piggybackRes = res;                                                              // capture *any* (latest) response for certain kernel promulgated errors
		
		logMsg('info', 'Established session ID=' + ecxSid);		

		var reqId = crypto.createHash('md5').update(req.url).update(req.method).update(new Date().getTime().toString()).digest('hex');
		req.reqId = reqId;
		res.reqId = reqId;
		
		logMsg('info', 'Established request ID=' + reqId + ' for request=' + req.url + ' in session ID=' + ecxSid);		
		
		getEcxSites(ecxSid, function(rc){                                                                   // either it works or it don't   !! MAYBE get ECX identities in same way                             
				next();				
		});		
	}	
	else                                                                                                 
	{
		var reqId = crypto.createHash('md5').update(req.url).update(req.method).update(new Date().getTime().toString()).digest('hex');                                        
		logMsg('info', 'Established request ID=' + reqId + ' for request=' + req.url + ' in session ID=' + ecxSid);		

		req.reqId = reqId;
		res.reqId = reqId;

		if(req.headers['x-forwarded-host'])
		{
			ecxSidHash[ecxSid].host = req.headers['x-forwarded-host'];
			
		}
		else {
			var hostVal = req.headers['host'].split(':');
			ecxSidHash[ecxSid].host = hostVal[0];
			
		}
		
		ecxSidHash[ecxSid].piggybackRes = res; 
	
		next();
	}
});


app.use(bodyParser.json({limit: '50mb'}));
app.use(bodyParser.urlencoded({limit: '50mb', extended: true}));
// This waypoint handles modifying the response body to conform to UI expectations.

app.use(modresp.json(                                                                                   //!!! revisit when CORS is active; body is typically the composite object from sendRequestToEcx
    function transform(body, req, res) {
		if ( body.statusCode && body.response )                                                             // FTI quite likely that this response came from sendRequestToEcx [you could always add unique marker]
		{
			//logMsg('warn', 'ORIGINAL response=' + JSON.stringify(body));

			if ( body.statusCode == 200 )                                                                     // per RS, return body.response *only*
			{
				body = body.response;
			}
		}

  		//logMsg('warn', 'Modified response=' + JSON.stringify(body));

        return body;
    }
));

/*
  ========================================
  Add middleware for other components here
  ========================================
*/

const swaggerSpec = swaggerJSDoc(options);

app.use(apiFirstStep + 'api-docs', swaggerUi.serve, swaggerUi.setup(swaggerSpec));
app.get(apiFirstStep + 'api-docs.json', (req, res) => {
	res.setHeader('Content-Type', 'application/json');
	res.send(swaggerSpec);
});

app.use(apiFirstStep + 'report', report);
app.use(apiFirstStep + 'catalog', catalog);

const licitQueries = {                                                                                  // !! probably qualify by HTTP verb, too
	report: {
		action: ['schedule'],
	},
	hypervisor: {
		action: ['applySLAPolicies', 'restore', 'applyOptions', 'restorefile', 'adhoc'],                     
	},
	hypervisor_job_get: {
		subtype: ['vmware', 'hyperv', 'awsec2'],
		filter: [/.*/,],
	},
	hypervisor_job_put: {
		action: ['restore',],
	},	
	hypervisor_job_delete: {
		name: [/.*/,],
	},				
	storage: {
		action: ['applySLAPolicies'],
	},
	appserver: {
		action: ['applySLAPolicies','findDiscovery'],
		appServerId: [/\d+/,],
	},
	scriptserver: {
		action: ['applySLAPolicies'],
	},
	application: {
		action: ['applySLAPolicies','applyOptions', 'restore', 'adhoc'],
	},
	application_job_get: {
		subtype: ['sql', 'oracle', 'db2', 'exch', 'office365', 'mongo', 'k8s', 'openshift', 'file'],
		filter: [/.*/,],
	},
	application_job_delete: {
		name: [/.*/,],
	},				
	slapolicy: {                                                                                          // as in GET slapolicy
		status: ['true', 'false'],                                                                          // if true, type must be present; else type must be absent
		action: ['applyPolicyOptions'],
		subtype: ['vmware', 'hyperv', 'awsec2', 'storage', 'sql', 'oracle',
				'saphana', 'iscache','catalog', 'db2', 'exch','office365', 'mongo', 'k8s', 'openshift', 'file'],                   // modify as new providers, apps come along !! ARE WE SURE ABOUT THESE TYPES?
		name: [/.*/,],
		embedoptions: ['true', 'false'],
		slaType: ['backup', 'snapshot_backup', 'snapshot']
	},	
	cloud: {                                                                                          // as in GET slapolicy
		bucket: ['true', 'false'],                                                                          // if true, type must be present; else type must be absent
	},	
	security: {
		action: ['add', 'update'],
		resourcePoolId: [/\d+/,],
	},
	schedule: {
		action: ['hold', 'release'],
	},	
};


// ***** Start of ALL ENDPOINTS (except for reports) *****

// Provider registration endpoints

/**
 * @swagger
 * /ngp/security/user:
 *   post:
 *     tags: ["Security"]
 */
app.post(apiFirstStep + 'security/user', function(req, res) {  
	stepsStream = crUser.init(req, res);                                                  
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/security/user/{id}:
 *   put:
 *     tags: ["Security"]
 */
app.put(apiFirstStep + 'security/user/:id', function(req, res) {  
	stepsStream = crUser.init(req, res);                                                  
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/security/user/{id}:
 *   delete:
 *     tags: ["Security"]
 */
app.delete(apiFirstStep + 'security/user/:id', function(req, res) {  
	var qsO = validateQs(req, licitQueries.security);
	
	if ( ! qsO.keyVals.resourcePoolId )
	{
		logMsg('error', 'Invalid query string specification');   		
		returnBadRequestError(res); 
    return;
	}

	stepsStream = crUser.init(req, res);                                                  
	startFunc(req, res, stepsStream);
});

app.post(apiFirstStep + 'vadp', function(req, res) {  
	stepsStream = regVADP.init(req, res);                                                  
	startFunc(req, res, stepsStream);
});

app.put(apiFirstStep + 'vadp/:id', function(req, res) {  
	stepsStream = updateVADP.init(req, res);                                                  
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/hypervisor:
 *   post:
 *     tags: ["Hypervisor"]
 */
app.post(apiFirstStep + 'hypervisor', function(req, res) {                                                                                                     
	var actionO = validateQs(req, licitQueries.hypervisor);
	
	if ( actionO.rc == 2 )                                                                                // FTI valid key with invalid value 	                                                                                                                                
	{
		logMsg('error', 'Detected invalid value among keys=' + JSON.stringify(actionO.keyVals));
		returnBadRequestError(res); 
    return;
	}

	var stepsStream = null;   

	if ( actionO.rc == 0 )                                                                                // FTI no action which means register the hypervisor
	{
		stepsStream = regProvider.init(req, res, 'hypervisor');                                             // 3rd arg is my resourceType; construe common 'type' property in context of resourceType     	
		startFunc(req, res, stepsStream);
	}
	else if ( actionO.keyVals.action == 'applySLAPolicies' )
	{
		stepsStream = crHyProtPol.init(req, res);                                                                  
		startFunc(req, res, stepsStream);	
	}
	else if ( actionO.keyVals.action == 'applyOptions' )
	{
		stepsStream = applyOptions.init(req, res, 'hypervisor');                                                                  
		startFunc(req, res, stepsStream);	
	}	
	else if ( actionO.keyVals.action == 'adhoc' )
	{
		stepsStream = adhocRun.init(req, res, 'hypervisor');                                                                  
		startFunc(req, res, stepsStream);	
	}	
	else if ( actionO.keyVals.action == 'restore' )
	{
		stepsStream = crHyRecovPol.init(req, res);                                                                  
		startFunc(req, res, stepsStream);	
	}	
	else if ( actionO.keyVals.action == 'restorefile' )
		{
			stepsStream = crRfPol.init(req, res);                                                                  
			startFunc(req, res, stepsStream);	
		}		
});

/**
 * @swagger
 * /ngp/cloud:
 *   post:
 *     tags: ["Cloud"]
 */
app.post(apiFirstStep + 'cloud', function(req, res) {   
	var actionO = validateQs(req, licitQueries.cloud);
	
	var getBucketsList = 'false';

	var stepsStream = null;   
		
	if(actionO.keyVals.bucket)
	getBucketsList = actionO.keyVals.bucket;

	stepsStream = regCloudProvider.init(req, res, getBucketsList);                                             // 3rd arg is my resourceType; construe common 'type' property in context of resourceType     	
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/hypervsior/{id}:
 *   put:
 *     tags: ["Hypervisor"]
 */
app.put(apiFirstStep + 'hypervisor/:id', function(req, res) {  
	stepsStream = updProv.init(req, res, 'hypervisor');                                                   // 3rd arg is my resourceType; construe common 'type' property in context of resourceType     	
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/cloud/{id}:
 *   put:
 *     tags: ["Cloud"]
 */
app.put(apiFirstStep + 'cloud/:id', function(req, res) {  
	stepsStream = updCloudProv.init(req, res);                                                   // 3rd arg is my resourceType; construe common 'type' property in context of resourceType     	
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/storage:
 *   post:
 *     tags: ["Storage"]
 */
app.post(apiFirstStep + 'storage', function(req, res) {  
	var actionO = validateQs(req, licitQueries.storage);
	
	if ( actionO.rc == 2 )      	                                                                                                                                
	{
		logMsg('error', 'Detected invalid value among keys=' + JSON.stringify(actionO.keyVals));
		returnBadRequestError(res); 
    return;
	}

	var stepsStream = null;   

	if ( actionO.rc == 0 )                                                                                // FTI no action which means register the storage provider
	{
		stepsStream = regProvider.init(req, res, 'storage');                                                // 3rd arg is my resourceType; construe common 'type' property in context of resourceType     	
		startFunc(req, res, stepsStream);
	}
	else if ( actionO.keyVals.action == 'applySLAPolicies' )
	{
		stepsStream = crStProtPol.init(req, res);                                                                  
		startFunc(req, res, stepsStream);	
	}	
});

/**
 * @swagger
 * /ngp/storage/{id}:
 *   put:
 *     tags: ["Storage"]
 */
app.put(apiFirstStep + 'storage/:id', function(req, res) {                                          
	stepsStream = updProv.init(req, res, 'storage');                                                                  
	startFunc(req, res, stepsStream);		
});

/**
 * @swagger
 * /ngp/appserver:
 *   post:
 *     tags: ["Application Server"]
 */
app.post(apiFirstStep + 'appserver', function(req, res) {                                                   // !!! NO type info in POST body for this endpoint (5/25/17)                                                                                     
	var actionO = validateQs(req, licitQueries.appserver);
	
	if ( actionO.rc == 2 )      	                                                                                                                                
	{
		logMsg('error', 'Detected invalid value among keys=' + JSON.stringify(actionO.keyVals));
		returnBadRequestError(res); 
    return;
	}

	var stepsStream = null;   

	if ( actionO.rc == 0 )                                                                                  // FTI no action which means register the hypervisor
	{
		stepsStream = regProvider.init(req, res, 'appserver');                                           // 3rd arg is my resourceType; construe common 'type' property in context of resourceType     	
		startFunc(req, res, stepsStream);
	}
	else if ( actionO.keyVals.action == 'findDiscovery' )
	{
		stepsStream = getInstancesStatus.init(req, res);                                                                  
		startFunc(req, res, stepsStream);	
	}	
});

/**
 * @swagger
 * /ngp/appserver/{id}:
 *   put:
 *     tags: ["Application Server"]
 */
app.put(apiFirstStep + 'appserver/:id', function(req, res) {                                          
	stepsStream = updProv.init(req, res, 'appserver');                                                                  
	startFunc(req, res, stepsStream);		
});

app.delete(apiFirstStep + 'appserver/:id', function(req, res) {  
	stepsStream = delAppServer.init(req, res);                                                  
	startFunc(req, res, stepsStream);
});

app.get(apiFirstStep + 'appserver', function(req, res) {  
	stepsStream = getAppServers.init(req, res);                                                  
	startFunc(req, res, stepsStream);
});

app.get(apiFirstStep + 'scriptserver', function(req, res) {  
	stepsStream = getScriptServers.init(req, res);                                                  
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/scriptserver:
 *   post:
 *     tags: ["Script Server"]
 */
app.post(apiFirstStep + 'scriptserver', function(req, res) {                                                   // !!! NO type info in POST body for this endpoint (5/25/17)                                                                                     
	var actionO = validateQs(req, licitQueries.scriptserver);
	
	if ( actionO.rc == 2 )      	                                                                                                                                
	{
		logMsg('error', 'Detected invalid value among keys=' + JSON.stringify(actionO.keyVals));
		returnBadRequestError(res); 
    return;
	}

	var stepsStream = null;   

	if ( actionO.rc == 0 )                                                                                  // FTI no action which means register the hypervisor
	{
		stepsStream = regProvider.init(req, res, 'scriptserver');                                           // 3rd arg is my resourceType; construe common 'type' property in context of resourceType     	
		startFunc(req, res, stepsStream);
	}
});

/**
 * @swagger
 * /ngp/scriptserver/{id}:
 *   put:
 *     tags: ["Script Server"]
 */
app.put(apiFirstStep + 'scriptserver/:id', function(req, res) {                                          
	stepsStream = updProv.init(req, res, 'scriptserver');                                                                  
	startFunc(req, res, stepsStream);		
});

app.delete(apiFirstStep + 'scriptserver/:id', function(req, res) {  
	stepsStream = delScriptServer.init(req, res);                                                  
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/application:
 *   post:
 *     tags: ["Application"]
 */
app.post(apiFirstStep + 'application', function(req, res) {                                                   // !!! NO type info in POST body for this endpoint (5/25/17)                                                                                     
	var actionO = validateQs(req, licitQueries.application);
	
	if ( actionO.rc == 2 )      	                                                                                                                                
	{
		logMsg('error', 'Detected invalid value among keys=' + JSON.stringify(actionO.keyVals));
		returnBadRequestError(res); 
    return;
	}

	var stepsStream = null;   

	if ( actionO.keyVals.action == 'applySLAPolicies' )	
	{
		stepsStream = crApProtPol.init(req, res);                                                                
		startFunc(req, res, stepsStream);	
	}	
	else if ( actionO.keyVals.action == 'applyOptions' )
	{
		stepsStream = applyOptions.init(req, res, 'application');                                                                  
		startFunc(req, res, stepsStream);	
	}
	else if ( actionO.keyVals.action == 'adhoc' )
	{
		stepsStream = adhocRun.init(req, res, 'application');                                                                  
		startFunc(req, res, stepsStream);	
	}	
	else if ( actionO.keyVals.action == 'restore' )
	{
		stepsStream = createAppRestorePol.init(req, res);                                                                  
		startFunc(req, res, stepsStream);	
	}	
});

// Storage profile ("SLA Policy") registration endpoints

/**
 * @swagger
 * /ngp/slapolicy:
 *   post:
 *     tags: ["SLA Policy"]
 *     description: Creates a new SLA policy or applies options to an existing SLA policy (if the `?action=applyPolicyOptions` query parameter is included).
 *     parameters:
 *       - in: query
 *         name: action
 *         required: false
 *         type: string
 *         enum: [applyPolicyOptions]
 *         description: >
 *           Applies a change to the policy options for an existing SLA policy.<br /><br />
 *
 *           When using this flag to associated a script policy to an existing SLA policy, 
 *           the request body should resemble the following:<br /><br />
 *          
 *           <pre>{"script": {"preGuest":null,"postGuest":null,"continueScriptsOnError":false},"policyOption":{"excluderesources":"","forcebaseresources":"","runInventory":false}}</pre><br /><br />
 *
 *           When changing the general options for an existing resource, the request body
 *           should resemble the following:<br /><br />
 *
 *           <pre>{"resources":[{"href":"https://71.100.14.11/api/application/oracle/instance/37573786e1ec1f871c7b96ca28014ce0/database/3968fd3fd2476b0c4151f71979c5b26f?from=hlo","id":"3968fd3fd2476b0c4151f71979c5b26f","metadataPath":"/DBINST:37573786e1ec1f871c7b96ca28014ce0/DB:3968fd3fd2476b0c4151f71979c5b26f"}],"subtype":"oracle","options":{"maxParallelStreams":1,"dbFilesForParallelStreams":"SINGLE_FILE","backupPreferredNode":"","logbackup":{"purgePrimaryLogs":false,"primaryLogRetentionDays":3,"performlogbackup":true,"rpo":{"frequency":1,"frequencyType":"DAY","triggerTime":"12:00:00 AM","metadata":{"activateDate":1540267200000}}}}}</pre><br /><br />
 *
 *           When associating an SLA policy with a specific resource, the request body should resemble the following:<br /><br />
 *
 *           <pre>{"subtype":"oracle","version":"1.0","resources":[{"href":"https://71.100.14.11/api/application/oracle/instance/37573786e1ec1f871c7b96ca28014ce0/database/3968fd3fd2476b0c4151f71979c5b26f?from=hlo","id":"3968fd3fd2476b0c4151f71979c5b26f","metadataPath":"/DBINST:37573786e1ec1f871c7b96ca28014ce0/DB:3968fd3fd2476b0c4151f71979c5b26f"}],"slapolicies":[{"href":"https://71.100.14.11/api/spec/storageprofile/2002","id":"2002","name":"Bronze"}]}</pre><br /><br />
 *       - in: query
 *         name: name
 *         required: false
 *         type: string
 *         description: The name of the SLA policy to change (should be used with the `applyPolicyOptions` flag).
 *       - in: query
 *         name: subtype
 *         type: string
 *         enum: [vmware, hyperv, storage, sql, oracle, saphana, iscache, catalog, db2, exch, mongo]
 *         description: The desired SLA policy subtype (should be used with the `applyPolicyOptions` flag).
 *       - in: body
 *         required: false
 *         type: object
 *         description: The SLA policy to create. This body should be included only when the `applyPolicyOptions` flag is <em>not</em> set.
 *         schema:
 *           type: object
 *           properties:
 *             name:
 *               type: string
 *             version:
 *               type: string
 *             spec:
 *               type: object
 *               properties:
 *                 simple:
 *                   type: boolean
 *                 subpolicy:
 *                   type: array
 *                   items:
 *                     type: object
 *                     properties:
 *                       type:
 *                         type: string
 *                       software:
 *                         type: string
 *                       retention:
 *                         type: object
 *                         properties:
 *                           age:
 *                             type: integer
 *                       useEncryption:
 *                         type: boolean
 *                       trigger:
 *                         type: object
 *                         properties:
 *                           frequency:
 *                             type: integer
 *                           type:
 *                             type: string
 *                           activateDate:
 *                             type: integer
 *                       site:
 *                         type: string
 *     responses:
 *       200:
 *         description: The applied options change was successful
 *       201:
 *         description: The desired SLA policy was created
 */
app.post(apiFirstStep + 'slapolicy', function(req, res) {       	                                                                                                     
	var stepsStream = null;   
	var qsO = validateQs(req, licitQueries.slapolicy);
	var statusSpec = null;
	var slaPolicyName = null;

	if ( qsO.keyVals.action == 'applyPolicyOptions' )
    { 
		slaPolicyName = qsO.keyVals.name;
		statusSpec = qsO.keyVals.subtype;
		stepsStream = applyPolicyOptions.init(req, res, slaPolicyName, statusSpec);                                                                  
		startFunc(req, res, stepsStream);	
    }	
	else
	{
	stepsStream = regSprof.init(req, res);                                                                  // NB: this registrar doesn't require any 'type' qualifier	
	startFunc(req, res, stepsStream);		
	}
});

/**
 * @swagger
 * /ngp/slapolicy/{id}:
 *   put:
 *     tags: ["SLA Policy"]
 *     parameters:
 *       - name: id
 *         in: path
 *         required: true
 *         description: ID of the SLA policy to update
 *     responses:
 *       200:
 *         description: SLA policy associated with the supplied ID was successfully updated
 *       404:
 *         description: No SLA policy found with the supplied ID
 */
app.put(apiFirstStep + 'slapolicy/:id', function(req, res) {                                              
	var stepsStream = null;   

	stepsStream = updSprof.init(req, res);                                                                  
	startFunc(req, res, stepsStream);		
});

/**
 * @swagger
 * /ngp/slapolicy:
 *   get:
 *     tags: ["SLA Policy"]
 *     parameters:
 *       - in: query
 *         name: pageSize
 *         type: integer
 *         description: The maximum number of entries returned
 *       - in: query
 *         name: sort
 *         description: >
 *           The sorting criteria for the returned entries. This criteria is a JSON 
 *           object based on the properties associated with an SLA policy. For example:<br /><br /> 
 *
 *           &nbsp;&nbsp;&nbsp;&nbsp;`?sort=%5B%7B%22property%22:%22name%22,%22direction%22:%22ASC%22%7D%5D`<br /><br />
 *
 *           (escaped form of `?sort={"propety": "name", "direction": "ACS"}`). Note that all
 *           special characters must be escaped.
 *       - in: query
 *         name: subtype
 *         type: string
 *         enum: [vmware, hyperv, storage, sql, oracle, saphana, iscache, catalog, db2, exch, mongo]
 *         description: The desired SLA policy subtype.
 *       - in: query
 *         name: status
 *         type: string
 *         enum: [true, false]
 *         description: Boolean flag denoting if the status associated with each SLA policy should be included in the response.
 *       - in: query
 *         name: embedoptions
 *         type: string
 *         enum: [true, false]
 *         description: Boolean flag denoting if the options associated with each SLA policy should be included in the response.
 *       - in: query
 *         name: name
 *         type: string
 *         description: The desired name of the SLA policy
 *     responses:
 *       200:
 *         description: All existing SLA policies were successfully returned
 *       500:
 *         description: An error occurred while retrieving the existing SLA policies
 */
app.get(apiFirstStep + 'slapolicy', function(req, res) { 
	var pageSize = req.query.pageSize;
	var slaType = undefined
	var pageStartIndex = null;

	if(req.query.pageStartIndex)
	pageStartIndex = req.query.pageStartIndex;

	var qsO = validateQs(req, licitQueries.slapolicy);
	if ( qsO.rc == 2 )      	                                                                                                                                
	{
		logMsg('error', 'Detected invalid value among keys=' + JSON.stringify(qsO.keyVals));
		returnBadRequestError(res); 
    return;
	}
	var subtype = undefined;
	var statusSpec = null;
	var embedOptions = null;
	
	if ( qsO.rc == 1 )                                                                                      // all we know is at least one valid k=v present, but could be mere status=false, or other, illicit combo
	{
		if ( qsO.numKv == 1 )                                                                               // only status=false permitted
		{
			// if ( ! qsO.keyVals.status || qsO.keyVals.status != 'false' )                                    // FTI it's a different key, e.g., subtype which ain't valid by itself; or it's present but not false
			// {
			// 	logMsg('error', '---------------------------------------');
			// 	logMsg('error', qsO.keyVals)
			// 	logMsg('error', '---------------------------------------');
			// 	logMsg('error', 'Invalid query string specification');                                      // hard to be more specific than this cuz other keys may be added in future; thus dependencies obscure
			// 	returnBadRequestError(res); 
    		// return;
			// }
			slaType = qsO.keyVals.slaType;
		}                                                                                                   // call getSla with null statusSpec; status=false effectively ignored
		else if ( qsO.numKv == 2 )                                                                          // FTI both status & subtype present. status=true is required
		{
			if ( qsO.keyVals.status != 'true' )
			{
				logMsg('error', 'Invalid query string specification');                                      // hard to be more specific than this cuz other keys may be added in future; thus dependencies obscure
				returnBadRequestError(res); 
    		return;
			}
			else
			{
				statusSpec = qsO.keyVals.subtype;                                                            // e.g., vmware
			}					
		}
		else if ( qsO.numKv == 3 )                                                                          // FTI both status & subtype present. status=true is required
		{
			if ( qsO.keyVals.status != 'true' )
			{
				logMsg('error', 'Invalid query string specification');                                      // hard to be more specific than this cuz other keys may be added in future; thus dependencies obscure
				returnBadRequestError(res); 
    		return;
			}
			else
			{
				statusSpec = qsO.keyVals.subtype;                                                            // e.g., vmware
				embedOptions = qsO.keyVals.embedoptions;
			}
		}
	}
	var stepsStream = null;   

	//throw new ReferenceError("GOODNIGHT IRENE"); // try to cause EXCEPTION (a thrown error)          // test exception and error handling out of the Express stack
	  //var a = noProp; // this causes a reference ERROR
	if(statusSpec != null)
	stepsStream = getBackupSla.init(req, res, statusSpec,embedOptions,pageSize,pageStartIndex); 
	else
	stepsStream = getSla.init(req, res, statusSpec,null,embedOptions, slaType);

	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/slapolicy:
 *   delete:
 *     tags: ["SLA Policy"]
 *     parameters:
 *       - name: name
 *         in: query
 *         required: true
 *         description: The of the SLA policy to delete.
 *     responses:
 *       204:
 *         description: SLA policy associated with the supplied name was successfully deleted
 *       404:
 *         description: No SLA policy found with the supplied name
 */
app.delete(apiFirstStep + 'slapolicy/', function(req, res){
		var stepsStream = null; 
		var actionO = validateQs(req, licitQueries.slapolicy);
		
		stepsStream = getSla.init(req, res, null, actionO.keyVals.name, null);                                                       
		startFunc(req, res, stepsStream);	
});


app.post(apiFirstStep + 'schedule/', function(req, res) {
	var stepsStream = null; 
	var actionO = validateQs(req, licitQueries.schedule);
	if ( actionO.rc === 2 ) {
		logMsg('error', 'Detected invalid value among keys=' + JSON.stringify(actionO.keyVals));
		returnBadRequestError(res); 
    	return;
	}
	stepsStream = schedule.init(req, res, actionO.keyVals.action);                                                       
	startFunc(req, res, stepsStream);	
});


/**
 * @swagger
 * /ngp/slapolicy/jobsessions:
 *   get:
 *     tags: ["SLA Policy"]
 *     description: Obtain the job sessions associated with an SLA policy.
 *     parameters:
 *       - in: query
 *         name: subtype
 *         type: string
 *         enum: [vmware, hyperv, storage, sql, oracle, saphana, iscache, catalog, db2, exch, mongo]
 *         description: The desired SLA policy subtype.
 *       - in: query
 *         name: name
 *         type: string
 *         description: The desired name of the SLA policy
 *     responses:
 *       200:
 *         description: Job sessions associated with the known SLA policies were successfully obtained
 */
app.get(apiFirstStep + 'slapolicy/jobsessions', function(req, res) { 
	var qsO = validateQs(req, licitQueries.slapolicy);  //!! name & subtype. can't check for name cuz it could be the name of any sprof (not just Gold, Silver, Bronze) OR use Rx

	if ( qsO.rc == 2 )      	                                                                                                                                
	{
		logMsg('error', 'Detected invalid value among keys=' + JSON.stringify(qsO.keyVals));
		returnBadRequestError(res); 
    return;
	}

	if ( qsO.rc == 0 || qsO.numKv != 2 )      	                                                             // NB: *presently* expect both name & subtype
	{
		logMsg('warn', 'Missing keys among keys=' + JSON.stringify(qsO.keyVals));
		returnBadRequestError(res); 
    return;
	}
	
	var stepsStream = null;   

	stepsStream = getJobSess.init(req, res, qsO.keyVals);                                                                 
	startFunc(req, res, stepsStream);
})

/**
 * @swagger
 * /ngp/smtp:
 *   post:
 *     tags: ["SMTP"]
 *     responses:
 *       200:
 *         description: Successfully obtained all SMTP instances
 */
app.post(apiFirstStep + 'smtp', function(req, res) {                                                                                                     
	stepsStream = regSmtpLdap.init(req, res, 'smtp');                                                            	
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/ldap:
 *   post:
 *     tags: ["LDAP"]
 *     responses:
 *       200:
 *         description: Successfully obtained all LDAP instances
 */
app.post(apiFirstStep + 'ldap', function(req, res) {                                                                                                     
	stepsStream = regSmtpLdap.init(req, res, 'ldap');                                                            	
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/smtp/{id}:
 *   put:
 *     tags: ["SMTP"]
 *     parameters:
 *       - name: id
 *         in: path
 *         required: true
 *         description: ID of the SMTP instance to update
 *     responses:
 *       200:
 *         description: Successfully obtained the desired SMTP instance
 */
app.put(apiFirstStep + 'smtp/:id', function(req, res) {                                                                                                     
	stepsStream = updSmtpLdap.init(req, res, 'smtp');                                                            	
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/ldap/{id}:
 *   put:
 *     tags: ["LDAP"]
 *     parameters:
 *       - name: id
 *         in: path
 *         required: true
 *         description: ID of the LDAP instance to update
 *     responses:
 *       200:
 *         description: Successfully obtained the desired LDAP instance
 */
app.put(apiFirstStep + 'ldap/:id', function(req, res) {    
	stepsStream = updSmtpLdap.init(req, res, 'ldap');                                                            	
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/security/permissions:
 *   post:
 *     tags: ["Security"]
 */
app.post(apiFirstStep + 'security/permissions', function(req, res) {                                                                                                     
	var qsO = validateQs(req, licitQueries.security);  
	
		if ( qsO.rc == 2 )      	                                                                                                                                
		{
			logMsg('error', 'Detected invalid value among keys=' + JSON.stringify(qsO.keyVals));
			returnBadRequestError(res); 
			return;
		}
	
		if ( qsO.rc == 0 || qsO.numKv != 1 )      	                                                        // NB: presently either add or update
		{
			logMsg('warn', 'Missing keys among keys=' + JSON.stringify(qsO.keyVals));
			returnBadRequestError(res); 
			return;
		}
		
		var stepsStream = null;   

		stepsStream = secPerms.init(req, res, qsO.keyVals);                                                                 
		startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/hypervisor/job:
 *   get:
 *     tags: ["Hypervisor"]
 */
app.get(apiFirstStep + 'hypervisor/job', function(req, res) { 
	var qsO = validateQs(req, licitQueries.hypervisor_job_get);  

//	if ( qsO.numKv != 1 )      	                                                                             // must be exactly one of subtype
	if ( qsO.numKv < 1 || qsO.numKv > 2 )                                                                    // at least 1 (subtype), never more than 2
	{
		logMsg('error', 'Detected missing or invalid keys in URL query string');
		returnBadRequestError(res); 
    return;
	}

	var stepsStream = null;   

	stepsStream = hyrJob.init(req, res, qsO.keyVals);                                                                 
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/analytics/vmbackuphistory:
 *   get:
 *     tags: ["Analytics"]
 */
app.get(apiFirstStep + 'analytics/vmbackuphistory', function(req, res, next) { 
	var filters = parsefilter(req.query,next)
	getbackuphistory(filters, res, req, next);
	//test                                                                 
	//startFunc(req, res, stepsStream);

});

/**
 * @swagger
 * /ngp/analytics/dbbackuphistory:
 *   get:
 *     tags: ["Analytics"]
 */
app.get(apiFirstStep + 'analytics/dbbackuphistory', function(req, res, next) { 
	var filters = dbparsefilter(req.query,next)
	getdbbackuphistory(filters, res, req, next);
});

/**
 * @swagger
 * /ngp/analytics/vmbackupslacom:
 *   get:
 *     tags: ["Analytics"]
 */
app.get(apiFirstStep + 'analytics/vmbackupslacom', function(req, res, next) { 
	getcompliance(res,req,next);                                                                 
});

/**
 * @swagger
 * /ngp/analytics/dbbackupslacom:
 *   get:
 *     tags: ["Analytics"]
 */
app.get(apiFirstStep + 'analytics/dbbackupslacom', function(req, res, next) { 
	getdbcompliance(res,req,next);                                                                 
});

/**
 * @swagger
 * /ngp/analytics/vmbackupdetail:
 *   get:
 *     tags: ["Analytics"]
 */
app.get(apiFirstStep + 'analytics/vmbackupdetail', function(req, res, next) { 
	getBillinfo(res,req,next,true);                                                                 
});

/**
 * @swagger
 * /ngp/analytics/applicationbackupdetail:
 *   get:
 *     tags: ["Analytics"]
 */
app.get(apiFirstStep + 'analytics/applicationbackupdetail', function(req, res, next) { 
	getBillinfo(res,req,next,false);                                                                 
});

/**
 * @swagger
 * /ngp/application/job:
 *   get:
 *     tags: ["Application"]
 */
app.get(apiFirstStep + 'application/job', function(req, res) { 
	var qsO = validateQs(req, licitQueries.application_job_get);  

//	if ( qsO.numKv != 1 )      	                                                                             // must be exactly one of subtype
	if ( qsO.numKv < 1 || qsO.numKv > 2 )                                                                    // at least 1 (subtype), never more than 2
	{
		logMsg('error', 'Detected missing or invalid keys in URL query string');
		returnBadRequestError(res); 
    return;
	}

	var stepsStream = null;   

	stepsStream = appRestoreJob.init(req, res, qsO.keyVals);                                                                 
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/application/job:
 *   delete:
 *     tags: ["Application"]
 */
app.delete(apiFirstStep + 'application/job', function(req, res) { 
	var qsO = validateQs(req, licitQueries.application_job_delete);  

	if ( qsO.numKv != 1 )      	                                                                             // must be exactly one of name
	{
		logMsg('error', 'Detected missing or invalid keys in URL query string');
		returnBadRequestError(res); 
    return;
	}

	var stepsStream = null;   

	stepsStream = appRestoreJob.init(req, res, qsO.keyVals);                                                                 
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/application:
 *   put:
 *     tags: ["Application"]
 */
app.put(apiFirstStep + 'application', function(req, res) { 
	var qsO = validateQs(req, licitQueries.application);  
	if ( qsO.numKv != 1 )      	                                                                             // must be exactly one of action
	{
		logMsg('error', 'Detected missing or invalid keys in URL query string');
		returnBadRequestError(res); 
    return;
	}

	var stepsStream = null;   

	stepsStream = appRestoreJob.init(req, res, qsO.keyVals);                                                                 
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/hypervisor:
 *   put:
 *     tags: ["Hypervisor"]
 */
app.put(apiFirstStep + 'hypervisor', function(req, res) { 
	var qsO = validateQs(req, licitQueries.hypervisor);  
	if ( qsO.numKv != 1 )      	                                                                             // must be exactly one of action
	{
		logMsg('error', 'Detected missing or invalid keys in URL query string');
		returnBadRequestError(res); 
    return;
	}

	var stepsStream = null;   

	stepsStream = hyrJob.init(req, res, qsO.keyVals);                                                                 
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/hypervisor/job:
 *   delete:
 *     tags: ["Hypervisor"]
 */
app.delete(apiFirstStep + 'hypervisor/job', function(req, res) { 
	var qsO = validateQs(req, licitQueries.hypervisor_job_delete);  

	if ( qsO.numKv != 1 )      	                                                                             // must be exactly one of name
	{
		logMsg('error', 'Detected missing or invalid keys in URL query string');
		returnBadRequestError(res); 
    return;
	}

	var stepsStream = null;   

	stepsStream = hyrJob.init(req, res, qsO.keyVals);                                                                 
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/metrics:
 *   get:
 *     tags: ["Miscellaneous"]
 */
app.get(apiFirstStep + 'metrics', function(req, res) { 
	var stepsStream = null;   
	stepsStream = metrics.init(req, res);                                                                      
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/ping:
 *   post:
 *     tags: ["Miscellaneous"]
 */
app.post(apiFirstStep + 'ping', function(req, res) { 
	var stepsStream = null;   
	setCorsHeaders(res);
	stepsStream = pingURL.init(req, res);                                                                      
	startFunc(req, res, stepsStream);
});

/**
 * @swagger
 * /ngp/version:
 *   get:
 *     tags: ["Miscellaneous"]
 */
app.get(apiFirstStep + 'version', function(req, res){
		var status = 200;
		var effResponse = null;

		try
		{
			effResponse = fs.readFileSync(versionFile, 'utf8');
		}
		catch(e)
		{
			logMsg('warn', "Failed to read version file'" + versionFile + "' because of error=" + e.toString());
			effResponse = "500 Failed to read version file";
		}	

		setCorsHeaders(res);
		res.status(status).json({statusCode: status, response: effResponse});              
		res.end();   
});

// Manual Swagger 1.2 documentation

app.get(apiFirstStep + 'docs/v1/vadp', function(req, res) {  
	res.status(200)
		.json(rootSwaggerJson('/default/vadp', 'API_DOCUMENTATION_OPERATIONS_VADP_TITLE', 'API_DOCUMENTATION_OPERATIONS_VADP_TITLE'))
		.end();
});

function rootSwaggerJson(path, title, description) {
	return {  
	   "apiVersion": "1.0",
	   "apis": [  
	      {  
	         "description": description,
	         "path": path,
	         "position": 0
	      }
	   ],
	   "authorizations": {},
	   "info": {
	      "title": title
	   },
	   "swaggerVersion": "1.2"
	};
}

function translatable(key, params = []) {

	if (params !== undefined && params !== null && params.length > 0) {
		var paramsAsStr = params.map(p => `'${p}'`).join(',');
		return `<$key:${key}, params: [${paramsAsStr}]$>`;
	}
	else {
		return `<$key:${key}$>`;
	}
}

function translatableOp(op, key, params = []) {
	return translatable(`API_DOCUMENTATION_OPERATIONS_${op.toUpperCase()}_${key}`, params);
}

app.get(apiFirstStep + 'docs/v1/vadp/default/vadp', function(req, res) {  

	var json = swaggerJsonWithApisAndModels(
		[
		    {
		      	"path": "/ngp/vadp",
		      	"operations": [
			        {  
			            "method": "POST",
			            "summary": translatableOp('VADP', 'CREATE_SUMMARY'),
			            "description": translatableOp('VADP', 'CREATE_DESCRIPTION'),
			            "requestBody": {  
			               "description": translatableOp('VADP', 'CREATE_REQUEST_BODY_DESCRIPTION'),
			               "required":true
			            },
			            "nickname": "registerVadp",
			            "consumes":[  
			               "application/json"
			            ],
			            "parameters":[  
			               {  
			                  "paramType":"query",
			                  "name":"action",
			                  "required":true,
			                  "type":"string",
			                  "enum":[  
			                     "register",
			                     "installandregister"
			                  ],
			                  "description": `${translatableOp('VADP', 'CREATE_PARAMS_ACTION_DESCRIPTION')}:<br /><br><ul><li><code>register</code> - ${translatableOp('VADP', 'CREATE_PARAMS_ACTION_DESCRIPTION_REGISTER')}</li><li><code>installandregister</code> - ${translatableOp('VADP', 'CREATE_PARAMS_ACTION_DESCRIPTION_INSTALL_AND_REGISTER')}</li></ul>`
			               },
			               {  
			                  "paramType":"body",
			                  "name":"body",
			                  "description": `${translatableOp('VADP', 'CREATE_PARAMS_BODY_DESCRIPTION_WITH_EXAMPLE', ['<br /><br /><pre>&#123;<br />&nbsp;&nbsp;&nbsp;&nbsp;"pushinstall": &#123;<br />&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"hostAddress": "&lt;host_address&gt;"<br />&nbsp;&nbsp;&nbsp;&nbsp;&#125;,<br />&nbsp;&nbsp;&nbsp;&nbsp;"identityId": &#123;<br />&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"username": "&lt;username&gt;",<br />&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"password": "&lt;password&gt;"<br />&nbsp;&nbsp;&nbsp;&nbsp;&#125;,<br />&nbsp;&nbsp;&nbsp;&nbsp;"registration": &#123;<br \>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"site": "&lt;site_id&gt;"<br />&nbsp;&nbsp;&nbsp;&nbsp;&#125;<br />&#125;</pre><br />', '<code>host_address</code>', '<code>username</code>', '<code>password</code>', '<code>site_id</code>', '<br /><br />', '<code>identityId</code>', '<code>string</code>', '<code>username</code>', '<code>password</code>', '<code>"identityId": "1001"</code>'])}\n`,
			                  "required":true,
			                  "type": "object",
			                  "schema":{  
			                     "type":"object",
			                     "properties":{  
			                        "pushinstall":{  
			                           "type":"object",
			                           "properties":{  
			                              "hostAddress":{  
			                                 "type":"string"
			                              }
			                           }
			                        },
			                        "identityId":{  
			                           "type":"object",
			                           "properties":{  
			                              "username":{  
			                                 "type":"string"
			                              },
			                              "password":{  
			                                 "type":"string"
			                              }
			                           }
			                        },
			                        "registration":{  
			                           "type":"object",
			                           "properties":{  
			                              "siteId":{  
			                                 "type":"string"
			                              }
			                           }
			                        }
			                     }
			                  }
			               }
			            ],
			            "responseMessages": [
			            	responseMessage(201, translatableOp('VADP', 'CREATE_RESPONSE_201')),
			            	responseMessage(400, translatableOp('VADP', 'CREATE_RESPONSE_400', ['<code>hostAddress</code>'])),
			          		responseMessage(500, translatableOp('VADP', 'CREATE_RESPONSE_201'))
			            ]
			        }
		      	]
		    },
		    {
			    "path": "/api/vadp",
			    "operations": [
			    	{
			            "method": "GET",
			            "summary": translatableOp('VADP', 'FIND_ALL_SUMMARY'),
			            "nickname": "getVadps",
			            "responseMessages": [
			            	responseMessage(200, translatableOp('VADP', 'FIND_ALL_RESPONSE_200')),
			          		responseMessage(500, translatableOp('VADP', 'FIND_ALL_RESPONSE_500'))
			            ]
			        }
			    ]
			},
			{
			    "path": "/api/vadp/{id}",
			    "operations": [
			    	{
			            "method": "PUT",
			            "summary": translatableOp('VADP', 'UPDATE_BY_ID_SUMMARY'),
			            "nickname": "updateVadp",
			            "parameters":[  
				            {  
				                "name": "id",
				                "paramType": "path",
				                "required": true,
				                "type": "integer",
				                "description": translatableOp('VADP', 'UPDATE_BY_ID_PARAMS_ID_DESCRIPTION')
				            },
				            {  
								"paramType": "body",
								"name": "vadp",
								"description": translatableOp('VADP', 'UPDATE_BY_ID_PARAMS_BODY_DESCRIPTION'),
								"required": true,
								"type": "Vadp"
			              	}
			            ],
			            "responseMessages": [
			            	responseMessage(200, translatableOp('VADP', 'UPDATE_BY_ID_RESPONSE_200', ['<code>id</code>'])),
			            	responseMessage(404, translatableOp('VADP', 'COMMON_BY_ID_RESPONSE_404', ['<code>id</code>'])),
			          		responseMessage(500, translatableOp('VADP', 'UPDATE_BY_ID_RESPONSE_500', ['<code>id</code>']))
			            ]
			        },
			        {
			        	"method": "DELETE",
			        	"summary": translatableOp('VADP', 'UNREGISTER_BY_ID_SUMMARY'),
			        	"nickname": "unregisterVadp",
			        	"parameters":[  
				            {  
				                "name": "id",
				                "paramType": "path",
				                "required": true,
				                "type": "integer",
				                "description": translatableOp('VADP', 'UNREGISTER_BY_ID_PARAMS_ID_DESCRIPTION')
				            }
			            ],
			            "responseMessages": [
			            	responseMessage(204, translatableOp('VADP', 'UNREGISTER_BY_ID_RESPONSE_200', ['<code>id</code>'])),
			            	responseMessage(404, translatableOp('VADP', 'COMMON_BY_ID_RESPONSE_404', ['<code>id</code>'])),
			          		responseMessage(500, translatableOp('VADP', 'UNREGISTER_BY_ID_RESPONSE_500', ['<code>id</code>']))
			            ]
			        },
			        {
			        	"method": "POST",
			        	"summary": translatableOp('VADP', 'PERFORM_ACTION_BY_ID_SUMMARY'),
			        	"nickname": "actionVadp",
			        	"parameters":[  
				            {  
				                "name": "id",
				                "paramType": "path",
				                "required": true,
				                "type": "integer",
				                "description": translatableOp('VADP', 'PERFORM_ACTION_BY_ID_PARAMS_ID_DESCRIPTION')
				            },
				            {  
				                "name": "action",
				                "paramType": "query",
				                "required": true,
				                "type": "string",
				                "enum": [
				                	"pushupdate",
				                	"autopushupdate"
				                ],
				                "description": translatableOp('VADP', 'PERFORM_ACTION_BY_ID_PARAMS_ACTION_DESCRIPTION')
				            },
				            {  
								"paramType": "body",
								"name": "pushUpdatePayload",
								"description": translatableOp('VADP', 'PERFORM_ACTION_BY_ID_PARAMS_BODY_DESCRIPTION', ['<code>action</code>','<code>pushupdate</code>']),
								"required": false,
								"type": "PushUpdatePayload"
			              	}
			            ],
			            "responseMessages": [
			            	responseMessage(204, translatableOp('VADP', 'PERFORM_ACTION_BY_ID_RESPONSE_200', ['<code>id</code>'])),
			            	responseMessage(404, translatableOp('VADP', 'COMMON_BY_ID_RESPONSE_404', ['<code>id</code>'])),
			          		responseMessage(500, translatableOp('VADP', 'PERFORM_ACTION_BY_ID_RESPONSE_500', ['<code>id</code>']))
			            ]
			        }
			    ]
			}
		],
		{
			"Vadp": {
				"id": "Vadp",
				"description": translatableOp('VADP', 'MODELS_VADP_DESCRIPTION'),
				"properties": {
					"identityId": {"required": true, "type": "string"},
					"siteId": {"required": true, "type": "string"},
					"options": {
						"required": true, 
						"type": "array",
						"items": {
							"type": "VadpOption"
						}
					}
				}
			},
			"VadpOption": {
				"id": "VadpOption",
				"description": translatableOp('VADP', 'MODELS_VADPOPTION_DESCRIPTION'),
				"properties": {
					"name": {"required": true, "type": "string"},
					"selectedValue": {"required": true, "type": "any"}
				}
			},
			"PushUpdatePayload": {
				"id": "PushUpdatePayload",
				"description": translatableOp('VADP', 'MODELS_PUSHUPDATEPAYLOAD_DESCRIPTION'),
				"properties": {
					"hostAddress": {"required": true, "type": "string"},
					"identityId": {"required": true, "type": "string"}
				}
			}
		}
    );
	
	res.status(200)
		.json(json)
		.end();
});

function swaggerJsonWithApis(apis) {
	return swaggerJsonWithApisAndModels(apis, {});
}

function swaggerJsonWithApisAndModels(apis, models) {
	return {
		"apiVersion": "1.0.0",
		"swaggerVersion": "1.2",
		"consumes": ["application/json"],
		"produces": ["application/json"],
		"basePath": "https://localhost",
		"resourcePath": "/",
		"apis": apis,
		"models": models
	};
}

function responseMessage(code, message) {
	return {"code": code, "message": message};
}

app.get(apiFirstStep + 'docs/v1/sla', function(req, res) {  
	res.status(200)
		.json(rootSwaggerJson('/default/sla', 'SLA Policy', 'SLA Policy'))
		.end();
});

app.get(apiFirstStep + 'docs/v1/sla/default/sla', function(req, res) {  

	var json = swaggerJsonWithApisAndModels([
		{ 
		    "path": "/ngp/slapolicy",
		    "operations": [
		        {
		            "method": "post",  
		            "nickname": "createSlaPolicy",
		            "summary": translatableOp('SLA', 'CREATE_SUMMARY'),
		            "description": translatableOp('SLA', 'CREATE_DESCRIPTION', ['<code>?action=applyPolicyOptions</code>']),
		            "parameters":[  
		               {  
		                  "paramType":"query",
		                  "name":"action",
		                  "required":false,
		                  "type":"string",
		                  "enum":[  
		                     "applyPolicyOptions"
		                  ],
		                  "description": translatableOp('SLA', 'CREATE_PARAMS_ACTION_DESCRIPTION', ['<br /><br />', '<br /><br /><pre>{\"script\": {\"preGuest\":null,\"postGuest\":null,\"continueScriptsOnError\":false},\"policyOption\":{\"excluderesources\":\"\",\"forcebaseresources\":\"\",\"runInventory\":false}}</pre><br /><br />', '<br /><br /><pre>{\"resources\":[{\"href\":\"https://71.100.14.11/api/application/oracle/instance/37573786e1ec1f871c7b96ca28014ce0/database/3968fd3fd2476b0c4151f71979c5b26f?from=hlo\",\"id\":\"3968fd3fd2476b0c4151f71979c5b26f\",\"metadataPath\":\"/DBINST:37573786e1ec1f871c7b96ca28014ce0/DB:3968fd3fd2476b0c4151f71979c5b26f\"}],\"subtype\":\"oracle\",\"options\":{\"maxParallelStreams\":1,\"dbFilesForParallelStreams\":\"SINGLE_FILE\",\"backupPreferredNode\":\"\",\"logbackup\":{\"purgePrimaryLogs\":false,\"primaryLogRetentionDays\":3,\"performlogbackup\":true,\"rpo\":{\"frequency\":1,\"frequencyType\":\"DAY\",\"triggerTime\":\"12:00:00 AM\",\"metadata\":{\"activateDate\":1540267200000}}}}}</pre><br /><br />', '<br /><br /><pre>{\"subtype\":\"oracle\",\"version\":\"1.0\",\"resources\":[{\"href\":\"https://71.100.14.11/api/application/oracle/instance/37573786e1ec1f871c7b96ca28014ce0/database/3968fd3fd2476b0c4151f71979c5b26f?from=hlo\",\"id\":\"3968fd3fd2476b0c4151f71979c5b26f\",\"metadataPath\":\"/DBINST:37573786e1ec1f871c7b96ca28014ce0/DB:3968fd3fd2476b0c4151f71979c5b26f\"}],\"slapolicies\":[{\"href\":\"https://71.100.14.11/api/spec/storageprofile/2002\",\"id\":\"2002\",\"name\":\"Bronze\"}]}</pre><br /><br />'])
		               },
		               {  
		                  "paramType":"query",
		                  "name":"name",
		                  "required":false,
		                  "type":"string",
		                  "description": translatableOp('SLA', 'CREATE_PARAMS_NAME_DESCRIPTION', ['<code>applyPolicyOptions</code>'])
		               },
		               {  
		                  "paramType":"query",
		                  "name":"subtype",
		                  "type":"string",
		                  "enum":[  
		                     "vmware",
		                     "hyperv",
		                     "storage",
		                     "sql",
		                     "oracle",
		                     "catalog",
		                     "db2",
							 "exch",
							 "office365",
							 "mongo",
               "openshift",
							 "k8s",
							 "file"
		                  ],
		                  "description": translatableOp('SLA', 'CREATE_PARAMS_SUBTYPE_DESCRIPTION', ['<code>applyPolicyOptions</code>'])
		               },
		               {  
		                  "paramType":"body",
		                  "name": "body",
		                  "required":false,
		                  "type":"Policy",
		                  "description": translatableOp('SLA', 'CREATE_PARAMS_BODY_DESCRIPTION', ['<code>applyPolicyOptions</code>'])
		               }
		            ],
		            "responseMessages": [ 
		                responseMessage(200, translatableOp('SLA', 'CREATE_RESPONSE_200')),
		                responseMessage(201, translatableOp('SLA', 'CREATE_RESPONSE_201'))
		            ]
		         },
		         {
		            "method": "get",  
		            "nickname": "getSlaPolicies",
		            "summary": translatableOp('SLA', 'FIND_ALL_SUMMARY'),
		            "parameters":[  
		               {  
		                  "paramType":"query",
		                  "name":"pageSize",
		                  "type":"integer",
		                  "description": translatableOp('SLA', 'FIND_ALL_PARAMS_PAGE_SIZE_DESCRIPTION')
		               },
		               {  
		                  "paramType":"query",
		                  "name":"sort",
		                  "type": "string",
		                  "enum": ["ACS", "DESC"],
		                  "description": ""
		               },
		               {  
		                  "paramType":"query",
		                  "name":"subtype",
		                  "type":"string",
		                  "enum":[  
		                     "vmware",
		                     "hyperv",
		                     "storage",
		                     "sql",
		                     "oracle",
		                     "saphana",
		                     "iscache",
		                     "catalog",
		                     "db2",
							 "exch",
							 'office365',
							 "mongo",
              "openshift",
							 "k8s",
							 "file"
		                  ],
		                  "description": translatableOp('SLA', 'FIND_ALL_PARAMS_SUBTYPE_DESCRIPTION')
		               },
		               {  
		                  "paramType":"query",
		                  "name":"status",
		                  "type":"string",
		                  "enum":[  
		                     true,
		                     false
		                  ],
		                  "description": translatableOp('SLA', 'FIND_ALL_PARAMS_STATUS_DESCRIPTION')
		               },
		               {  
		                  "paramType":"query",
		                  "name":"embedoptions",
		                  "type":"string",
		                  "enum":[  
		                     true,
		                     false
		                  ],
		                  "description": translatableOp('SLA', 'FIND_ALL_PARAMS_EMBED_OPTIONS_DESCRIPTION')
		               },
		               {  
		                  "paramType":"query",
		                  "name":"name",
		                  "type":"string",
		                  "description": translatableOp('SLA', 'FIND_ALL_PARAMS_NAME_DESCRIPTION')
		               }
		            ],
		            "responseMessages": [
		                responseMessage(200, translatableOp('SLA', 'FIND_ALL_RESPONSE_200')),
		                responseMessage(500, translatableOp('SLA', 'FIND_ALL_RESPONSE_500'))
		            ]
		         },
		        {
		            "method": "delete",  
		            "nickname": "deleteSlaPolicy",
		            "summary": translatableOp('SLA', 'DELETE_BY_NAME_SUMMARY'),
		            "parameters":[  
		               {  
		                  "name":"name",
		                  "paramType":"query",
		                  "type": "string",
		                  "required":true,
		                  "description": translatableOp('SLA', 'DELETE_BY_NAME_PARAMS_NAME_DESCRIPTION')
		               }
		            ],
		            "responseMessages": [
		                responseMessage(204, translatableOp('SLA', 'DELETE_BY_NAME_RESPONSE_204')),
		                responseMessage(404, translatableOp('SLA', 'DELETE_BY_NAME_RESPONSE_404'))
		            ]
		        }
		    ]
		},
		{  
		    "path": "/ngp/slapolicy/{id}",
		    "operations": [
		        {
		            "method": "put",  
		            "nickname": "updateSlaPolicy",
		            "summary": translatableOp('SLA', 'UPDATE_BY_ID_SUMMARY'),
		            "parameters":[  
			            {  
			                "name":"id",
			                "paramType":"path",
			                "required":true,
			                "type": "integer",
			                "description": translatableOp('SLA', 'UPDATE_BY_ID_PARAMS_ID_DESCRIPTION')
			            }
		            ],
		            "responseMessages": [
		                responseMessage(200, translatableOp('SLA', 'UPDATE_BY_ID_RESPONSE_200')),
		                responseMessage(404, translatableOp('SLA', 'UPDATE_BY_ID_RESPONSE_404'))
		            ]
		        }
		    ]
		},
		{  
		    "path": "/ngp/slapolicy/jobsessions",
		    "operations": [
		        {
		            "method": "get",  
		            "nickname": "getSlaPolicyJobSessions",
		            "summary": translatableOp('SLA', 'FIND_JOB_SESSIONS_SUMMARY'),
		            "description": translatableOp('SLA', 'FIND_JOB_SESSIONS_DESCRIPTION'),
		            "parameters": [  
			            {  
			                "paramType":"query",
			                "name":"subtype",
			                "type":"string",
			                "enum":[  
			                    "vmware",
			                    "hyperv",
			                    "storage",
			                    "sql",
			                    "oracle",
			                    "saphana",
			                    "iscache",
			                    "catalog",
			                    "db2",
								"exch",
								'office365',
								"mongo",
                "openshift",
								"k8s",
								"file"
			                ],
			                "description": translatableOp('SLA', 'FIND_JOB_SESSIONS_PARAMS_SUBTYPE_DESCRIPTION')
			            },
			            {  
			                "paramType":"query",
			                "name":"name",
			                "type":"string",
			                "description": translatableOp('SLA', 'FIND_JOB_SESSIONS_PARAMS_NAME_DESCRIPTION')
			            }
		            ],
		            "responseMessages": [
		                responseMessage(200, translatableOp('SLA', 'FIND_JOB_SESSIONS_RESPONSE_200'))
		            ]
		        }
		    ]
		}
	],
	{
		"Policy": {
			"id": "Policy",
			"description": translatableOp('SLA', 'MODELS_POLICY_DESCRIPTION'),
			"properties": {
				"name": {"required": true, "type": "string"},
				"version": {"required": true, "type": "string"},
				"spec": {"required": true, "type": "PolicySpec"}
			}
		},
		"PolicySpec": {
			"id": "PolicySpec",
			"description": translatableOp('SLA', 'MODELS_POLICYSPEC_DESCRIPTION'),
			"properties": {
				"simple": {"required": true, "type": "boolean"},
				"subpolicy": {
					"required": true, 
					"type": "array", 
					"items": {
						"type": "SubPolicy"
					}
				}
			}
		},
		"SubPolicy": {
			"id": "SubPolicy",
			"description": translatableOp('SLA', 'MODELS_SUBPOLICY_DESCRIPTION'),
			"properties": {
				"type": {"required": true, "type": "string"},
				"software": {"required": true, "type": "boolean"},
				"retention": {
					"required": true, 
					"type": "object",
					"properties": {
						"age": {"required": true, "type": "integer"}
					}
				},
				"useEncryption": {"required": true, "type": "boolean"},
				"site": {"required": true, "type": "string"},
				"trigger": {"required": true, "type": "Trigger"}
			}
		},
		"Trigger": {
			"id": "Trigger",
			"description": translatableOp('SLA', 'MODELS_TRIGGER_DESCRIPTION'),
			"properties": {
				"frequency": {"required": true, "type": "integer"},
				"type": {"required": true, "type": "string"},
				"activateDate": {"required": true, "type": "string"}
			}
		}
    }
	)
	
	res.status(200)
		.json(json)
		.end();
});

app.get(apiFirstStep + 'docs/v1/storage', function(req, res) {  
	res.status(200)
		.json(rootSwaggerJson('/default/storage', 'API_DOCUMENTATION_OPERATIONS_STORAGE_TITLE', 'API_DOCUMENTATION_OPERATIONS_STORAGE_TITLE'))
		.end();
});

app.get(apiFirstStep + 'docs/v1/storage/default/storage', function(req, res) {  

	var json = swaggerJsonWithApisAndModels(
		[
		    {
		      	"path": "/ngp/storage",
		      	"operations": [
			        {  
			            "method": "POST",
			            "summary": translatableOp('STORAGE', 'REGISTER_SUMMARY'),
			            "description": translatableOp('STORAGE', 'REGISTER_DESCRIPTION'),
			            "nickname": "registerVadp",
			            "consumes":[  
			               "application/json"
			            ],
			            "parameters":[  
			                {  
			                    "paramType": "query",
			                    "name": "action",
			                    "required": false,
			                    "type": "string",
			                    "enum":[
			                        "applySLAPolicies"
			                    ],
			                    "description": translatableOp('STORAGE', 'REGISTER_PARAMS_ACTION_DESCRIPTION')
			                },
			                {  
			                    "paramType": "body",
			                    "name": "storage",
			                    "description": translatableOp('STORAGE', 'REGISTER_PARAMS_BODY_DESCRIPTION', ['<code>applySLAPolicies</code>', '<br /><br /><pre>{\"hostAddress\": \"100.0.0.1\", \"siteId\": \"1000\", \"type\": \"vsnap\", \"username\": \"foo\", \"password\": \"bar\", \"portNumber\": 8900, \"sslConnection\": true}</pre><br /><br />', 'HREF', '<br /><br /><pre>{\"hostAddress\": \"172.20.2.136\", \"siteId\": \"1000\", \"type\": \"vsnap\", \"username\": \"https://localhost/api/identity/user/2000\", \"portNumber\":8900, \"sslConnection\": true}</pre>']),
			                    "required": false,
			                    "type": "Storage",
			                    "allowMultiple": false,
			                    "defaultValue": ""
			                }
			            ],
			            "responseMessages": [
			            	responseMessage(201, translatableOp('STORAGE', 'REGISTER_RESPONSE_201')),
			            	responseMessage(400, translatableOp('STORAGE', 'REGISTER_RESPONSE_400')),
			          		responseMessage(500, translatableOp('STORAGE', 'REGISTER_RESPONSE_500'))
			            ]
			        }
		      	]
		    },
		    {
			    "path": "/api/storage",
			    "operations": [
			    	{
			            "method": "GET",
			            "summary": translatableOp('STORAGE', 'FIND_ALL_SUMMARY'),
			            "nickname": "getStorages",
			            "responseMessages": [
			            	responseMessage(200, translatableOp('STORAGE', 'FIND_ALL_RESPONSE_200')),
			          		responseMessage(500, translatableOp('STORAGE', 'FIND_ALL_RESPONSE_500'))
			            ]
			        }
			    ]
			},
			{
			    "path": "/ngp/storage/{id}",
			    "operations": [
			    	{
			            "method": "PUT",
			            "summary": translatableOp('STORAGE', 'UPDATE_BY_ID_SUMMARY'),
			            "nickname": "updateStorage",
			            "parameters": [
			            	{  
								"paramType": "path",
								"required": true,
								"type": "integer",
								"description": translatableOp('STORAGE', 'UPDATE_BY_ID_PARAMS_ID_DESCRIPTION')
			                },
			                {  
								"paramType": "body",
								"name": "storage",
								"description": translatableOp('STORAGE', 'UPDATE_BY_ID_PARAMS_BODY_DESCRIPTION', ['<br /><br /><pre>{\"hostAddress\": \"100.0.0.1\", \"siteId\": \"1000\", \"type\": \"vsnap\", \"username\": \"foo\", \"password\": \"bar\", \"portNumber\": 8900, \"sslConnection\": true}</pre><br /><br />', 'HREF', '<br /><br /><pre>{\"hostAddress\": \"172.20.2.136\", \"siteId\": \"1000\", \"type\": \"vsnap\", \"username\": \"https://localhost/api/identity/user/2000\", \"portNumber\":8900, \"sslConnection\": true}</pre>']),
								"required": true,
								"type": "Storage"
			                }
			            ],
			            "responseMessages": [
			            	responseMessage(200, translatableOp('STORAGE', 'UPDATE_BY_ID_RESPONSE_200')),
			            	responseMessage(404, translatableOp('STORAGE', 'COMMON_BY_ID_RESPONSE_404', ['<code>id</code>'])),
			          		responseMessage(500, translatableOp('STORAGE', 'UPDATE_BY_ID_RESPONSE_500'))
			            ]
			        }
			    ]
			},
			{
				"path": "/api/storage/{id}",
				"operations": [
			        {
			            "method": "DELETE",
			            "summary": translatableOp('STORAGE', 'UNREGISTER_BY_ID_SUMMARY'),
			            "nickname": "unregisterStorage",
			            "parameters": [
			            	{  
								"paramType": "path",
								"required": true,
								"type": "integer",
								"description": translatableOp('STORAGE', 'UNREGISTER_BY_ID_PARAMS_ID_DESCRIPTION')
			                }
			            ],
			            "responseMessages": [
			            	responseMessage(200, translatableOp('STORAGE', 'UNREGISTER_BY_ID_RESPONSE_200')),
			            	responseMessage(404, translatableOp('STORAGE', 'COMMON_BY_ID_RESPONSE_404', ['<code>id</code>'])),
			          		responseMessage(500, translatableOp('STORAGE', 'UNREGISTER_BY_ID_RESPONSE_500'))
			            ]
			        }
			    ]
			}
		],
		{
			"Storage": {
				"id": "Storage",
				"description": translatableOp('STORAGE', 'MODELS_STORAGE_DESCRIPTION'),
				"properties": {
					"hostname": {"required": true, "type": "string"},
					"siteId": {"required": true, "type": "string"},
					"type": {"required": true, "type": "string"},
					"portNumber": {"required": true, "type": "integer"},
					"sslConnection": {"required": true, "type": "boolean"},
					"username": {"required": true, "type": "string"},
					"password": {"required": false, "type": "string"}
				}
			}
	    }
	)
	
	res.status(200)
		.json(json)
		.end();
});

bindSwaggerRoot('analytics', 'API_DOCUMENTATION_OPERATIONS_ANALYTICS_TITLE');

function bindSwaggerRoot(key, name) {
	app.get(`${apiFirstStep}docs/v1/${key}`, function(req, res) {  
		res.status(200)
			.json(rootSwaggerJson(`/default/${key}`, name, name))
			.end();
	});
}

bindSwaggerDocumentation('analytics', function(req, res) {  

	var json = swaggerJsonWithApis(
		[
		    {
		      	"path": "/ngp/analytics/dbbackuphistory",
		      	"operations": [
			        {  
			            "method": "GET",
			            "summary": translatableOp('ANALYTICS', 'FIND_DB_BACKUP_HISTORY_SUMMARY'),
			            "description": translatableOp('ANALYTICS', 'FIND_DB_BACKUP_HISTORY_DESCRIPTION'),
			            "nickname": "getDbBackupHistory",
			            "responseMessages": [
			            	responseMessage(200, translatableOp('ANALYTICS', 'FIND_DB_BACKUP_HISTORY_RESPONSE_200')),
			            	responseMessage(500, translatableOp('ANALYTICS', 'FIND_DB_BACKUP_HISTORY_RESPONSE_500'))
			            ]
			        }
		      	]
		    },
		    {
			    "path": "/ngp/analytics/vmbackuphistory",
			    "operations": [
			    	{  
			            "method": "GET",
			            "summary": translatableOp('ANALYTICS', 'FIND_VM_BACKUP_HISTORY_SUMMARY'),
			            "description": translatableOp('ANALYTICS', 'FIND_VM_BACKUP_HISTORY_DESCRIPTION'),
			            "nickname": "getVmBackupHistory",
			            "responseMessages": [
			            	responseMessage(200, translatableOp('ANALYTICS', 'FIND_VM_BACKUP_HISTORY_RESPONSE_200')),
			            	responseMessage(500, translatableOp('ANALYTICS', 'FIND_VM_BACKUP_HISTORY_RESPONSE_500'))
			            ]
			        }
			    ]
			},
			{
			    "path": "/ngp/analytics/vmbackupslacom",
			    "operations": [
			    	{  
			            "method": "GET",
			            "summary": translatableOp('ANALYTICS', 'FIND_VM_BACKUP_SLA_COM_SUMMARY'),
			            "description": translatableOp('ANALYTICS', 'FIND_VM_BACKUP_SLA_COM_DESCRIPTION'),
			            "nickname": "getVmBackupSlaCom",
			            "parameters": [
			            	{  
								"paramType": "query",
								"name": "download",
								"required": false,
								"type": "boolean",
								"enum": ["true", "false"],
								"description": translatableOp('ANALYTICS', 'FIND_VM_BACKUP_SLA_COM_PARAMS_DOWNLOAD_DESCRIPTION')
			                }
			            ],
			            "responseMessages": [
			            	responseMessage(200, translatableOp('ANALYTICS', 'FIND_VM_BACKUP_SLA_COM_RESPONSE_200')),
			            	responseMessage(500, translatableOp('ANALYTICS', 'FIND_VM_BACKUP_SLA_COM_RESPONSE_500'))
			            ]
			        }
			    ]
			},
			{
			    "path": "/ngp/analytics/dbbackupslacom",
			    "operations": [
			    	{  
			            "method": "GET",
			            "summary": translatableOp('ANALYTICS', 'FIND_DB_BACKUP_SLA_COM_SUMMARY'),
			            "description": translatableOp('ANALYTICS', 'FIND_DB_BACKUP_SLA_COM_DESCRIPTION'),
			            "nickname": "getDbBackupSlaCom",
			            "responseMessages": [
			            	responseMessage(200, translatableOp('ANALYTICS', 'FIND_DB_BACKUP_SLA_COM_RESPONSE_200')),
			            	responseMessage(500, translatableOp('ANALYTICS', 'FIND_DB_BACKUP_SLA_COM_RESPONSE_500'))
			            ]
			        }
			    ]
			}
		]
	)
	
	res.status(200)
		.json(json)
		.end();
});

function bindSwaggerDocumentation(key, handler) {
	app.get(`${apiFirstStep}docs/v1/${key}/default/${key}`, handler);
}

// ***** End of ALL ENDPOINTS *****


/*
  NB: validateQs ignores invalid *keys* in the query string.
  qs0.rc value 0-> no query key present, 1-> 1 or more present & valid, 2->valid key w/invalid value
*/
function validateQs(req, licitQs)                                                                           // obj with licit query string keys whose value is array of licit values for those keys                                       
{
	var qsO = {
		rc: 0,                                                                                              // 0-> no query key present, 1-> 1 or more present & valid, 2->valid key w/invalid value
		numKv: 0,                                                                                           // number objects in keyVals
		keyVals: {                                                                                          // hash of valid K=V pairs

		},
	};

	var illicitKeys = [];
	
    for ( queryKey in req.query )                                                                           // what's in the URL, e.g., jive=true&action=applySLAPolicies
	{
		if ( ! licitQs[queryKey] )                                                                          // FTI invalid key
		{
			logMsg('warn', 'Query string key=' + queryKey + ' is invalid and will be ignored');
			illicitKeys.push(queryKey);
		}
	}

	illicitKeys.forEach(function(illicitKey){
		delete req.query[illicitKey];                                                                       // ignore it for now  
		logMsg('debug', 'Removed from query string illicit key=' + illicitKey);
	});

	for ( queryKey in req.query )                                                                           // key perforce licit
	{
		var licitVals = licitQs[queryKey];                                                                  // array of licit values for the key

		var okKeyVal = false;

		licitVals.forEach(function(licitVal){
			var match = null;

			if ( typeof licitVal == 'object' )                                                              //!! FTI a Rx
			{
				match = req.query[queryKey].match(licitVal);
			}
			else
			{
				match = ( licitVal == req.query[queryKey] );
			}
			
			if ( match )                                                                                      // e.g., key 'action' has value 'applySLAPolicies'			
			{
				qsO.keyVals[queryKey] = req.query[queryKey];
				okKeyVal = true;
				return;                                                                                         // a bit wastefully continue with next licit value
			}
		});

		if ( ! okKeyVal )                                                                                   // e.g., action=applySpackle
		{
			qsO.rc = 2;
			return qsO;                                                                                       // valid key, invalid value; short-circuit, we're done
		}
	}

	qsO.numKv = Object.keys(qsO.keyVals).length;

	if ( qsO.numKv > 0 )                                                                                  // FTI at least 1 licit K=V present in query string 
		qsO.rc = 1;

	logMsg('debug', 'Query object=' + JSON.stringify(qsO));
	return qsO;
}


function startFunc(req, res, stepsStream)                                    
{
	var ngState = getSessionState(req, res);

	if ( ngState.stale )                                                                                  // FTI we got bounced while client remained up. AND response already sent.
		return;                                                                                             // !!! R U CERTAIN THAT RESPONSE SENT IN ALL CASES HERE?

	var ecxSid = ngState.ecxSid;
	res.ecxSid = ecxSid;

// NB: the cb function in the initial call to runSerialFuncs will NOT be driven until all the steps in the serial function stream have completed (one way or another).

	runSerialFuncs(stepsStream, ecxSid, escapeChars(req.body), function(){                                              
		var temp = ngState.result;                                                                          // assume we've object (usable with .json() call)
		
		if ( typeof(temp) == 'string' && temp.length > 0 )                                                  // it can happen that this string is zero length, parse not
			temp = JSON.parse(temp);

		//if ( req.method == 'GET' && req.url.contains('slapolicy') )                                         //!! check for status & subtype, too??
		//	logMsg('warn', 'CACHE GETSLAP=' + JSON.stringify(temp));

		logMsg('debug', 'GOOD PLACE FOR A "SENDING RESPONSE TO CLIENT" MSG=');

		setCorsHeaders(res);
		res.status(ngState.result.statusCode).json(temp);                                                   // NB: 'internal' (non-ECX call) failure must also set these values		
		res.end();                                                                                          // SEND FINAL response to client (but will be intercepted by modresp waypoint)

		logMsg('info', "RESPONSE returned to client for function '" + getCurrentFunc(req.reqId, ecxSid) + "' in session ID='" + ecxSid + "'");
		
		delete ngState.endPoints[req.reqId];                                                                // KA-BOOM
	});
}


// For static content, e.g., http://71.100.14.30:9444/index.html. But file actually resides in /home/ecxcc/ecxmon/public

app.use(express.static(PUBLICDIR));

app.use(logErrors);


function logErrors (err, req, res, next) {                                                              // EHI error in Express stack
	appendErrorToLog(err);
  next(err);
}


function appendErrorToLog(err)
{
		fs.appendFileSync('/data/log/node-cdm-service/' + LOGFN, err.stack);
}


/*
  runSerialFuncs is a recursive function that drives each of the steps defined by the 'name' property in a serial function stream array. As each step finishes, it will re-enter runSerialFuncs.
  The loop terminator is an empty function stream array. Note also that the cb argument is the 'final' callback which, when called, will send HTTP response back to client.
*/

function runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb)                                     // the 'cb' arg here came from the 1st call to here                   
{
	if ( funcStream.length < 1 )
	{
		//logMsg('warn', 'HOWNOW=' + JSON.stringify(JSON.parse(ecxSidHash[ecxSessionId].result.response)));

		cb();                                                                                      // callback driven only after all constituent calls finished (hence the clever name)
		return;
	}

	logMsg('debug', 'Step stream=' + JSON.stringify(funcStream));                                  // NB: this will show at top function about to be driven, followed by subsequent steps

	var step = funcStream.shift();                                                                 // func is step we're about to drive

	if ( step.continue == 'fail' )                                                                 // FTI step we're about to drive told by previous step to give up the ghost (can't happen with 1st step)
	{
		//logMsg('error', 'FAILURE=' + JSON.stringify(ecxSidHash[ecxSessionId].result));             // make damn sure this is always set
		cb();
		return;
	}

	var currentFunc = getCurrentFunc(step.reqId, ecxSessionId);

	var ngState = ecxSidHash[ecxSessionId];        

	if ( step.continue == 'skip' )                                                                 // FTI step we're about to drive told by previous step to skip (1st step can't be skipped)
	{
		logMsg('info', "Skipping step '" + currentFunc + '(' + step.name + ")' in session ID '" + ecxSessionId + "'");
	
		if ( funcStream.length > 0 )                                                               // FTI another step after one we're now skipping
		{
			//logMsg('warn', 'RECURSING FROM SKIPPED STEP');
			runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
			return;
		}
		else                                                                                       // EHI no more steps, we're done
		{
			cb();
			return;
		}
	}
	
	var hostIp = step.ip;

	logMsg('info', "Started step '" + currentFunc + '(' + step.name + ")' in session ID '" + ecxSessionId + "'");

	step.func(step, funcStream, ecxSessionId, oFromNg, cb);	
}

 
// Validate user mode (use existing | create)

var userMode = function(step, funcStream, ecxSessionId, oFromNg, cb) {
	var userMode = validateNgUserInput(ecxSessionId, oFromNg);                                   // either use extant user (by its identity name) or create via username & password - both is wrong
	var userLit = 'user';
	var umModeName = 'userMode';
	var umHrefName = 'userHref';

	if ( oFromNg.ve )		
	{
		userLit = 'VE user';
		umModeName = 'veUserMode';
		umHrefName = 'veUserHref';
	}

	if ( ! userMode.valid )                                                                      // FTI fatal error
	{
		logMsg('error', 'Detected invalid ' + userLit + ' input from NG for function=' + step.name);

/*
  Failures can be internal to this module or returned by ECX. In all cases we return an object that consists of an HTTP status code and an ECX response. If the failure is internal, we obviously
  manufacture the values for these properties. 'Return' here means that we set property 'result' in the session state (ngState).
*/

		var failResult = {                                                                       // try to choose code that reflects actual cause of error
			statusCode: 400,
			response: {rc: 'Input to provider registration contains invalid data'},                     
		};

		setSerialStreamResult(ecxSessionId, failResult);                                         // do NOT use setStepFailResult here because it wants to continue driving runSerialSteps  
		step.cb(ecxSessionId, null);                                                             // drive cb after step (largely pro forma)     
		cb();                                                                                    // drive FINAL (before sending response to client) cb in order to send response to client
		return;
	}

	//@#@setEndpointState(ecxSessionId, umModeName, userMode);                                       // remember if mere 'name' XOR 'username&&password'
	setRequestState(step.reqId, ecxSessionId, umModeName, userMode);                                       // remember if mere 'name' XOR 'username&&password'

	if ( ! userMode.createUser )                                                               // FTI gui gave us href of an extant user
	{
		//@#@setEndpointState(ecxSessionId, umHrefName, require('querystring').unescape(userMode.username));

		setRequestState(step.reqId, ecxSessionId, umHrefName, require('querystring').unescape(userMode.username));
		logMsg('debug', 'Extant user stored under state key=' + umHrefName + '=' + userMode.username);

		if ( funcStream.length > 0 )                                                             // FTI there is a next step (always first element in what remains of array)
			funcStream[0].continue = 'skip';                                                       // skip user creation step
	}

	delete oFromNg.ve;
	delete oFromNg.ao;

	step.cb(ecxSessionId, null);                                                               //!! IN EACH CASE (extant user | create ) just drive next function
	runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 		
};


// setSerialStreamResult is called by any failure case (internal or external) and by successful *terminal* steps, i.e., last step in a serial function stream.

function setSerialStreamResult(ecxSid, result)
{
	var ngState = ecxSidHash[ecxSid];                                                          // assume we target persisted value
	ngState.result = result;                                                                   // NB: property 'result' is magical, i.e., conventional    
}


function getSerialStreamResult(ecxSid)
{
	var ngState = ecxSidHash[ecxSid];                                                              
	return ngState.result;
}


// NB: despite the 'reg' in its name, this function is called by updateprovider.js & regsmtpldap.js. BE CAREFUL

var regDeleteUser = function(step, funcStream, ecxSessionId, oFromNg, cb)                          // EHI provider registration failed, IF we created a user, delete it
{
	//@#@var createdIdentities = getEndpointState(ecxSessionId, 'createdIdentities');                   // e.g., https://71.100.14.58:8443/api/identity/user/1311
	var createdIdentities = getRequestState(step.reqId, ecxSessionId, 'createdIdentities');                   // e.g., https://71.100.14.58:8443/api/identity/user/1311

	if ( createdIdentities && createdIdentities.length > 0 )                                       // recursively delete contents 
	{
		logMsg('debug', 'Created identities to delete=' + JSON.stringify(createdIdentities));
		
		deleteOneUser(createdIdentities, ecxSessionId, function(){
			//logMsg('warn', 'BACK FROM IDENTITY DELETION');
			
			step.cb(ecxSessionId, null);  //!! BE CAREFUL WITH THIS, should be OK. we are probably driving the pre-pended delete user step. next step probably marked for failure
			runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 			
		});
	}
	else 
	{
		step.cb(ecxSessionId, null); 
		runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 
	}
};


function deleteOneUser(identitiesToDelete, ecxSessionId, cb)
{
	if ( identitiesToDelete.length < 1 )
		cb();
	else
	{
		var identityToDelete = identitiesToDelete.shift();
		
		const urlParse = require('url-parse');
		var path = urlParse(identityToDelete).pathname;

		sendRequestToEcx('DELETE', ecxExternal, path, ecxSessionId, null, function(dlObj){           
			if ( dlObj.response.statusCode != 204 )
				logMsg('error', "Deletion of SPP identity '" + identityToDelete + "' failed because of error=" + JSON.stringify(dlObj.rO));                         
			
			deleteOneUser(identitiesToDelete, ecxSessionId, cb);
		});
	}
}


function getOrCreateEcxUser(step, funcStream, ecxSessionId, oFromNg, cb)
{
	var ngState = ecxSidHash[ecxSessionId];

	var matchedUser = null;

	//@#@if ( getEndpointState(ecxSessionId, 'userMode').createUser )                                   // FTI generate a 'name' for the identity & create it, so simply drive the user creation step
	if ( getRequestState(step.reqId, ecxSessionId, 'userMode').createUser )                                   // FTI generate a 'name' for the identity & create it, so simply drive the user creation step
	{
		//logMsg('warn', 'DRIVING TO CREATE');
	}
	else                                                                                           // EHI use an extant identity; find it via given 'name'
	{
		//logMsg('warn', 'EXTANT USER');

		var effNameProp = oFromNg.username;

		if ( ! ngState.userHashByName[effNameProp] )                                              // FTI identity name we were told to use doesn't exist
		{
			var ngState = ecxSidHash[ecxSessionId];

			if ( funcStream.length > 0 )
				funcStream[0].continue = 'fail';                                                   // Abort the stream; MUST SET stream result if 'internal' (non-ECX) failure
						
			var failResult = {                                                                     // try to choose code that reflects actual cause of error
				statusCode: 400,
				response: {rc: "User identity with name '" + effNameProp + "' does not exist"},
			};

			setSerialStreamResult(ecxSessionId, failResult);                 
		}
		else                                                                                       // EHI matched user: capture its user.href and FORCE SKIP of user creation step
		{
			matchedUser = ngState.userHashByName[effNameProp];

			//@#@setEndpointState(ecxSessionId, 'userHref', matchedUser.links.self.href);
			setRequestState(step.reqId, ecxSessionId, 'userHref', matchedUser.links.self.href);
			if ( funcStream.length > 0 )                                                           // FTI there is a next step (always first element in what remains of array)
				funcStream[0].continue = 'skip';                                                   // cause skip of next serial step (creation of user)
		}
	}

// From ECX callback now drive the recursive driver to move to next constituent function

	step.cb(ecxSessionId, matchedUser);                                                          
	runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 
}


function getEcxIdentities(ecxSessionId, step, cb)
{
	sendRequestToEcx('GET', ecxExternal, step.path, ecxSessionId, null, function(rObj){            // know nothing about the dupl ident; no choice but to get all identities
		if ( rObj.e || rObj.response.statusCode != 200 )
			var returnObj = getGenericEcxErrorResponse(rObj, ecxSessionId, step);
		else
		{
			returnObj = {
				statusCode: rObj.response.statusCode,                                              // leave response null
			};

			var ecxUsers = rObj.rO;
			
			var ngState = ecxSidHash[ecxSessionId];
			ngState.userHashByName = arrayToHash(ecxUsers.users, 'name');
		}

		cb(returnObj);
	});			
}


function getEcxSites(ecxSid, cb)
{
	var rc = false;

	sendRequestToEcx('GET', ecxExternal, '/api/site', ecxSid, null, function(rObj){
		var step = {
			verb: 'GET',
			path: '/api/site',                                                                       // keep happy common routine
		};

		if ( rObj.e || rObj.response.statusCode != 200 )  
		{
			var returnObj = getGenericEcxErrorResponse(rObj, ecxSid, step);
		}  
		else
		{
			var ecxSites = rObj.rO;

			var ngState = ecxSidHash[ecxSid];
			ngState.siteHashByName = arrayToHash(ecxSites.sites, 'name');
			rc = true;
		}
		
		//logMsg('error', 'SITES=' + JSON.stringify(ngState.siteHashByName));
		cb(rc);
	});			
}


function sendRequestToEcx(method, port, path, ecxSid, postBody, cb)                             // GENERIC requester to ECX (**ALL** calls to ECX come here)
{
	var headers = {                                                                                
			'x-endeavour-sessionid': ecxSid,
			'x-endeavour-locale': getEcxLocaleHeader.get('ecxLocale'),
			'Content-Type' : 'application/json',
	};   
	
	    var options = { 
			host: ecxSidHash[ecxSid].host,
			port: port,
			path: path,
			method: method,
			headers: headers,
			rejectUnauthorized: false                                                               // w/o true, bad creds won't be handled by the on error handler                                                  
		};
		
		if(process.env.CORE_IP)
		options.host = process.env.CORE_IP;

	var guiReqProtocol = ecxSidHash[ecxSid].reqProtocol;                                        // set by waypoint up top                  

	if ( guiReqProtocol == 'https' )
			effProtocol = https;
	else
			effProtocol = http;

	if ( ! useHttps )
			options.port = ecxInternal;

	var rObj = {
			rO: null,
			response: null,
	};

// *** START CHANGE

	if ( productionMode )
	{
			options.port = null;                                                                    // https protocol will assume 443
			effProtocol = https;
			guiReqProtocol = 'https';                                                               // diagnostic only from here on
	}

// *** END CHANGE

	var request = effProtocol.request(options, function(res) {
			var response = '';

			res.on('data', function(d) {
				response += d;
			});

			res.on('error', function(e) {
					logMsg('error', "API call '" + method + ' ' + path + "' failed because of error='" + JSON.stringify(e));
					rObj.e = JSON.stringify(e);                                                        
					cb(rObj);                                                               
			});

			res.on('end', function(){                                                               
					if ( res.statusCode != 200 && res.statusCode != 201 && res.statusCode != 204 )      // e.g., GET&PUT, POST, and DELETE ( make damn sure this doesn't include any other non-error cases )
					{
						logMsg('error', options.method + ' request against ' + options.path + ' failed with status code=' + res.statusCode);
					}

					var rO = null;

					if ( response )                                                                     // not all ECX API calls return a response, e.g., DELETE; fti this call did return something
					{
						var rO = JSON.parse(response);                                                    // rO is raw response from ECX
						logMsg('debug', 'Generic requester returning=' + JSON.stringify(rO));
					}

					rObj.rO = rO;                                                                       // !!! REALLY CONFUSING: .rO is what came back from ECX; .response is the Express.js response object
					rObj.response = res;                                                                   

					cb(rObj);
				});
	});


/*
	Express waypoint (middleware) processing above captures the last client response object in the session state as property piggybackRes. If there is a network partition or if SPP crashes
	our HTTP(S) request will receive an error from the network stack. These errors are handled by this request event handler. Use the last captured response to send an error back to the client.
*/

	request.on('error', function(e){                                                            // e.g., ECONNREFUSED, ECONNRESET, EPROTO, etc... from kernel via network stack
			var piggybackRes = ecxSidHash[ecxSid].piggybackRes;
			
			var errMsg = 'SPP server unavailable in session ID=' + ecxSid + ' because of error=' + e;
			
			if ( piggybackRes )                                                                     // should always be set
			{
					logMsg('error', errMsg);
					piggybackRes.status(500).json({statusCode: 500, response: errMsg});              
					piggybackRes.end();  
			}
			else
			{
					logMsg('error', 'Unable to respond to client after ' + errMsg);
			}
	});


	if ( postBody )
	{
		request.write(JSON.stringify(postBody));
	}

	logMsg('info', 'Sending ' + guiReqProtocol.toUpperCase() + ' request for session ID(' + ecxSid + '):');
	logMsg('info', '    Request options for session ID(' + ecxSid + '):' + JSON.stringify(options));

	if ( postBody )                                                                             // FTI POST or PUT, mayhap contain password info
	{
			var scrubbedPostBody = scrubPostBody(postBody);
			logMsg('info', '    Request body to SPP(' + ecxSid + '):' + JSON.stringify(scrubbedPostBody));
	}
	
	request.end();
}


function arrayToHash(array, key)                                                                // call with key null to create hash whose key is same as value (useful for simple, non-object arrays)
{
	var hash = {};
	
	array.forEach(function(el){
		if ( ! key )                                                                            // e.g. [1,2,3] becomes {"1":1, "2":2, "3":3}
			hash[el] = el;
		else
			hash[el[key]] = el;
	});
	
	return hash;
}


function generateNameForUserIdentity(oFromNg)                                                   // EHI create user case, perforce a register call
{
	return oFromNg.username + '_' + oFromNg.hostAddress;                                          // our convention, perhaps questionable
}


function createRequestStore(stepsStream, ecxSid, reqId, funcName)                               // called only by the several init methods
{
		var ngState = ecxSidHash[ecxSid];
		
		ngState.endPoints[reqId] = {
				currentFunc: funcName,
		};

		stepsStream.forEach(function(step){
				step.reqId = reqId;
		});
}


function setRequestState(reqId, ecxSid, key, value)
{
		//logMsg('warn', 'SET reqId=' + reqId + ', key=' + key + ', value=' + value);
		var ngState = ecxSidHash[ecxSid];
		ngState.endPoints[reqId][key] = value;
}


function getRequestState(reqId, ecxSid, key)
{
		var ngState = ecxSidHash[ecxSid];
		//logMsg('warn', 'GET reqId=' + reqId + ', key=' + key);

		if ( ngState.endPoints[reqId] && ngState.endPoints[reqId][key] )
			return ngState.endPoints[reqId][key];

		return null;
}


function getCurrentFunc(reqId, ecxSid)
{
		return getRequestState(reqId, ecxSid, 'currentFunc');
}


function HIDE_getEndpointState(ecxSid, key)
{
	var ngState = ecxSidHash[ecxSid];

	if ( ngState.endPoints[ngState.currentFunc] && ngState.endPoints[ngState.currentFunc][key] )
		return ngState.endPoints[ngState.currentFunc][key];

	return null;
}


function HIDE_setEndpointState(ecxSid, key, value)                                                   // generic state capture, anything caller wants          
{
	var ngState = ecxSidHash[ecxSid];

	if ( ! ngState.endPoints[ngState.currentFunc] )
		ngState.endPoints[ngState.currentFunc] = {};

	ngState.endPoints[ngState.currentFunc][key] = value;
	
	try                                                                                          // in case object we stringify has circular refs
	{
		logMsg('debug', 'Set state for func=' + ngState.currentFunc + '.' + key + '=' + JSON.stringify(value));
	}
	catch(e)
	{
		;
	}

	return getEndpointState(ecxSid, key);
}


function getEcxHeaders(ecxSid)
{
	return {                                                                           
		'x-endeavour-sessionid': ecxSid,
		'Content-Type' : 'application/json',
	};   
}


function getSessionState(req, res)                                                                          
{
	var ecxSid = req.get(ECXSESSIONHEADER);
	var ngState = app.get(ecxSid);
	
	if ( ! ngState )                                                                                        // FTI ecomon probably got bounced while client up
	{
		setCorsHeaders(res);
		res.status(400).json({statusCode: 500, response: '500 Stale state; please restart client'});   	
		res.end();    
	}
	
	if ( ngState.stale )
		logMsg('warn', 'Returning stale state for session ID=' + ecxSid);
	
	return ngState;
}


/*
	NB: I removed the log message printing from here because, depending on caller's behavior, these messages could be painfully misleading. Specifically, the messages
	depend on the contents of the step object's verb and path properties. But it often happens that a caller will hardwire either of both of these properties in the call to
	SPP, but then use the unchanged step object to obtain a generic error description. This leads to bizarro phenomena like doing a PUT but the message generated here says 'GET',
	etc.
*/

function getGenericEcxErrorResponse(rObj, ecxSid, step)                                                // rObj is thing returned by SRTE; ** step is unused
{
	var failResult = {
			statusCode: rObj.response.statusCode,
			response: rObj.rO,                                                       
	};

	if ( rObj.e )                                
	{
			//logMsg('debug', "Failed to " + step.verb + " against " + step.path + ' because of error=' + JSON.stringify(rObj.e));
			failResult.response = rObj.e;
	}
	else
	{
			//logMsg('debug', "Failed to " + step.verb + " against " + step.path + ' because of error=' + JSON.stringify(rObj.rO));
	}

	return failResult;
}


function getConfiguration()
{
	try
	{
		var cfg = yaml.safeLoad(fs.readFileSync(cfgFile, 'utf8'))
	}
	catch(e)
	{
		logMsg('warn', "Failed to read configuration file'" + cfgFile + "' because of error=" + e.toString());
		return;
	}

	cors = JSON.parse(JSON.stringify(cfg.cors));
	useHttps = JSON.parse(JSON.stringify(cfg.https));

	if ( cfg.port )                                                                                         // NB: if specified, we listen here, either HTTPS | HTTP
		effectivePort = JSON.parse(JSON.stringify(cfg.port));

	if ( cfg.ecxInternal )
		ecxInternal = JSON.parse(JSON.stringify(cfg.ecxInternal));

	if ( cfg.ecxExternal )
		ecxExternal = JSON.parse(JSON.stringify(cfg.ecxExternal));	

	try                                                                                                     // stage all policy templates (and sprof)
	{
		var temp = fs.readFileSync(sprofTplFile, 'utf8');
		sprofTpl = JSON.parse(temp);
		
		temp = fs.readFileSync(hyppolTplFile, 'utf8');
		hyppolTpl = JSON.parse(temp);

		temp = fs.readFileSync(applicationProtectionPolicyTemplateFile, 'utf8');
		applicationProtectionPolicyTemplate = JSON.parse(temp);

		temp = fs.readFileSync(applicationRestorePolicyTemplateFile, 'utf8');
		applicationRestorePolicyTemplate = JSON.parse(temp);

		temp = fs.readFileSync(jobTplFile, 'utf8');
		jobTpl = JSON.parse(temp);

		temp = fs.readFileSync(stppolTplFile, 'utf8');
		stppolTpl = JSON.parse(temp);

		temp = fs.readFileSync(hyrpolTplFile, 'utf8');                                                       // common part of hypervisor recovery template
		hyrpolTpl = JSON.parse(temp);		

		temp = fs.readFileSync(triggerTplFile, 'utf8');
		triggerTpl = JSON.parse(temp);

		temp = fs.readFileSync(rfpolTplFile, 'utf8');
		rfpolTpl = JSON.parse(temp);
		
		temp = fs.readFileSync(rptpolTplFile, 'utf8');
		rptpolTpl = JSON.parse(temp);
	}
	catch(e)
	{
		logMsg('error', "Failed to read template file because of error=" + e.toString()); //!!! FATAL, NO? MAYBE THIS IS FATAL!
		return;
	}			
}


function getStartupArgs(args)
{
	for ( var n=0; n < args.length; n+=2 )
	{
		if ( ! startupArgs[args[n]] )
		{
			logMsg('error', "Command line argument '" + args[n] + "' is invalid");
			process.exit(1);
		}
		
		startupArgs[args[n]](args[1+n]);
	}
}


function logMsg(lvl, msg)
{
	if (arguments.length == 2){
		logger.log(lvl, msg);
    }else if (arguments.length > 2){
		var interpolatedString = msg;
        for (i=2;i<arguments.length;i++) {
            interpolatedString = interpolatedString.replace(/%s/i, arguments[i]);
        }
        logger.log(lvl, interpolatedString);
	}
}

function scrubPostBody(pb, fromClient)                                                          // NB: pb could be oFromNG, or POST, or PUT bodies
{
	var effPb = pb;                                                                               // assume it's POST or PUT

	if ( fromClient )                                                                             // FTI oFromNg being scrubbed; can't muck with it
	{
		effPb = JSON.parse(JSON.stringify(pb));

		if ( effPb.opProperties && effPb.opProperties.veServerInfo )                                // FTI veuser info which may itself contain password 
		{
			try
			{
				effPb.opProperties.veServerInfo.password = '********';
			}
			catch(e)
			{
				;
			}
		}

		if ( effPb.options )                                                                        // FTI likley applyOptions call with user creation; won't happen with a .ve call 
		{
			try
			{
				effPb.options.password = '********';
			}
			catch(e)
			{
				;
			}
		}

		if ( effPb.password )                                                                       // still check top-level property as well (also handles create user case)
		{
			effPb.password = '********';
		}

		if ( effPb.clientSecret )                                                                       // still check top-level property as well (also handles create user case)
		{
			effPb.clientSecret = '********';
		}

		if ( effPb.adminUserPass )                                                                       // still check top-level property as well (also handles create user case)
		{
			effPb.adminUserPass = '********';
		}
		
		if ( effPb.accesskey )                                                                       // still check top-level property as well (also handles create user case)
		{
			if(effPb.accesskey.access)
			effPb.accesskey.access = '********';

			if(effPb.accesskey.secret)
			effPb.accesskey.secret = '********';
		}
	}
	else  {
		if ( effPb.password )                                                                    // EHI POST | PUT body that's already been sent, no need to clone it.
		{
			effPb.password = '********';
		}

		if ( effPb.clientSecret )                                                                       // still check top-level property as well (also handles create user case)
		{
			effPb.clientSecret = '********';
		}

		if ( effPb.adminUserPass )                                                                       // still check top-level property as well (also handles create user case)
		{
			effPb.adminUserPass = '********';
		}

		if ( effPb.access )                                                                       // still check top-level property as well (also handles create user case)
		{
			effPb.access = '********';
		}

		 if ( effPb.secret )                                                                       // still check top-level property as well (also handles create user case)
		{
			effPb.secret = '********';
		}
	}
	return effPb;
}


function getLocalIp()
{
    var interfaces = os.networkInterfaces();
    var addresses = [];
    
    var localIp = '';
    var gotLocalIp = false;

    for ( var iface in interfaces )
    {
    	if ( gotLocalIp )
    		break;
     	
    	if ( iface.startsWith('ens') || iface.startsWith('enp') || iface.startsWith('eth') )                 // ens for Centos7, enp for vBox, eth for Centos6 and Ubuntu...
    	{
	        for ( var k2 in interfaces[iface] )
	        {
	            var address = interfaces[iface][k2];
	            
	            if ( address.family === 'IPv4' && ! address.internal )                                       // iface offers no way to get its name, e.g., ens160, eth0, docker0, etc. so must record all IPv4s
	            {
	            	localIp = address.address;
	            	gotLocalIp = true;
	            	break;
	            }
	        }
    	}
    }

    app.set('localIp', localIp);
    return localIp;
}

function setCorsHeaders(res)
{
	if ( cors )
	{
	    res.header("Access-Control-Allow-Origin", "*");                                                      // or more restrictive browser_IP:port here for origin
	    res.header('Access-Control-Allow-Methods', 'GET,PUT,POST,DELETE');
	    res.header("Access-Control-Allow-Headers", "Origin, X-Requested-With, Content-Type, Accept, Authorization, Access-Control-Allow-Origin, x-endeavour-sessionid, x-endeavour-locale");
	    res.header("Access-Control-Expose-Headers", "x-endeavour-sessionid");
	}
}


function initNgState(ecxSid)                                                                               // NB: UPDATE WHEN A NEW ENDPOINT IS WRITTEN
{
	return {
		ecxSid: ecxSid,
		accumState: {},	                                                                                       // used to communicate between steps; !! likely vestigial, replaced by .endPoints
		result: {},	
		userHashByName: null,
		endPoints: {                                                                                           // dynamically populated                                                                             
		},
	};
}


function validateNgRecovInput(ecxSessionId, oFromNg)
{
	var rMode = {
		valid: false,
		scheduled: false,
	};

	if ( oFromNg.name && oFromNg.trigger )                                                                   // FTI scheduled job
	{
		rMode.valid = true;
		rMode.scheduled = true;
	}

	if ( ! oFromNg.name && ! oFromNg.trigger )                                                               // FTI onDemand job
	{
		rMode.valid = true;
		rMode.scheduled = false;                          
	}

	logMsg('warn', 'Recovery mode=' + JSON.stringify(rMode));
	return rMode;
}


function validateNgUserInput(ecxSessionId, oFromNg)
{
	var userMode = {
		valid: false,
		createUser: false,
		veUser: false,
		username: oFromNg.username,
		password: oFromNg.password,
	};

	var effUserObj = oFromNg;

	if ( oFromNg.ve ) 
	{
		userMode.veUser = true;

		try
		{
			effUserObj = {
				username: oFromNg.opProperties.veServerInfo.username,
				password: oFromNg.opProperties.veServerInfo.password,
			};

			userMode.username = effUserObj.username;
			userMode.password = effUserObj.password;			
		}
		catch(e)                                                                              // EHI we were asked to validate non-existent VE user property, etc.
		{
			return userMode;
		}
	}

	if ( oFromNg.ao )                                                                         // EHI apply options case, user info elsewhere...again
	{
		try
		{
			effUserObj = {
				username: oFromNg.options.username,
				password: oFromNg.options.password,
			};

			userMode.username = effUserObj.username;
			userMode.password = effUserObj.password;			
		}
		catch(e)                                                                              // EHI we were asked to validate non-existent user property, etc.
		{
			return userMode;
		}
	}

	if ( effUserObj.username )                                                                // by itself -> user exists; w/password -> create user
	{
		if ( effUserObj.password )
		{
			userMode.valid = true;
			userMode.createUser = true;
		}
		else
		{
			userMode.valid = true;
		}
	}
	
	logMsg('debug', 'User mode=' + JSON.stringify(userMode));
	return userMode;
}


const escapePropsHash = {
	username: true,
};


const escapeRepls = {
	'%08': 'b',
	'%09': 't',
	'%0A': 'n',
	'%0B': 'v',
	'%0C': 'f',
	'%0D': 'r',		
	'%5C': '',	                                                                                 // NB: querystring.escape() will convert '\' to %5C
}

var escapeRxStr = null;


function escapeChars(oFromNg)                                                                  //!! BUG: won't work for nested props. DON'T FORGET
{
	if ( ! escapeRxStr )
	{
		escapeRxStr = '(';

		for ( code in escapeRepls )
		{
			escapeRxStr += code + '|';
		}

		escapeRxStr = escapeRxStr.replace(/\|$/, '');
		escapeRxStr += ')';
	}
	
	traverse(oFromNg).forEach(function(){
    if ( this.node )                                                                           // no reason this.node is defined, e.g., key with null or blank property
		{
			if ( escapePropsHash[this.key] )                                                   
			{
				logMsg('debug', 'Input property=' + this.key + ' has value=' + this.node);

				var escaped = require('querystring').escape(this.node);                                // encode any special escape sequences

				var escapeRx = new RegExp(escapeRxStr, 'g');                           

				escaped = escaped.replace(escapeRx, function($1){                                      // e.g., '%0D' -> \\r 				
					return '\\'+escapeRepls[$1];                                                         // NB: escapeRepls value for %05C is empty because this is an identity function for that code
				});      

				this.update(escaped);

				logMsg('debug', 'Updated property=' + this.key + ' with value=' + this.node);
			}
		}
	});

	return oFromNg;
}


function setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, result, skipSetFail)   
{
	if ( ! skipSetFail )                                                                       // FTI caller wants us to force failure in NEXT step (if any)
	{
		if ( funcStream.length > 0 )     
			funcStream[0].continue = 'fail';                                                       // MUST SET stream result if 'internal' (non-ECX) failure
	}

	if ( step.failsafe )
	{
		logMsg('warn', 'Driving failsafe function in step=' + step.name);
		step.failsafe(step, funcStream, ecxSessionId, oFromNg, cb)
	}

	setSerialStreamResult(ecxSessionId, result);                                               // NB: this behavior depends on this being last step, eh? Else it could be overriden by subseq
	step.cb(ecxSessionId, result.response);                                                   
	runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	                                   // let main driver end things via 'fail'	
}


// NB: present successful ending status, i.e., we're about to send response back to client

function setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, result)                       
{
	setSerialStreamResult(ecxSessionId, result);                                               // NB: this behavior depends on this being last step, eh? Else it could be overriden by subseq
	step.cb(ecxSessionId, result.response);                                                   
	runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	                                   	
}


// NB: present successful ending status, i.e., we're about to send response back to client

function returnBadRequestError(res)
{
	res.status(400).json({statusCode: 400, response: '400 Bad Request'});              
	res.end();   
}


function returnMissingSidError(res)
{
	res.status(401).json({statusCode: 401, response: '401 Session ID is missing'});              
	res.end();   
}


// NB: we use regIsSprof for both storage profile register (POST) and update (PUT) requests

var isSprof = function(step, funcStream, ecxSessionId, oFromNg, cb) {  
	var scrubbedNgInput = scrubPostBody(oFromNg, true);	                                   // prob not needed for sprof, cuz no user info...               
	
	logMsg('debug', "Input from client for storage profile in session ID='" + ecxSessionId + "'=" + JSON.stringify(scrubbedNgInput));

	sendRequestToEcx(step.verb, ecxExternal, step.path, ecxSessionId, null, function(sprofObj){   
		if ( sprofObj.e || sprofObj.response.statusCode != 200 )                               
		{
			var failResult = getGenericEcxErrorResponse(sprofObj, ecxSessionId, step);
			setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult);
		}
		else                                                                                   // EHI examine extant sprofs and respond based on function (create | update)
		{
			var ngState = ecxSidHash[ecxSessionId];    
			
			//@#@ if ( ngState.currentFunc == 'get_slapolicy' )                                      // piggy-backing on is_sprof, needs all sprofs
			if ( getCurrentFunc(step.reqId, ecxSessionId) == 'get_slapolicy' )               // piggy-backing on is_sprof, needs all sprofs
			{
				//@#@setEndpointState(ecxSessionId, 'sprofs', sprofObj);                            // preserve for next step; NB: this is the composite response & NOT that which comes from ECX alone				
				setRequestState(step.reqId, ecxSessionId, 'sprofs', sprofObj);                 // preserve for next step; NB: this is the composite response & NOT that which comes from ECX alone
				step.cb(ecxSessionId, sprofObj.rO);                                            // drive step cb     
				runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	                       // let main driver end things via 'fail'					
				return;					
			}
			
			var matchedSprof = null;

			sprofObj.rO.storageprofiles.forEach(function(sprof){
				if ( matchedSprof )
					return;

				if ( sprof.name == oFromNg.name )                                              //!! BUG, albeit unlikely: null == undefined
				{
					matchedSprof = sprof;
					return;
				}
			});

			//@#@ if ( ngState.currentFunc.startsWith('register') )                                  // as opposed to 'update' or maybe 'delete'
			if ( getCurrentFunc(step.reqId, ecxSessionId).startsWith('register') )                                  // as opposed to 'update' or maybe 'delete'
			{
				if ( matchedSprof )                                                            // FTI an error: register of extant storage profile
				{
					var failResult = {                                                         // try to choose code that reflects actual cause of error
						statusCode: 400,
						response: {rc: "Storage profile with name '" + oFromNg.name + "' already exists"},                    
					};

					setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult);				
				}
				else                                                                           // create the sprof by letting next step run
				{
					step.cb(ecxSessionId, null);                                               // drive step cb     
					runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	                   // let main driver end things via 'fail'					
					return;					
				}
			}
			//@#@ else if ( ngState.currentFunc.startsWith('update') ) 
			else if ( getCurrentFunc(step.reqId, ecxSessionId).startsWith('update') ) 
			{
				if ( ! matchedSprof )
				{
					var failResult = {                                                         // try to choose code that reflects actual cause of error
						statusCode: 400,
						response: {rc: "Storage profile with name '" + oFromNg.name + "' does not exist"},                    
					};

					setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult);					
				}
				else                                                                           // EHI update and sprof exists, let *next* step run
				{
					//@#@setEndpointState(ecxSessionId, 'editHref', matchedSprof.links.edit.href);
					setRequestState(step.reqId, ecxSessionId, 'editHref', matchedSprof.links.edit.href);					
					step.cb(ecxSessionId, sprofObj.rO);                                                   
					runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	   
				}				
			}
		}
	});	
};

function convertReplSubpolicies(sprofTpl, oFromNg, ecxSessionId, updateSprofInfo) {
		sprofTpl.name = oFromNg.name; 
		sprofTpl.description = oFromNg.description; 
		sprofTpl.type = oFromNg.type;                                            

	    if ( updateSprofInfo )                                                                   // FTI update case, keep description already in the sprof, i.e., GUI won't
		sprofTpl.description = updateSprofInfo.sprofDesc;                                    // alter desc and we don't want value in sprof template to govern ECC-11991

	    var subpolTplHash = arrayToHash(sprofTpl.spec.subpolicy, 'type');                             
	    oFromNg.spec.subpolicy.forEach(function(subpolicy) {
			if ( subpolicy.type == 'SNAPSHOT' ) {
				var tplTarget = subpolTplHash.SNAPSHOT;
				tplTarget.spec.option.retention = subpolicy.retention;
				tplTarget.participant = true;
				tplTarget.trigger = subpolicy.trigger;
				tplTarget.spec.option.label = subpolicy.label;
		    }

		    if ( subpolicy.type == 'REPLICATION' && ! subpolicy.software ) {
				var tplTarget = subpolTplHash.REPLICATION;
				tplTarget.trigger = subpolicy.trigger;
				tplTarget.participant = true;
				tplTarget.name = 'Backup Storage Replication';
				tplTarget.description = 'Backup Storage Replication';
				tplTarget.spec.destination = { 
					target:{
						href: subpolicy.site,
						resourceType: 'site',
						id: subpolicy.site,
					},
 					option: {
						retention: subpolicy.retention,
						useEncryption: subpolicy.useEncryption || false
					}
				};			
		    }

		    if ( subpolicy.type == 'REPLICATION' && subpolicy.software ) {
				var tplTarget = subpolTplHash.BACKUP;
				tplTarget.spec.option.retention = subpolicy.retention;
				tplTarget.participant = true;

				delete sprofTpl.spec.option;
				tplTarget.trigger = subpolicy.trigger;                                             // @@ SPROF - per RS and ECC-10979, only BACKUP is target of this trigger

				tplTarget.spec.destination = { 
					target:{
						href: subpolicy.site,
						resourceType: 'site',
						id: subpolicy.site,
					},
                    option: {
                        useEncryption: subpolicy.useEncryption || false
                    }
				};
		    }

		    if ( subpolicy.type == 'SP' ) {
				var tplTarget = subpolTplHash.SP;
				tplTarget.trigger = subpolicy.trigger;

				if ( subpolicy.useVsnap )
					tplTarget.useVsnap = subpolicy.useVsnap;
				
				if ( subpolicy.destination )
					tplTarget.destination = subpolicy.destination;

				tplTarget.participant = true;
			}
			
			if ( subpolicy.type == 'SPPOFFLOAD' ) {
				var tplTarget = subpolTplHash.SPPOFFLOAD;
				tplTarget.spec.destination.option.retention = subpolicy.retention;
				tplTarget.trigger = subpolicy.trigger;
				tplTarget.source = subpolicy.source;
				tplTarget.spec.destination.target = subpolicy.target;
				tplTarget.participant = true;	
			}
			
			if ( subpolicy.type == 'SPPARCHIVE' ) {
				var tplTarget = subpolTplHash.SPPARCHIVE;
				tplTarget.spec.destination.option.retention = subpolicy.retention;
				tplTarget.trigger = subpolicy.trigger;
				tplTarget.source = subpolicy.source;
				tplTarget.spec.destination.target = subpolicy.target;
				tplTarget.participant = true;	
	     	}	 
	    });

	    var effPolicies = sprofTpl.spec.subpolicy;

	    for ( var n = effPolicies.length-1; n >= 0; n-- ) {
			if ( ! effPolicies[n].participant ) 
			effPolicies.splice(n, 1);
		    else
			delete effPolicies[n].participant;
	    }
}


function KEEP_convertReplSubpolicies(sprofTpl, oFromNg, ecxSessionId)
{
	sprofTpl.name = oFromNg.name;
	sprofTpl.description = oFromNg.description;

	//logMsg('error', 'HMR=' + JSON.stringify(sprofTpl));

//!!! ADD TEST for simple = false; if so (i.e., advanced mode), then pass contents of oFromNg directly to ECX

	var subpolHash = arrayToHash(sprofTpl.spec.subpolicy, 'type');                             
	var sites = ecxSidHash[ecxSessionId].siteHashByName;

	sprofTpl.spec.option = {
		rpo: oFromNg.spec.rpo,
	}	

	oFromNg.spec.subpolicy.forEach(function(subpolicy){                                        // NB: 1-1, override tpl with live values from NG; in 'simple' mode can be only
		if ( subpolicy.type == 'SNAPSHOT' )
		{
			var tplTarget = subpolHash.SNAPSHOT;
			tplTarget.spec.option.retention = subpolicy.retention;	
			tplTarget.participant = true;	 ///!!! EXPERIMENT	
			tplTarget.trigger = subpolicy.trigger;
			tplTarget.spec.option.label = subpolicy.label;	
		}

		if ( subpolicy.type == 'REPLICATION' && ! subpolicy.software )                                                 
		{
			var tplTarget = subpolHash.REPLICATION;
			tplTarget.spec.destination.option.retention = subpolicy.retention;			
			tplTarget.participant = true;
			
			tplTarget.spec.destination = { 
				target:{
					href: sites[subpolicy.site].links.self.href,
					resourceType: 'site',
					id: sites[subpolicy.site].id,
				}
			};	
			
			tplTarget.trigger = subpolicy.trigger;
		}

		if ( subpolicy.type == 'REPLICATION' && subpolicy.software )                                                 
		{
			var tplTarget = subpolHash.BACKUP;
			tplTarget.spec.option.retention = subpolicy.retention;			
			tplTarget.participant = true;
			
			tplTarget.spec.destination = { 
				target:{
					href: sites[subpolicy.site].links.self.href,
					resourceType: 'site',
					id: sites[subpolicy.site].id,
				}
			};			
		}

		if ( subpolicy.type == 'SP' )                                                 
		{
			var tplTarget = subpolHash.SP;

			//!! DON'T SEEM TO EXIST FOR SP: tplTarget.spec.option.retention = subpolicy.retention;			
			tplTarget.trigger = subpolicy.trigger;
			tplTarget.participant = true;	
		}		
	});

	//logMsg('warn', 'BEFORE=' + JSON.stringify(sprofTpl));
	var effPolicies = sprofTpl.spec.subpolicy;

	//!! EXPERIMENT: KEEP for ( var n = effPolicies.length-1; n > 0; n-- )                                       // in simple mode this will examine indices 2 and 1 (0 is always SNAPSHOT and is required)
	for ( var n = effPolicies.length-1; n >= 0; n-- ) 		
	{
		if ( ! effPolicies[n].participant ) 
			effPolicies.splice(n, 1);
		else
			delete effPolicies[n].participant;
	}
}


//  If policy IS, then the ECX policy&job creation step is not required. However, you still must present the NG POST body to ECX at its endpoint /api/hypervisor?action=applyStorageProfiles

var isProtectionPolicy = function(step, funcStream, ecxSessionId, oFromNg, cb) {           
	var scrubbedNgInput = scrubPostBody(oFromNg, true);	                               // prob not needed...
	logMsg('warn', "Input from client for protection policy in session ID='" + ecxSessionId + "'=" + JSON.stringify(scrubbedNgInput));

	sendRequestToEcx(step.verb, ecxExternal, step.path, ecxSessionId, null, function(policyObj){   
		if ( policyObj.e || policyObj.response.statusCode != 200 )                               
		{
			var failResult = getGenericEcxErrorResponse(policyObj, ecxSessionId, step);			
			setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult);
		}
		else                                                                             // EHI examine extant sprofs and respond based on function (create | update)
		{
			var ngState = ecxSidHash[ecxSessionId];    
			var createEcxPolicyNames = {};
			var policiesList = [];

			oFromNg.slapolicies.forEach(function(slap){
				createEcxPolicyNames[oFromNg.subtype + '_' + slap.name] = 1;                 // e.g., 'vmware_Gold'
				logMsg('debug', 'Added policy/job name=' + oFromNg.subtype + '_' + slap.name);
			});

			logMsg('debug', "Extant hypervisor protection policy='" + ecxSessionId + "'=" + JSON.stringify(policyObj.rO));

			policyObj.rO.policies.forEach(function(policy){
				if ( createEcxPolicyNames[policy.name] )                                     // FTI ECX policy (and job) exist; no need to create 
				{
					policiesList.push(policy.name);
					delete createEcxPolicyNames[policy.name];
					logMsg('debug', 'Deleted policy/job name=' + policy.name);
					return;
				}
			});
			setRequestState(step.reqId, ecxSessionId, 'policiesList', policiesList);
			if (policiesList.length === 0) {
				funcStream[1].continue = 'skip';
			}
			logMsg('info', 'Policy/job pairs to create=' + JSON.stringify(createEcxPolicyNames));

// EHI createEcxPolicyNames hash contains names of ECX policies (and their jobs) that must be created.

			if ( Object.keys(createEcxPolicyNames).length < 1 )                            // FTI all indicated policies already exist, so skip creation step & drive link_resources
			{
				if ( funcStream.length > 0 )                                             
					funcStream[0].continue = 'skip';                                           // don't try to re-create extant ECX policies & jobs

				step.cb(ecxSessionId, null);                                               
				runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 
			}
			else                                                                           // create the ECX policies by letting next step run
			{
				//@#@setEndpointState(ecxSessionId, 'createPolicyNames', createEcxPolicyNames);
				setRequestState(step.reqId, ecxSessionId, 'createPolicyNames', createEcxPolicyNames);
				step.cb(ecxSessionId, null);                                                 // drive step cb     
				runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	                     // let main driver end things via 'fail'					
				return;					
			}
		}
	});	
};


// NB: Here only for scheduled recovery policies. If we find one with same name, that's an error

var isRecoveryPolicy = function(step, funcStream, ecxSessionId, oFromNg, cb) { 
	var scrubbedNgInput = scrubPostBody(oFromNg, true);                                 // prob not needed	                           
	logMsg('warn', "Input from client for recovery policy in session ID='" + ecxSessionId + "'=" + JSON.stringify(scrubbedNgInput));

	var path = step.path + '?filter=';
	var filter = '[{"property":"serviceId","value":"com.syncsort.dp.xsb.serviceprovider.recovery.hypervisor"},{"property":"subType", "value":"' + oFromNg.subType + '"}]';
	var escFilter = escape(filter);	

	sendRequestToEcx(step.verb, ecxExternal, path+escFilter, ecxSessionId, null, function(policyObj){   
		if ( policyObj.e || policyObj.response.statusCode != 200 )                               
		{
			var failResult = getGenericEcxErrorResponse(policyObj, ecxSessionId, step);			
			setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult);
		}
		else                                                                               // EHI examine extant sprofs and respond based on function (create | update)
		{
			var ngState = ecxSidHash[ecxSessionId];    

			//var createEcxPolicyNames = {};

			//logMsg('warn', "Extant hypervisor recovery policy='" + ecxSessionId + "'=" + JSON.stringify(policyObj.rO));
			
			var policyIs = false;

			policyObj.rO.policies.forEach(function(policy){
				if ( policyIs )
					return;

				if ( policy.name == oFromNg.name )                                             // FTI recovery policy IS; that's an error 
				{
					policyIs = true;
					return;
				}
			});

			if ( policyIs )
			{
				var failResult = {
					statusCode: 400,                                                           
					response: {rc: "Scheduled recovery policy '" + oFromNg.name + "' already exists"},   
				};		
				
				setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
			}
			else                                                                             // create the ECX policies by letting next step run
			{
				step.cb(ecxSessionId, null);                                                   // drive step cb     
				runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	                       // continue with likely policy creation step					
				return;					
			}
		}
	});	
};


var recoveryMode = function(step, funcStream, ecxSessionId, oFromNg, cb)
{
	var rMode = validateNgRecovInput(ecxSessionId, oFromNg);                             // either use extant user (by its identity name) or create via username & password - both is wrong

	if ( ! rMode.valid )                                                                 // FTI fatal error
	{
		logMsg('error', 'Detected invalid recovery mode input from NG for function=' + step.name);

		var failResult = {                                                                 // try to choose code that reflects actual cause of error
			statusCode: 400,
			response: {rc: 'Input to recovery policy creation contains invalid data'},                     
		};

		setSerialStreamResult(ecxSessionId, failResult);                                   // do NOT use setStepFailResult here because it wants to continue driving runSerialSteps  
		step.cb(ecxSessionId, null);                                                       // drive cb after step (largely pro forma)     
		cb();                                                                              // short circuit recursive stepper - drive FINAL (before response to client) cb in order to send response to client
		return;
	}

	if ( ! rMode.scheduled )                                                             // FTI onDemand recovery, no point in looking for an extant policy                                       
	{
		if ( funcStream.length > 0 )                                             
			funcStream[0].continue = 'skip';                                            
	}

	setRequestState(step.reqId, ecxSessionId, 'rMode', rMode);                            

	step.cb(ecxSessionId, null);                                                                
	runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	
}


var createTrigger = function(step, funcStream, ecxSessionId, oFromNg, cb)            
{
	var rMode = getRequestState(step.reqId, ecxSessionId, 'rMode');

	if ( ! rMode || ! rMode.scheduled )                                                  // FTI no need to create schedule (!rMode.scheduled ->an onDemand recovery is in play)
	{
		step.cb(ecxSessionId, null);                                                                
		runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
	}
	else
	{
		var triggerTplClone = JSON.parse(JSON.stringify(triggerTpl));
		triggerTplClone.name = oFromNg.name;
		triggerTplClone.properties = oFromNg.trigger;
		//!!triggerTplClone.properties.activateDate = (new Date).getTime();
		triggerTplClone.properties.activateDate = oFromNg.trigger.activateDate;            // ahem, ECC-11717
		triggerTplClone.properties.deactivateDate = null;
	
		sendRequestToEcx(step.verb, ecxExternal, step.path, ecxSessionId, triggerTplClone, function(trigObj){    // Create trigger
			if ( trigObj.e || trigObj.response.statusCode != 201 )                                
			{
				var failResult = getGenericEcxErrorResponse(trigObj, ecxSessionId, step);
				setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);	          
			} 
			else
			{
				setRequestState(step.reqId, ecxSessionId, 'triggerId', trigObj.rO.id);         // set intermediate status in case someone interested
				setRequestState(step.reqId, ecxSessionId, 'trigger', trigObj.rO);              // YES, a bit wasteful; want to return this along with PJ in success case

				step.cb(ecxSessionId, null);                                                                
				runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
			} 
		});
	}
}


// 8/2 ADDING 'NEW' logic to update & re-use an extant identity
var createUser = function(step, funcStream, ecxSessionId, oFromNg, cb) {  
	var postBody = null;

	var userLit = 'user';
	var umHrefName = 'userHref';

	var duplIdentityType = 'pr';                                                     // as in 'provider', e.g., a hypervisor password

	var regularUser = false;
	var createUserInstances = false;

// NB: is it not necessary to include type or ostype in creating a user. However, their respective defaults are 'system' and empty string. Sencha UI seems to impose 'LINUX' as osType

	if ( oFromNg.ve )
	{
		var fauxFromNg = oFromNg.opProperties.veServerInfo;

		postBody = {  
			name: require('querystring').unescape(generateNameForUserIdentity(fauxFromNg)),
			password: fauxFromNg.password,
			username: require('querystring').unescape(fauxFromNg.username),
		};

		userLit = 'VE user';
		umHrefName = 'veUserHref';	
		
		duplIdentityType = 've';
		
		delete oFromNg.ve;                                                             // !! A MUST        
	}
	else if ( oFromNg.ao )                                                           // sigh...applyOptions case, look elsewhere for the user info
	{
		var fauxFromNg = oFromNg.options;

		postBody = {  
			name: require('querystring').unescape(fauxFromNg.username + '_' + oFromNg.resources[0].id),    // ****NB: can't create standard username_resourcename here****
			password: fauxFromNg.password,
			username: require('querystring').unescape(fauxFromNg.username),
		};

		duplIdentityType = 'ao';
	}	
	else
	{
		//regularUser = true; //uncomment to force error in the creation of a regular (non-VE) user. Sort of assumes that we're here by hypervisor registration path
		postBody = {  
			name: (oFromNg.name && oFromNg.name.length > 0) ? oFromNg.name : require('querystring').unescape(generateNameForUserIdentity(oFromNg)),
			password: oFromNg.password,
			username: require('querystring').unescape(oFromNg.username),
		};
	}

	if ( regularUser ) //!! HERE FOR ERROR TESTING ONLY
	{
		delete postBody.name;
		delete postBody.username;
	}

// NB: we've not yet cleaned oFromNg.ao (if set). All success branches below must be sure to do this...

	sendRequestToEcx(step.verb, step.port, step.path, ecxSessionId, postBody, function(rObj){           // know nothing about the dupl ident; no choice but to get all identities
		if ( rObj.e ) 
		{
			var failResult = getGenericEcxErrorResponse(rObj, ecxSessionId, step); 
			logMsg('error', 'Failed to create SPP ' + userLit + ' for session ID=' + ecxSessionId + ' because of error=' + JSON.stringify(e)); //!! MAYBE LOSE THIS MSG
			setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);			
			return;
		}   
		
		var crO = rObj.rO;

		if ( rObj.response.statusCode != 201 && crO && (crO.id == 'IdentityUserExistsException' || crO.id === 'XSBException'))        // FTI identity exists, UPDATE IT & USE IT
		{
			if ( oFromNg.ao )                                                            // applyOptions doesn't fit our normal identity naming scheme cuz no hostAddress
			{
				oFromNg.identityNameAo = postBody.name;                                    // so use the identity name that just elicited the user exists exception
				delete oFromNg.ao;
			}

			updateIdentity(postBody, umHrefName, step, funcStream, ecxSessionId, oFromNg, duplIdentityType, createUserInstances, null, cb);     // but need to get it first 
		}
		else if ( rObj.response.statusCode != 201 )                                                                                                      
		{
			var failResult = getGenericEcxErrorResponse(rObj, ecxSessionId, step); 
			setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
		}
		else                                                                           // EHI we actually created user in crO; NB: IF registration fails, we delete this identity
		{
			delete oFromNg.ao;    

			var ngState = ecxSidHash[ecxSessionId];                                      // push userHref property ahead to next step
			setRequestState(step.reqId, ecxSessionId, umHrefName, crO.links.self.href);
			logMsg('warn', 'Created identity with name=' + postBody.name + ' and href=' + JSON.stringify(crO.links.self.href));

			var createdIdentities = getRequestState(step.reqId, ecxSessionId, 'createdIdentities');

			if ( ! createdIdentities )
			{
				createdIdentities = [crO.links.self.href];
				setRequestState(step.reqId, ecxSessionId, 'createdIdentities', createdIdentities); 
			}
			else 
			{
				createdIdentities.push(crO.links.self.href);
			}

			step.cb(ecxSessionId, crO);                                                   
			runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 			
		}
	});
};

var createUserInstances = function(step, funcStream, ecxSessionId, oFromNg, postBody, currentArrayObject, arrayLength, cb) {  
	
	var instancePassword = postBody.password;
	var umHrefName = 'userHref';
	var userLit = 'user';
	var umHrefName = 'userHref';
	var createUserInstances = true;

	var duplIdentityType = 'pr';                                                     // as in 'provider', e.g., a hypervisor password

	sendRequestToEcx(step.verb, step.port, step.path, ecxSessionId, postBody, function(rObj) {           // know nothing about the dupl ident; no choice but to get all identities
		if ( rObj.e ) 
		{
			var failResult = getGenericEcxErrorResponse(rObj, ecxSessionId, step); 
			logMsg('error', 'Failed to create SPP ' + userLit + ' for session ID=' + ecxSessionId + ' because of error=' + JSON.stringify(e)); //!! MAYBE LOSE THIS MSG
			setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);			
			return;
		}   
		
		var crO = rObj.rO;

		if ( rObj.response.statusCode != 201 && crO && (crO.id == 'IdentityUserExistsException' || crO.id === 'XSBException'))        // FTI identity exists, UPDATE IT & USE IT
		{
			if ( oFromNg.ao )                                                            // applyOptions doesn't fit our normal identity naming scheme cuz no hostAddress
			{
				oFromNg.identityNameAo = postBody.name;                                    // so use the identity name that just elicited the user exists exception
				delete oFromNg.ao;
			}

			updateIdentity(postBody, umHrefName, step, funcStream, ecxSessionId, oFromNg, duplIdentityType, createUserInstances, instancePassword, cb);     // but need to get it first 
		}
		else if ( rObj.response.statusCode != 201 )                                                                                                      
		{
			var failResult = getGenericEcxErrorResponse(rObj, ecxSessionId, step); 
			setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
		}
		else                                                                           // EHI we actually created user in crO; NB: IF registration fails, we delete this identity
		{
			delete oFromNg.ao;    

			var ngState = ecxSidHash[ecxSessionId];                                      // push userHref property ahead to next step
			logMsg('warn', 'Created identity with name=' + postBody.name + ' and href=' + JSON.stringify(crO.links.self.href));

			var createdIdentities = getRequestState(step.reqId, ecxSessionId, 'createdIdentities');

			if ( ! createdIdentities )
			{
				createdIdentities = [crO.links.self.href];
				setRequestState(step.reqId, ecxSessionId, 'createdIdentities', createdIdentities); 
			}
			else 
			{
				createdIdentities.push(crO.links.self.href);
			}

			step.cb(ecxSessionId, crO);  
			if(currentArrayObject==arrayLength-1){
				runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 			

			}                                                 
		}
	});
};

function updateIdentity(duplFailPostBody, umHrefName, step, funcStream, ecxSessionId, oFromNg, duplIdentityType, createUserInstances, instancePassword, cb)    // EHI identity create failed cuz user exists
{
	sendRequestToEcx('GET', step.port, step.path, ecxSessionId, null, function(usersObj){           // know nothing about the dupl ident; no choice but to get all identities
		if ( usersObj.e || usersObj.response.statusCode != 200 )                         
		{
			var failResult = getGenericEcxErrorResponse(usersObj, ecxSessionId, step);
			delete oFromNg.identityNameAo;                                                              // pointless but clean
			setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
		}
		else                                                                                          // look thru identities to find one that just caused duplicate error
		{
			var ngState = ecxSidHash[ecxSessionId];                                                     // push userHref property ahead to next step

			var duplName = duplFailPostBody.name;                                                       // from POST body name that got duplicate identity (OK for update & register)

			if ( oFromNg.identityNameAo )
			{
				duplName = oFromNg.identityNameAo;
				logMsg('warn', 'HELPMERHONDA=' + JSON.stringify(oFromNg));
			}
			
			if ( usersObj.rO.users.length < 1 )                                                         // FTI we're running as an RBAC user that can't see users
			{
					logMsg('warn', 'Unable to update password for user name=' + duplName + ' because no user identities were found');
					delete oFromNg.identityNameAo;                                                          // !!! IF LETTING NEXT STEP RUN, DO THIS
					
					step.cb(ecxSessionId, null);                                                                
					runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);                                  // !!! SHOULD THIS FUNCTION BE ALLOWED TO CONTINUE?                                    
					return;
			}

			logMsg('warn', 'Looking for extant identity=' + duplName);

			var duplIdentity = null;

			usersObj.rO.users.forEach(function(user){
				if ( duplIdentity )
					return;

				if ( user.name == duplName )
					duplIdentity = user;
			});

			//!!!! FORCE THIS TEST using admin user
			if ( ! duplIdentity )                                                                       // getting a bit overwrought, no?
			{
					logMsg('warn', 'Unable to update password for user name=' + duplName + ' because its identity could not be found');
					
					var failResult = {
							statusCode: 404,
							response: {rc: 'Failed to find extant identity for user name=' + duplName},     
					};
					
					setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);			
					return;
			}
			
			if(createUserInstances == false)
			setRequestState(step.reqId, ecxSessionId, umHrefName, duplIdentity.links.self.href);

			logMsg('warn', 'Using extant identity=' + JSON.stringify(duplIdentity));
			
			var path = step.path + '/' + duplIdentity.id; 


// !!! NB: you could use duplIdentityType entirely and lose the identityNameAo object; also clean the .ao flag much earlier
			
			//if ( hasPropertyPath(oFromNg, 'opProperties.veServerInfo'.split('.') ) )                    // we removed the .ve flag set by caller; so simply check again
			if(createUserInstances == false) { 
				if ( duplIdentityType == 've' )                                                             // get correct password 
				{
					duplFailPostBody.password	= oFromNg.opProperties.veServerInfo.password;                 // update VE identity password
				}
				else
					duplFailPostBody.password = oFromNg.password;                                           // otherwise update, say, hypervisor identity

				if ( oFromNg.identityNameAo )
				{
					duplFailPostBody.password = oFromNg.options.password;				
					delete oFromNg.identityNameAo;
				}
			}
			else
			duplFailPostBody.password = instancePassword;

			logMsg('warn', 'UPDATING EXTANT IDENTITY'); 
	
			sendRequestToEcx('PUT', step.port, path, ecxSessionId, duplFailPostBody, function(uuObj){   // update extant identity with what may be a new password...vraiment?
				if ( uuObj.e || uuObj.response.statusCode != 200 )   
				{
					var failResult = getGenericEcxErrorResponse(uuObj, ecxSessionId, step); 
					logMsg('error', 'Failed to update user password for session ID=' + ecxSessionId + ' because of error=' + JSON.stringify(uuObj.e)); //!! MAYBE LOSE THIS MSG
					setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);			
					return;
				}
				
				step.cb(ecxSessionId, null);                                                                
				runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);                                    // AFTER async GET call completes (here, and nowhere else) do this to drive next step
			});  
		}
	}); 
}

var createKey = function(step, funcStream, ecxSessionId, oFromNg, postBody, cb) {  
	
	sendRequestToEcx(step.verb, step.port, step.path, ecxSessionId, postBody, function(rObj) {           // know nothing about the dupl ident; no choice but to get all identities
		if ( rObj.e ) 
		{
			var failResult = getGenericEcxErrorResponse(rObj, ecxSessionId, step); 
			logMsg('error', 'Failed to create SPP key for session ID=' + ecxSessionId + ' because of error=' + JSON.stringify(e)); //!! MAYBE LOSE THIS MSG
			setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);			
			return;
		}   
		
		var crO = rObj.rO;

		if ( rObj.response.statusCode != 200 && crO && crO.id == 'IdentityKeyExistsException' )        // FTI identity exists, UPDATE IT & USE IT
		{
			ngpCoreHelper.getKeys(ecxSessionId, function(getKeysResObj) {
				
				var path = step.path;

				var arrayOfKeys = getKeysResObj.rO['keys'];
				for (var i = 0; i < arrayOfKeys.length; i++) {
					if((arrayOfKeys[i].name == postBody.name) && (arrayOfKeys[i].keytype == postBody.keytype)) {
						path = step.path + '/' + arrayOfKeys[i].id;
						break;
					}
				}
				delete postBody.name;
				delete postBody.keytype;
				
				logMsg('warn', 'UPDATING KEY'); 

				sendRequestToEcx('PUT', step.port, path, ecxSessionId, postBody, function(uuObj){   // update extant identity with what may be a new password...vraiment?
					if ( uuObj.e || uuObj.response.statusCode != 200 )   
					{
						var failResult = getGenericEcxErrorResponse(uuObj, ecxSessionId, step); 
						logMsg('error', 'Failed to update key for session ID=' + ecxSessionId + ' because of error=' + JSON.stringify(uuObj.e)); //!! MAYBE LOSE THIS MSG
						setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);			
						return;
					}
				
					step.cb(ecxSessionId, null);                                                                
					runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);                                    // AFTER async GET call completes (here, and nowhere else) do this to drive next step
				}); 
			}); 
		} 
		else if ( rObj.response.statusCode != 200 )                                                                                                      
		{
			var failResult = getGenericEcxErrorResponse(rObj, ecxSessionId, step); 
			setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
		}
		else                                                                           // EHI we actually created user in crO; NB: IF registration fails, we delete this identity
		{
			logMsg('warn', 'Created Key with name=' + postBody.name + ' and href=' + JSON.stringify(crO.links.self.href));

			var createdKeys = getRequestState(step.reqId, ecxSessionId, 'createdKeys');

			if ( ! createdKeys )
			{
				createdKeys = [crO.links.self.href];
				setRequestState(step.reqId, ecxSessionId, 'createdKeys', createdKeys); 
			}
			else 
			{
				createdKeys.push(crO.links.self.href);
			}

			step.cb(ecxSessionId, crO);  
			runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 			

		}
	});
};


function LAST_updateIdentity(duplFailPostBody, umHrefName, step, funcStream, ecxSessionId, oFromNg, cb)    // EHI identity create failed cuz user exists
{
	sendRequestToEcx('GET', step.port, step.path, ecxSessionId, null, function(usersObj){           // know nothing about the dupl ident; no choice but to get all identities
		if ( usersObj.e || usersObj.response.statusCode != 200 )                         
		{
			var failResult = getGenericEcxErrorResponse(usersObj, ecxSessionId, step);
			delete oFromNg.identityNameAo;                                                              // pointless but clean
			setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
		}
		else                                                                                          // look thru identities to find one that just caused duplicate error
		{
			var ngState = ecxSidHash[ecxSessionId];                                                     // push userHref property ahead to next step

			var duplName = duplFailPostBody.name;                                                       // from POST body name that got duplicate identity (OK for update & register)

			if ( oFromNg.identityNameAo )
			{
				duplName = oFromNg.identityNameAo;
				logMsg('warn', 'HELPMERHONDA=' + JSON.stringify(oFromNg));
			}
			
			if ( usersObj.rO.users.length < 1 )                                                         // FTI we're running as an RBAC user that can't see users
			{
					logMsg('warn', 'Unable to update password for user name=' + duplName + ' because no user identities were found');
					delete oFromNg.identityNameAo;                                                          // !!! IF LETTING NEXT STEP RUN, DO THIS
					
					step.cb(ecxSessionId, null);                                                                
					runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);                                  // !!! SHOULD THIS FUNCTION BE ALLOWED TO CONTINUE?                                    
					return;
			}

			logMsg('warn', 'Looking for extant identity=' + duplName);

			var duplIdentity = null;

			usersObj.rO.users.forEach(function(user){
				if ( duplIdentity )
					return;

				if ( user.name == duplName )
					duplIdentity = user;
			});

			//!!!! FORCE THIS TEST using admin user
			if ( ! duplIdentity )                                                                       // getting a bit overwrought, no?
			{
					logMsg('warn', 'Unable to update password for user name=' + duplName + ' because its identity could not be found');
					
					var failResult = {
							statusCode: 404,
							response: {rc: 'Failed to find extant identity for user name=' + duplName},     
					};
					
					setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);			
					return;
			}

			setRequestState(step.reqId, ecxSessionId, umHrefName, duplIdentity.links.self.href);
			logMsg('warn', 'Using extant identity=' + JSON.stringify(duplIdentity));
			
			var path = step.path + '/' + duplIdentity.id; 

			if ( hasPropertyPath(oFromNg, 'opProperties.veServerInfo'.split('.') ) )                    // we removed the .ve flag set by caller; so simply check again
			{
					duplFailPostBody.password	= oFromNg.opProperties.veServerInfo.password;                 // update VE identity password
			}
			else
					duplFailPostBody.password = oFromNg.password;                                           // otherwise update, say, hypervisor identity

			if ( oFromNg.identityNameAo )
			{
				duplFailPostBody.password = oFromNg.options.password;				
				delete oFromNg.identityNameAo;
			}
			

			logMsg('warn', 'UPDATING EXTANT IDENTITY'); 


			sendRequestToEcx('PUT', step.port, path, ecxSessionId, duplFailPostBody, function(uuObj){   // update extant identity with what may be a new password...vraiment?
				if ( uuObj.e || uuObj.response.statusCode != 200 )   
				{
					var failResult = getGenericEcxErrorResponse(uuObj, ecxSessionId, step); 
					logMsg('error', 'Failed to update user password for session ID=' + ecxSessionId + ' because of error=' + JSON.stringify(uuObj.e)); //!! MAYBE LOSE THIS MSG
					setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);			
					return;
				}
				
				step.cb(ecxSessionId, null);                                                                
				runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);                                    // AFTER async GET call completes (here, and nowhere else) do this to drive next step
			});  
		}
	}); 
}


function getRegistrationTypology(req, providerType)                                               // NB: providerType is storage | hypervisor | appserver | scriptserver; comes from API endpoint
{
	var typology = null;

	traverse(req.body).forEach(function(){
		if ( typology )
			return;

        if ( this.node )                                                                          // no reason this.node is defined, e.g., key with null or blank property
		{
			if ( providerType == 'storage' && this.key == 'type' )			                   
			{
				typology = {
					resourceType: 'storage',
					type: this.node,
					path: '/api/storage',
				};
				
				return;
			}

			if ( providerType == 'hypervisor' && this.key == 'type' )                                   // hypervisor POST body supported type as of 5/26
			{
				typology = {
					resourceType: 'hypervisor',
					type: this.node,
					path: '/api/hypervisor',		 
				};
				
				return;
			}

			if ( providerType == 'appserver' )			                                        // NB: appserver provides NO type info
			{
				typology = {
					resourceType: 'appserver',
					type: 'app',                                                                // or 'unknown' (which is true cuz no type info)
					path: '/api/appserver',					
				};								
				
				return;
			}

			if ( providerType == 'scriptserver' )			                                        // NB: scriptserver provides NO type info
			{
				typology = {
					resourceType: 'scriptserver',
					type: 'script',                                                                // or 'unknown' (which is true cuz no type info)
					path: '/api/scriptserver',					
				};								
				
				return;
			}

			if ( providerType == 'smtp' )    	                            
			{
				typology = {
					resourceType: 'smtp',
					type: 'smtp',                                                       
					path: '/api/smtp',					
				};								
				
				return;
			}

			if ( providerType == 'ldap' )    	                            
			{
				typology = {
					resourceType: 'ldap',
					type: 'ldap',                                                 
					path: '/api/ldap',					
				};								
				
				return;
			}			
		}
	});	
	
	logMsg('debug', 'Typology=' + JSON.stringify(typology));
	return typology;
}


function service_header() {
    addDoubleSpace();
    logMsg('info', "Starting Node-cdm-service ...");
    logMsg('debug', "Log level=" + winston.level);
}


function addDoubleSpace(){                                                                      // FZM: please leave this be; thanks. -PMB
    logMsg('info', '');
    logMsg('info', '');
}


function hasPropertyPath(o, pathSteps)
{
	if ( traverse.has(o, pathSteps) )
		return true;

	return false;
}


function escape(s)
{
	var queryString =  require('querystring').escape(s.replace(/\\/g,"\\\\"));
	return queryString;
}

exports.userMode = userMode;
exports.regDeleteUser = regDeleteUser;
exports.isSprof = isSprof; 
exports.isProtectionPolicy = isProtectionPolicy;
exports.isRecoveryPolicy = isRecoveryPolicy;
exports.SECUREPORT = SECUREPORT;
exports.escape = escape;
exports.recoveryMode = recoveryMode;
exports.fs = fs;
exports.PUBLICDIR = PUBLICDIR;
exports.createTrigger = createTrigger;
exports.createUser = createUser;
exports.createUserInstances = createUserInstances;
exports.createKey = createKey;
exports.getRegistrationTypology = getRegistrationTypology;
exports.scrubPostBody = scrubPostBody;
exports.hasPropertyPath = hasPropertyPath;
exports.si = si;
exports.createRequestStore = createRequestStore;
exports.setRequestState = setRequestState;
exports.getRequestState = getRequestState;
exports.getCurrentFunc = getCurrentFunc;
exports.productionMode = productionMode;
