if (!('contains' in String.prototype)) {
	  String.prototype.contains = function(str, startIndex) {
	    return ''.indexOf.call(this, str, startIndex) !== -1;
	  };
	}

if (typeof String.prototype.startsWith != 'function') {
    String.prototype.startsWith = function(prefix) {
        return this.slice(0, prefix.length) == prefix;
    };
}
		 
if (typeof String.prototype.endsWith != 'function') {
    String.prototype.endsWith = function(suffix) {
        return this.slice(-suffix.length) == suffix;
    };
}

    
var ngp = require('./ecxngp');   

var functionSteps = [                                                                                  
    {                                                                                              
        continue: 'next',                                                                              
        name: 'get_appServer',                                                                    
        verb: 'GET',
        path: '/api/appserver/',                                                            
        ip: null, 
        port: null, 
        type: null,
        cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
    }
];

const getAppServersStepFuncs = {
    get_appServer: function(step, funcStream, ecxSessionId, oFromNg, cb){                                   
      getAppServer(step, funcStream, ecxSessionId, oFromNg, cb);    
    }
};


exports.init = init;


function init(req, res, subType)                                                                       
{
    var ngState = ngp.getSessionState(req, res);
    
    if ( ngState.stale )                                                                                 
      return;                                                                                            

    var funcSteps = JSON.parse(JSON.stringify(functionSteps));
    var genericFuncName = 'get_appServer';

    var ecxSid = ngState.ecxSid;
    res.ecxSid = ecxSid;

    var stepsStream = JSON.parse(JSON.stringify(funcSteps));                                    

    stepsStream.forEach(function(step){
      step.ip = ngp.localIp;
      step.port = ngp.ecxExternal;

      step.func = getAppServersStepFuncs[step.name];     
      
      step.cb = function(ecxSid, data){ 
        ngp.logMsg('info', "Finished step '" + ngp.getCurrentFunc(step.reqId, ecxSid) + '(' + this.name + ")' in session ID '" + ecxSid + "'");

        if ( data )
        {
          ngp.logMsg('debug', "Data returned by step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'=" + JSON.stringify(data));
        }
      };
    });

    ngp.createRequestStore(stepsStream, ecxSid, req.reqId, genericFuncName);                            // add reqId to steps, create store for the req & set currentFunc to genericFuncName
    ngp.setRequestState(req.reqId, ecxSid, 'reqUrl', JSON.stringify(req.url));                                       // used by delete_userrpool steps
    
    ngState.currentFunc = genericFuncName;                                                               
    return stepsStream;
}

function getAppServer(step, funcStream, ecxSessionId, oFromNg, cb)                                       
{
    var reqUrl = ngp.getRequestState(step.reqId, ecxSessionId, 'reqUrl'); 
    
    var fixedReqUrl = reqUrl.replace('ngp','api');
    
    ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, fixedReqUrl.replace(/['"]+/g, ''), ecxSessionId, null, function(appServerInfo){   
        if ( appServerInfo.e || ( appServerInfo.response.statusCode != 200 ) ) {
            ngp.logMsg('warn', 'Failed to get the app server Info because of error=' + JSON.stringify(appServerInfo.rO)); 
            var failResult = ngp.getGenericEcxErrorResponse(appServerInfo, ecxSessionId);  
            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
        } else {
            if(fixedReqUrl.contains('office365')) {
                
                var response = appServerInfo.rO;
                var appServers = appServerInfo.rO['appservers'];
                var finalAppServers = [];
                for(var i=0; i<appServers.length; i++) {
                    if(appServers[i].applicationType != null) 
                    finalAppServers.push(appServers[i]);
                }

                var totalAppServers = finalAppServers.length;
                var appServerCount = 0;

                if(totalAppServers > 0) {
                    getAppServersKeyInfo(step, funcStream, oFromNg, ecxSessionId, finalAppServers, appServerCount, totalAppServers, response, function(finalResponseReturned){   
                        
                        if(!finalResponseReturned.statusCode) {
                            var successResult = {
                                statusCode: appServerInfo.response.statusCode,
                                response: finalResponseReturned   
                            };
                
                            ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);    
                        } else {
                            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, finalResponseReturned, false);
                        }
                        
                        
                    });
                }
                else {
                    var successResult = {
                        statusCode: appServerInfo.response.statusCode,
                        response: appServerInfo.rO  
                    };
                    
                    ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);	
                }
                
            }	
            else {
                var successResult = {
                    statusCode: appServerInfo.response.statusCode,
                    response: appServerInfo.rO  
                };
                
                ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);	
            }      
        }
    });    
}

function getAppServersKeyInfo(step, funcStream, oFromNg, ecxSessionId, appServers, appServerCount, totalAppServers, response, cb) {

    var path = appServers[appServerCount]['appCredentials'][0]['appuser']['href'];

    if (path) {
        ngp.sendRequestToEcx('GET', ngp.ecxExternal, path, ecxSessionId, null, function(keyDetailInfo){   

            if ( keyDetailInfo.e || ( keyDetailInfo.response.statusCode != 200 ) ) {
                ngp.logMsg('warn', 'Failed to get the app server Info because of error=' + JSON.stringify(keyDetailInfo.rO)); 
                  
                var failResult = {
                    statusCode: keyDetailInfo.response.statusCode,
                    response: { rc: (JSON.stringify(keyDetailInfo.rO.description).replace(/\"/g, "")) }
                };

                cb(failResult);
            } else {
                var keyCredentials = {
                    name: keyDetailInfo.rO['name'],
                    keytype: 'exch_key',
                    tenant: keyDetailInfo.rO['tenant'],
                    tenantID: keyDetailInfo.rO['tenantID'],
                    clientID: keyDetailInfo.rO['clientID'],
                    clientSecret: keyDetailInfo.rO['clientSecret'],
                    adminUser: keyDetailInfo.rO['adminUser'],
                    adminUserPass: keyDetailInfo.rO['adminUserPass']
                }
                var instances = [];
                instances.push(keyCredentials);              
            
                response['appservers'][appServerCount].instances = instances;

                var totalAppServersCheck = totalAppServers - 1;
           
                if(appServerCount == totalAppServersCheck)
                cb(response);
                else {
                    appServerCount = appServerCount +1;
                    getAppServersKeyInfo(step, funcStream, oFromNg, ecxSessionId, appServers, appServerCount, totalAppServers, response, cb);
                }
            }
        })
    }
    else {
        var totalAppServersCheck = totalAppServers - 1;

        if(appServerCount == totalAppServersCheck)
        cb(response);
        else {
            appServerCount = appServerCount +1;
            getAppServersKeyInfo(step, funcStream, oFromNg, ecxSessionId, appServers, appServerCount, totalAppServers, response, cb);
        }
    }
}