if (!('contains' in String.prototype)) {
    String.prototype.contains = function (str, startIndex) {
        return ''.indexOf.call(this, str, startIndex) !== -1;
    };
}

if (typeof String.prototype.startsWith != 'function') {
    String.prototype.startsWith = function (prefix) {
        return this.slice(0, prefix.length) == prefix;
    };
}

if (typeof String.prototype.endsWith != 'function') {
    String.prototype.endsWith = function (suffix) {
        return this.slice(-suffix.length) == suffix;
    };
}

var ngp = require('./ecxngp');
var jobs = {}, action = undefined, promiseCounter = 0;

var functionSteps = [
    {
        continue: 'next',
        name: 'get_all_jobs',
        verb: 'GET',
        path: '/api/endeavour/job',
        ip: null,
        port: null,
        type: null,
        cb: function (res, data) { ngp.logMsg('info', 'Step=' + this.name + ' finished'); }
    },
    {
        continue: 'next',
        name: 'hold_or_release_schedules',
        verb: 'POST',
        path: null,
        ip: null,
        port: null,
        type: null,
        cb: function (res, data) { ngp.logMsg('info', 'Step=' + this.name + ' finished'); }
    }
];

const scheduleStepFuncs = {
    get_all_jobs: function (step, funcStream, ecxSessionId, oFromNg, cb) {
        getAllJobs(step, funcStream, ecxSessionId, oFromNg, cb);
    },
    hold_or_release_schedules: function (step, funcStream, ecxSessionId, oFromNg, cb) {
        holdOrReleaseSchedules(step, funcStream, ecxSessionId, oFromNg, cb);
    }
};

exports.init = init;

function init(req, res, scheduleAction) {
    var ngState = ngp.getSessionState(req, res);
    if (ngState.stale) {
        return;
    }
    // Set action to scheduleAction.
    action = scheduleAction;
    var funcSteps = JSON.parse(JSON.stringify(functionSteps));
    var genericFuncName = 'get_all_jobs';
    var ecxSid = ngState.ecxSid;
    res.ecxSid = ecxSid;

    var stepsStream = JSON.parse(JSON.stringify(funcSteps));

    stepsStream.forEach(function (step) {
        step.ip = ngp.localIp;
        step.port = ngp.ecxExternal;
        step.func = scheduleStepFuncs[step.name];
        step.cb = function (ecxSid, data) {
            ngp.logMsg('info', "Finished step '" + ngp.getCurrentFunc(step.reqId, ecxSid) + '(' + this.name + ")' in session ID '" + ecxSid + "'");
            if (data) {
                ngp.logMsg('debug', "Data returned by step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'=" + JSON.stringify(data));
            }
        };
    });

    ngp.createRequestStore(stepsStream, ecxSid, req.reqId, genericFuncName);
    // ngp.setRequestState(req.reqId, ecxSid, 'reqUrl', JSON.stringify(req.url));

    ngState.currentFunc = genericFuncName;
    return stepsStream;
}

function getAllJobs(step, funcStream, ecxSessionId, oFromNg, cb) {
    ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path, ecxSessionId, null, function (response) {
        if (response.e || response.response.statusCode != 200) {
            ngp.logMsg('warn', 'Failed to get jobs. Reason = ' + JSON.stringify(response.rO));
            var failResult = ngp.getGenericEcxErrorResponse(response, ecxSessionId, step);
            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
        } else {
            jobs = response.rO.jobs;
            step.cb(ecxSessionId, null);
            ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
        }
    })
}

function holdOrReleaseSchedules(step, funcStream, ecxSessionId, oFromNg, cb) {
    let hrefs = [], failedJobs = [];
    if (action === 'hold') {
        jobs.forEach(job => {
            if (job.links['pause Schedule']) {
                hrefs.push(
                    {
                        href: job.links['pause Schedule'].href,
                        name: job.name
                    });
            }
        })
    } else if (action === 'release') {
        jobs.forEach(job => {
            if (job.links['release Schedule']) {
                hrefs.push(
                    {
                        href: job.links['release Schedule'].href,
                        name: job.name
                    });
            }
        })
    }
    if (hrefs.length === 0) {
        var successResult = {
            statusCode: 200,
            response: {message: `No job(s) to be ${action === 'hold' ? 'held.' : 'released.'}`},
        };
        ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);
    }
    new Promise((resolve, reject) => {
        promiseCounter = 0;
        for (let index = 0; index < hrefs.length; index++) {
            ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, hrefs[index].href, ecxSessionId, null, function (response) {
                promiseCounter++;
                if (response.e || response.response.statusCode != 200) {
                    failedJobs.push(hrefs[index].name);
                    if (promiseCounter > hrefs.length - 1) {
                        reject(response);
                    }
                } else {
                    if (promiseCounter > hrefs.length - 1) {
                        var successResult = {
                            statusCode: response.response.statusCode,
                            response: { message: `All jobs have been successfully ${action === 'hold' ? 'HELD.' : 'RELEASED.'}` }
                        };
                        ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);
                    }
                }
            })
        }
    }).catch((response) => {
        var failResult = {
            statusCode: response.response.statusCode,
            response: {message: `The following jobs IDs have failed to be ${action === 'hold' ? 'HELD.' : 'RELEASED. - '} ${failedJobs.toString()}`},
        }
        ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
        return;
    })
}