if (!('contains' in String.prototype)) {
	  String.prototype.contains = function(str, startIndex) {
	    return ''.indexOf.call(this, str, startIndex) !== -1;
	  };
	}

if (typeof String.prototype.startsWith != 'function') {
    String.prototype.startsWith = function(prefix) {
        return this.slice(0, prefix.length) == prefix;
    };
}
		 
if (typeof String.prototype.endsWith != 'function') {
    String.prototype.endsWith = function(suffix) {
        return this.slice(-suffix.length) == suffix;
    };
}


var ngp = require('./ecxngp');
var HashMap = require('hashmap');

var funcSteps = [                                                                                  // generalized user handling + provider registration
	{                                                                                              // NB: nothing in this block will change
		continue: 'next',                                                                          // first constituent always continues
		name: 'reg_issprof',
		verb: 'GET',
		path: '/api/spec/storageprofile', 
		ip: null,                                                                                  //!! set by ecxngp. but maybe you don't really need per step ip and ort                                                     
		port: null, 
		type: null,
		func: null,                                                                                // see initialization below 
		cb: function(res, data){ logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              // NB: nothing in this block will change
		continue: 'next',                                                                          // first constituent always continues
		name: 'reg_create',
		verb: 'POST',
		path: '/api/spec/storageprofile', 
		ip: null,                                                                                  //!! set by ecxngp. but maybe you don't really need per step ip and ort                                                     
		port: null, 
		type: null,
		func: null,                                                                               
		cb: function(res, data){ logMsg('info', 'Step=' + this.name + ' finished');}
	}, 	
];


const regSprofStepFuncs = {
	reg_issprof: function(step, funcStream, ecxSessionId, oFromNg, cb){                                // if we're registering sprof then it mustn't exist
		ngp.isSprof(step, funcStream, ecxSessionId, oFromNg, cb); 
	},
	reg_create: function(step, funcStream, ecxSessionId, oFromNg, cb){                                // if we're registering sprof then it mustn't exist
		regCreateSprof(step, funcStream, ecxSessionId, oFromNg, cb); 		
	},	
};


exports.init = init;


function init(req, res, providerType)                                    
{
	var ngState = ngp.getSessionState(req, res);

	if ( ngState.stale )                                                                            // FTI we got bounced while client remained up. AND response already sent.
		return;                                                                                     // !!! R U CERTAIN THAT RESPONSE SENT IN ALL CASES HERE?

	var ecxSid = ngState.ecxSid;
	res.ecxSid = ecxSid;

	var stepsStream = JSON.parse(JSON.stringify(funcSteps));                                    

	stepsStream.forEach(function(step){
		step.ip = ngp.localIp;
		step.port = ngp.ecxExternal;

		step.func = regSprofStepFuncs[step.name];                                                   // hook up the step's real workhorse
		
		step.cb = function(ecxSid, data){ 
			//@#@ ngp.logMsg('info', "Finished step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'");
			ngp.logMsg('info', "Finished step '" + ngp.getCurrentFunc(step.reqId, ecxSid) + '(' + this.name + ")' in session ID '" + ecxSid + "'");

			if ( data )
			{
				ngp.logMsg('debug', "Data returned by step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'=" + JSON.stringify(data));
			}
		};
	});

	ngp.createRequestStore(stepsStream, ecxSid, req.reqId, 'register_storageprofile');  

	ngState.currentFunc = 'register_storageprofile';                                           // as opposed to, say, update_storageprofile

	return stepsStream;
}

// Vide note 116: if the POST body to register OR UPDATE sprof  specifies "offload to Spectrum Protect" but lacks a subpolicy definition for REPLICATION w/software: true, that's an error.

var regCreateSprof = function(step, funcStream, ecxSessionId, oFromNg, cb) { 
	var scrubbedNgInput = ngp.scrubPostBody(oFromNg, true);                      
	ngp.logMsg('warn', "Input from client for storage profile registration in session ID='" + ecxSessionId + "'=" + JSON.stringify(scrubbedNgInput));
	
	const slaName = oFromNg.name;
//	const slaNameAfterReplace = slaName.replace(/[&\/\\#,+()$~%.'":*?<>{}]/g, '');
	const slaNameAfterReplace = slaName.replace(/[,'":.&]/g, '');

	if(slaName.length != slaNameAfterReplace.length) {
		var failResult = {
			statusCode: 400,
			response: { rc: 'Cannot create SLA Policy which contains following characters in it\'s name \n\nDouble Quote(\")\nComma(,)\nApostrophe(\')\nColon(:)\nDot(.)\nAmpersand(&)'}
		};
		ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult,false);	
	} else {
		var sprofTplClone = JSON.parse(JSON.stringify(ngp.sprofTpl));
		ngp.convertReplSubpolicies(sprofTplClone, oFromNg, ecxSessionId);
		sprofTplClone['name'] = (sprofTplClone['name']).trim();

		ngp.ngpCoreHelper.getSites(ecxSessionId, function(getSitesResObj) {
			var arrayOfSites = getSitesResObj.rO['sites'];
			var mapSitesNameId = new HashMap();
			var mapSitesNameLinks = new HashMap();
					
			for (var i = 0; i < arrayOfSites.length; i++) {
				mapSitesNameId.set(arrayOfSites[i]['name'],arrayOfSites[i]['id']);
				mapSitesNameLinks.set(arrayOfSites[i]['name'],arrayOfSites[i].links.self.href);
			}
		
			sprofTplClone.spec.subpolicy.forEach(function(subpolicy) {
		    	if ( subpolicy.type == 'REPLICATION' ) {
					subpolicy.spec.destination.target.href = mapSitesNameLinks.get(subpolicy.spec.destination.target.href);
					subpolicy.spec.destination.target.id = mapSitesNameId.get(subpolicy.spec.destination.target.id);			
		    	}

		    	if ( subpolicy.type == 'BACKUP'  ) {
					subpolicy.spec.destination.target.href = mapSitesNameLinks.get(subpolicy.spec.destination.target.href);
					subpolicy.spec.destination.target.id = mapSitesNameId.get(subpolicy.spec.destination.target.id);			
		    
		    	}	
	   	 	});

	    	ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path, ecxSessionId, sprofTplClone, function(sprofObj) {
				if ( sprofObj.e || sprofObj.response.statusCode != 201 ) {
					var failResult = ngp.getGenericEcxErrorResponse(sprofObj, ecxSessionId, step);			
			    	ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult);	
		    	}
		    	else {
					var successResult = {
						statusCode: sprofObj.response.statusCode,
						response: sprofObj.rO,
					};
			
			    	ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);	
		    	}
	    	});	
   	 	});
	}
};