if (!('contains' in String.prototype)) {
	  String.prototype.contains = function(str, startIndex) {
	    return ''.indexOf.call(this, str, startIndex) !== -1;
	  };
	}

if (typeof String.prototype.startsWith != 'function') {
    String.prototype.startsWith = function(prefix) {
        return this.slice(0, prefix.length) == prefix;
    };
}
		 
if (typeof String.prototype.endsWith != 'function') {
    String.prototype.endsWith = function(suffix) {
        return this.slice(-suffix.length) == suffix;
    };
}

var ngp = require('../ecxngp');
var url = require('url');

const reportTriggerTplFile = ngp.PUBLICDIR + '/triggertpl.json';
var reportTriggerTpl = null;

const reportPolicyTplFile = ngp.PUBLICDIR + '/rptpoltpl.json';
var reportPolicyTpl = null;

const reportJobTplFile = ngp.PUBLICDIR + '/jobtpl.json';
var reportJobTpl = null;

const securityPermissionsTplFile = ngp.PUBLICDIR + '/secpermtpl.json';
var securityPermissionsTpl = null;

var reportScheduleFunctionSteps = [
	{
		continue: 'next',
		name: 'rpt_createschedule',
		verb: null,  //Leaving this as null because of nested callbacks
		path: null,  //Leaving this as null because of nested callbacks
		ip: null,
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
];

var getReportsFunctionSteps = [
	{
		continue: 'next',
		name: 'rpt_getreports',
		verb: null,  //Leaving this as null because of nested callbacks
		path: null,  //Leaving this as null because of nested callbacks
		ip: null,
		port: null, 
        type: null,
        userId: null,
        queryParameters: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
];

var deleteReportScheduleFunctionSteps = [
	{
		continue: 'next',
		name: 'rpt_deletejob',
		verb: null,  //Leaving this as null because of nested callbacks
		path: null,  //Leaving this as null because of nested callbacks
		ip: null,
		port: null, 
        type: null,
        reportId: null,
        userId: null,
        queryParameters: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
    },
	{
		continue: 'next',
		name: 'rpt_deletepolicy',
		verb: null,  //Leaving this as null because of nested callbacks
		path: null,  //Leaving this as null because of nested callbacks
		ip: null,
		port: null, 
        type: null,
        reportId: null,
        userId: null,
        queryParameters: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
    },
	{
		continue: 'next',
		name: 'rpt_deletetrigger',
		verb: null,  //Leaving this as null because of nested callbacks
		path: null,  //Leaving this as null because of nested callbacks
		ip: null,
		port: null, 
        type: null,
        reportId: null,
        userId: null,
        queryParameters: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
    },
	{
		continue: 'next',
		name: 'rpt_deletereport',
		verb: null,  //Leaving this as null because of nested callbacks
		path: null,  //Leaving this as null because of nested callbacks
		ip: null,
		port: null, 
        type: null,
        reportId: null,
        userId: null,
        queryParameters: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
    },           
];

const reportScheduleFunctions = {
    rpt_createschedule: function(step, funcStream, ecxSessionId, oFromNg, cb){
		createReportSchedule(step, funcStream, ecxSessionId, oFromNg, cb);
	},
};

const getReportsFunctions = {
    rpt_getreports: function(step, funcStream, ecxSessionId, oFromNg, cb){
		getReports(step, funcStream, ecxSessionId, oFromNg, cb);
	},
};

const deleteReportScheduleFunctions = {
    rpt_deletejob: function(step, funcStream, ecxSessionId, oFromNg, cb){
		deleteReportScheduleJob(step, funcStream, ecxSessionId, oFromNg, cb);
    },
    rpt_deletepolicy: function(step, funcStream, ecxSessionId, oFromNg, cb){
		deleteReportSchedulePolicy(step, funcStream, ecxSessionId, oFromNg, cb);
    },
    rpt_deletetrigger: function(step, funcStream, ecxSessionId, oFromNg, cb){
		deleteReportScheduleTrigger(step, funcStream, ecxSessionId, oFromNg, cb);
    },
    rpt_deletereport: function(step, funcStream, ecxSessionId, oFromNg, cb){
		deleteReportScheduleReport(step, funcStream, ecxSessionId, oFromNg, cb);
    },    
};

exports.init = init;

getConfiguration();

function getConfiguration()
{
    try
    {
        var rpttriggertemp = ngp.fs.readFileSync(reportTriggerTplFile, 'utf8');
        reportTriggerTpl = JSON.parse(rpttriggertemp);
    }
    catch(e)
    {
        ngp.logMsg('error', "Failed to read report trigger template file because of error=" + e.toString());
        return;
    }

    try
    {
        var rptpolicytemp = ngp.fs.readFileSync(reportPolicyTplFile, 'utf8');
        reportPolicyTpl = JSON.parse(rptpolicytemp);
    }
    catch(e)
    {
        ngp.logMsg('error', "Failed to read report policy template file because of error=" + e.toString());
        return;
    }

    try
    {
        var rptjobtemp = ngp.fs.readFileSync(reportJobTplFile, 'utf8');
        reportJobTpl = JSON.parse(rptjobtemp);
    }
    catch(e)
    {
        ngp.logMsg('error', "Failed to read report job template file because of error=" + e.toString());
        return;
    }

    try
    {
        var securitypermissionstemp = ngp.fs.readFileSync(securityPermissionsTplFile, 'utf8');
        securityPermissionsTpl = JSON.parse(securitypermissionstemp);
    }
    catch(e)
    {
        ngp.logMsg('error', "Failed to read report job template file because of error=" + e.toString());
        return;
    }

}

/**
 * Init function thats sets the steps and its values
 *
 * @param req
 * @param res
 * @param functionType
 */
function init(req, res, functionType)                                    
{
    ngp.logMsg('info', 'init()  ' + JSON.stringify(req.query));
	var ngState = ngp.getSessionState(req, res);

	if ( ngState.stale ) 
		return;

	var ecxSid = ngState.ecxSid;
	res.ecxSid = ecxSid;

    switch (functionType){
        case 'createReportSchedule':
            var stepsStream = JSON.parse(JSON.stringify(reportScheduleFunctionSteps));
            ngp.createRequestStore(stepsStream, ecxSid, req.reqId, 'reports_' + 'schedule');
            ngState.currentFunc = 'reports_' + 'schedule';
            break;
        case 'getReports': 
            var stepsStream = JSON.parse(JSON.stringify(getReportsFunctionSteps));
            ngp.createRequestStore(stepsStream, ecxSid, req.reqId, 'reports_' + 'get');
            ngState.currentFunc = 'reports_' + 'get';
            break;
        case 'deleteReportSchedule': 
            var stepsStream = JSON.parse(JSON.stringify(deleteReportScheduleFunctionSteps));
            ngp.createRequestStore(stepsStream, ecxSid, req.reqId, 'report_' + 'scheduledelete');
            ngState.currentFunc = 'report_' + 'scheduledelete';            
            break;                                
    }

	stepsStream.forEach(function(step){
		step.ip = ngp.localIp;
        step.port = ngp.ecxExternal;

        switch (functionType){
            case 'createReportSchedule':
                step.func = reportScheduleFunctions[step.name];
                break;
            case 'getReports':
                step.queryParameters = ngp.ngpUtils.formatRequestQueryObject(req.query);
                step.userId = req.query['userId'];
                step.func = getReportsFunctions[step.name];
                break;
            case 'deleteReportSchedule':
                step.queryParameters = req.query;
                step.reportId = req.params.id;
                step.userId = req.query['userId'];
                step.func = deleteReportScheduleFunctions[step.name];
                break;
        }
		
		step.cb = function(ecxSid, data){ 
			//@#@ ngp.logMsg('info', "Finished step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'");
			ngp.logMsg('info', "Finished step '" + ngp.getCurrentFunc(step.reqId, ecxSid) + '(' + this.name + ")' in session ID '" + ecxSid + "'");

			if ( data )
			{
				ngp.logMsg('debug', "Data returned by step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'=" + JSON.stringify(data));
			}
		};
	});

	return stepsStream;
}

/**
 * CREATE POST BODY FOR TRIGGER
 *
 * @param reportResource - individual report resource object from the POST body from the UI
 * @param oFromNg - complete POST body from the UI
 */
function createTriggerPostBody(reportResource, oFromNg){
    var reportTriggerTplClone = JSON.parse(JSON.stringify(reportTriggerTpl));
    reportTriggerTplClone.name = 'Report' + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR + reportResource.name;
    reportTriggerTplClone.properties = oFromNg.trigger;
    reportTriggerTplClone.properties.deactivateDate = null;
	return reportTriggerTplClone;
}

/**
 *  CREATE POST BODY FOR POLICY
 *
 * @param reportResource - individual report resource object from the POST body from the UI
 * @param oFromNg - complete POST body from the UI
 */
function createPolicyPostBody(reportResource, oFromNg){
    var reportPolicyTplClone = JSON.parse(JSON.stringify(reportPolicyTpl));
    reportPolicyTplClone.name = 'Report' + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR  + reportResource.name;
    reportPolicyTplClone.spec.source[0].href = reportResource.href;
    reportPolicyTplClone.spec.source[0].name = reportResource.name;
    reportPolicyTplClone.spec.source[0].id = reportResource.id;
    reportPolicyTplClone.spec.notification[0].recipients = oFromNg.notification;
    if (oFromNg.option) {
        reportPolicyTplClone.spec.option = oFromNg.option;    
    }
    return reportPolicyTplClone;
}

/**
 *  CREATE POST BODY FOR JOB
 *
 * @param pjEntry
 */
function createJobPostBody(pjEntry){

    var reportJobTplClone = JSON.parse(JSON.stringify(reportJobTpl));

    reportJobTplClone.name = pjEntry.policyName;
    reportJobTplClone.description += pjEntry.policyName;
    reportJobTplClone.policyId = pjEntry.policyId;
    reportJobTplClone.triggerIds[0] = pjEntry.triggerId;

    return reportJobTplClone;
}

/**
 *  CREATE POST BODY FOR SECURITY PERMISSIONS
 *
 * @param rbacpath
 * @param permissions
 */
function createSecurityPermissionsPostBody(rbacpath, permissions){
    
        var securityPermissionsTplClone = JSON.parse(JSON.stringify(securityPermissionsTpl));
    
        securityPermissionsTplClone.rbacPath = rbacpath;
        securityPermissionsTplClone.permissions = permissions;
    
        return securityPermissionsTplClone;
    }

/**
 * RETURN THE REPORT ID FROM THE URL
 *
 * Notes: Example URL: http://ecxbed3/api/endeavour/report/10
 *
 * @param url
 */
function getReportIdFromUrl(url){
    var arrayOfStrings = url.split('/');
    return arrayOfStrings[arrayOfStrings.length -1];
}

/**
 * GET REPORTS
 *
 * @param step
 * @param funcStream
 * @param ecxSessionId
 * @param oFromNg
 * @param cb
 */
function getReports(step, funcStream, ecxSessionId, oFromNg, cb){
    ngp.addDoubleSpace();
    ngp.logMsg('info', 'Get Report List ==>');

    var crSummaryResults = [];

    ngp.ngpCoreHelper.getTriggers(ecxSessionId, function(getTriggerResObj){
        if (getTriggerResObj.e || getTriggerResObj.response.statusCode != 200) {
            ngp.logMsg('warn', "Get trigger list failed because of error=" + JSON.stringify(getTriggerResObj.rO));
            var failResult = {
                statusCode: 500,
                response: {
                    response: getTriggerResObj.response
                },
            };            

            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
        }else{
            var arrayOfTriggers = getTriggerResObj.rO['triggers'];
            ngp.logMsg('info', 'Report Query Parameters ' + JSON.stringify(step.queryParameters));
            var queryString = ngp.ngpUtils.getQueryString(ngp.ngpUtils.removeEcxngpQueryParameters(step.queryParameters),'begin');
            ngp.logMsg('info', 'Fixed Query Parameters ' + queryString);
            ngp.ngpCoreHelper.getReportList(ecxSessionId, queryString, function(reportListResObj){
                if (reportListResObj.e || reportListResObj.response.statusCode != 200) {
                    ngp.logMsg('warn', "Get report list failed because of error=" + JSON.stringify(reportListResObj.rO));
                    var failResult = {
                        statusCode: 500,
                        response: {
                            response: reportListResObj.rO
                        },
                    };

                    ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
                }else {
                    var reportListResponse = reportListResObj.rO;
                    reportListResponse['reports'].forEach(function(report){
                        //If trigger is found for report, it needs to be added as part of the response
                        var triggerName = 'Report' + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR + report.name;
                        var triggerDisplayName = 'Report' + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR + report.displayName;
                        var oldTriggerName = 'Report' + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR + report.id + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR + step.userId;
                        for (var i = 0; i < arrayOfTriggers.length; i++){
                            if ((arrayOfTriggers[i]['name'] == triggerName) || (arrayOfTriggers[i]['name'] == triggerDisplayName) || (arrayOfTriggers[i]['name'] == oldTriggerName)){
                                report['trigger'] = arrayOfTriggers[i]['properties'];
                                break;
                            }
                        }
                        //Fix links for custom reports
                        if(ngp.ngpUtils.checkNestedPropertyExists(report, 'links.customReports.href')) {
                            var appendedEcxngpQueryString = ngp.ngpUtils.getQueryString(ngp.ngpUtils.getEcxngpQueryParameters(step.queryParameters),'append');
                            var customReportHrefFromBackend = report['links']['customReports']['href'];
                            ngp.logMsg('info', 'Custom Report Url From Backend ' + customReportHrefFromBackend);                            
                            var customReportQueryParamsArrayFromBackend = customReportHrefFromBackend.split("?");
                            if (customReportQueryParamsArrayFromBackend.length == 2){
                                var qParams = require('querystring').parse(customReportQueryParamsArrayFromBackend[1]);
                                if (qParams['filter']){
                                    ngp.escape(qParams['filter']);
                                }
                                var modifiedCustomReportHrefFromBackend = customReportQueryParamsArrayFromBackend[0] + "?" +
                                     require('querystring').stringify(qParams) + appendedEcxngpQueryString;
                                report['links']['customReports']['href'] = ngp.ngpUtils.convertCoreServiceUrlToEcxngpUrl(modifiedCustomReportHrefFromBackend,
                                        ngp.ngpConstants.REPORT_API, ngp.ngpConstants.ECXNGP_REPORT_API);                                
                            }
                        }

                        //Fix delete link for report
                        if(ngp.ngpUtils.checkNestedPropertyExists(report, 'links.delete.href')) {
                            var appendedEcxngpQueryString = ngp.ngpUtils.getQueryString(ngp.ngpUtils.getEcxngpQueryParameters(step.queryParameters),'begin');
                            var deleteReportHref = report['links']['delete']['href'] + appendedEcxngpQueryString;
                            report['links']['delete']['href'] = ngp.ngpUtils.convertCoreServiceUrlToEcxngpUrl(deleteReportHref,
                                    ngp.ngpConstants.REPORT_API, ngp.ngpConstants.ECXNGP_REPORT_API);
                        }
                    });

                    //Notifications and option need to be added to the response
                    //which is got from the asscociated report policy
                    ngp.ngpCoreHelper.getPolicies(ecxSessionId, function(getPoliciesResObj){
                        if (getPoliciesResObj.e || getPoliciesResObj.response.statusCode != 200) {
                            ngp.logMsg('warn', "Get policies list failed because of erro r= " + JSON.stringify(getPoliciesResObj.rO));
                            var failResult = {
                                statusCode: 500,
                                response: {
                                    response: getPoliciesResObj.rO
                                },
                            };

                            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
                        }else{
                            var finishedReportsArray = [];
                            var policiesArrayClone = JSON.parse(JSON.stringify(getPoliciesResObj.rO['policies']));
                            var reportsArrayClone = JSON.parse(JSON.stringify(reportListResponse['reports']));                            
                            addNotificationAndOption(step, funcStream, ecxSessionId, oFromNg, policiesArrayClone, reportsArrayClone, finishedReportsArray, function(finishedReports){

                                delete reportListResponse['reports'];
                                reportListResponse['reports'] = finishedReports;
                                var queryParams = step.queryParameters;

                                crSummaryResults.push(reportListResponse);
                                //@#@ ngp.setEndpointState(ecxSessionId, 'getReportListResults', crSummaryResults);
                                ngp.setRequestState(step.reqId, ecxSessionId, 'getReportListResults', crSummaryResults);

                                ngp.addDoubleSpace();
                                ngp.logMsg('info', '==> Done Getting Report List');

                                var successResult = {
                                    statusCode: 201,
                                    response: crSummaryResults,
                                };

                                ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);
                            });
                        }
                    });
                }
            });
        }
    });
}

/**
 * Returns report array with notification and option fields if found.
 * 
 * @param {*} step 
 * @param {*} funcStream 
 * @param {*} ecxSessionId 
 * @param {*} oFromNg 
 * @param {*} policyArray 
 * @param {*} reportsArray 
 * @param {*} finishedReportsArray 
 * @param {*} cb 
 */
function addNotificationAndOption(step, funcStream, ecxSessionId, oFromNg, policyArray, reportsArray, finishedReportsArray, cb){

    if (reportsArray.length < 1 ){
        ngp.addDoubleSpace();
        ngp.logMsg('info', 'addNotification ==> Done adding notifications');
        ngp.logMsg('info', '    Number of finished reports = ' + finishedReportsArray.length);
        cb(finishedReportsArray);
    }else{
        ngp.addDoubleSpace();
        ngp.logMsg('info', 'addNotification ==> ');
        ngp.logMsg('info', '    Number of reports to process = ' + reportsArray.length);
        ngp.logMsg('info', '    Number of reports processed  = ' + finishedReportsArray.length);
        var report  = reportsArray.shift();
        var oldPolicyName = 'Report' + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR + report.id + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR + step.userId;
        var policyName = 'Report' + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR + report.name;
        var policyDisplayName = 'Report' + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR + report.displayName;
        var policyId = null;
        for (var i = 0; i < policyArray.length; i++){
            if ((policyArray[i]['name'] == policyName) || (policyArray[i]['name'] == oldPolicyName) || (policyArray[i]['name'] == policyDisplayName)){                
                policyId = policyArray[i]['id'];
                ngp.logMsg('info', '    Found Report Policy(' + policyName + ")" + " PolicyId(" + policyId + ")");
                break;
            }
        }

        if (policyId){
            ngp.logMsg('info', '    Processing Report(' + report['id'] + ') Name('+ report['name'] +') Policy(' + policyName + ')' + ' PolicyId(' + policyId + ') for Notifications');
            getNotificationAndOptionFromPolicy(step, funcStream, ecxSessionId, oFromNg, policyName, policyId, function(notificationArray, option){
                if (notificationArray || option){
                    var emptyArray = [];
                    report['notification'] = emptyArray;                    
                    report['notification'] = notificationArray;
                    report['option'] = option;
                    finishedReportsArray.push(report);
                }
                addNotificationAndOption(step, funcStream, ecxSessionId, oFromNg, policyArray, reportsArray, finishedReportsArray, cb);
            });
        }else{
            ngp.logMsg('info', '    Skipping Report(' + report['name'] + ') for Notifications');
            finishedReportsArray.push(report);
            addNotificationAndOption(step, funcStream, ecxSessionId, oFromNg, policyArray, reportsArray, finishedReportsArray, cb);
        }
    }
}

/**
 * Returns notification and option information for a policy
 * 
 * @param {*} step 
 * @param {*} funcStream 
 * @param {*} ecxSessionId 
 * @param {*} oFromNg 
 * @param {*} policyName 
 * @param {*} policyId 
 * @param {*} cb 
 */
function getNotificationAndOptionFromPolicy(step, funcStream, ecxSessionId, oFromNg, policyName, policyId, cb){
    ngp.logMsg('info', 'getNotificationFromPolicy ==>')
    var notificationArray = null;
    ngp.ngpCoreHelper.getPolicy(ecxSessionId, policyId, function(getPolicyResObj){
        if (getPolicyResObj.e || getPolicyResObj.response.statusCode != 200) {
            ngp.logMsg('warn', '    Get policy failed because of error = ' + JSON.stringify(getPolicyResObj.rO));
            var failResult = {
                statusCode: 500,
                response: {
                    response: getPolicyResObj.rO
                },
            };
            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
        }else{
            var getPolicyResponse = getPolicyResObj.rO;
            if (getPolicyResponse['spec']['notification'][0]['recipients']){
                notificationArray = getPolicyResponse['spec']['notification'][0]['recipients'];
                ngp.logMsg('info', '    Found notification for Policy(' + policyName + ') PolicyId(' + policyId + ') ' + JSON.stringify(notificationArray));
            }
            if (getPolicyResponse['spec']['option']) {
                option = getPolicyResponse['spec']['option'];
                ngp.logMsg('info', '    Found option for Policy(' + policyName + ') PolicyId(' + policyId + ') ' + JSON.stringify(option));
            }
            cb(notificationArray, option);
        }
    });
}

/////////////////////////////////////////////////////// START: DELETION OF REPORT SCHEDULE ////////////////////////////////////////////////////////////////////////////

/**
 * Deletes the report schedule Job
 * 
 * @param step 
 * @param funcStream 
 * @param ecxSessionId 
 * @param oFromNg 
 * @param cb 
 */
function deleteReportScheduleJob(step, funcStream, ecxSessionId, oFromNg, cb){
    ngp.addDoubleSpace();
    ngp.logMsg('info', 'Deleting Report Schedule ==> Report Id(' + step.reportId + ')');
    ngp.ngpCoreHelper.getReport(ecxSessionId, step.reportId, function(getReportResObj){
        if (getReportResObj.e || getReportResObj.response.statusCode != 200) {
            ngp.logMsg('error', "Get report failed because of error ==> " + JSON.stringify(getReportResObj.rO));
            var failResult = {
                statusCode: 500,
                response: {
                    response: getReportResObj.rO
                }
            };
            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
        }else {
            var reportJobFound = false;
            var getReportResponse = getReportResObj.rO;
            if (getReportResponse.hasOwnProperty('name')) {
                var report_id = getReportResponse['name'];
                ngp.logMsg('info', 'Report Id(' + report_id + ')');
                ngp.ngpCoreHelper.getJobs(ecxSessionId, function(getJobsResObj){
                    if (getJobsResObj.e || getJobsResObj.response.statusCode != 200) {                        
                        ngp.logMsg('error', "Get jobs error  ==> " + JSON.stringify(getJobsResObj.rO));
                        var failResult = {
                            statusCode: 500,
                            response: {
                                response: getJobsResObj.rO
                            },
                        };
                        ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);                        
                    }else{
                        var reportJobIdToDelete = null;
                        var job_name = 'Report' + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR + report_id;
                        ngp.logMsg('info', 'Report Job Name To Delete(' + job_name + ')');                     
                        var arrayOfJobs = getJobsResObj.rO['jobs'];
                        for (var i = 0; i < arrayOfJobs.length; i++){                            
                            if (arrayOfJobs[i]['name'] == job_name){
                                reportJobIdToDelete = arrayOfJobs[i]['id'];
                                ngp.logMsg('info', 'Found Report Job(' + job_name + ')');
                                ngp.logMsg('info', 'Report Job Id To Delete (' + reportJobIdToDelete + ')');
                                reportJobFound = true;
                                break;
                            }
                        }

                        if (reportJobFound){
                            ngp.ngpCoreHelper.deleteJob(ecxSessionId, reportJobIdToDelete, function(deleteJobResObj){
                                if (deleteJobResObj.e || ( deleteJobResObj.response.statusCode != 204 && deleteJobResObj.response.statusCode != 404 && deleteJobResObj.response.statusCode != 500))
                                {                                        
                                    ngp.logMsg('error', "Delete job failed because of error ==> " + JSON.stringify(deleteJobResObj.rO));
                                    var failResult = {
                                        statusCode: 500,
                                        response: {
                                            response: deleteJobResObj.rO
                                        }
                                    };
                                    ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
                                }else{
                                    //@#@ ngp.setEndpointState(ecxSessionId, 'reportId', report_id);
                                    ngp.setRequestState(step.reqId, ecxSessionId, 'reportId', report_id);
                                    step.cb(ecxSessionId, null);				
                                    ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);                                    
                                }                            
                            });                        
                        }else{
                            //@#@ ngp.setEndpointState(ecxSessionId, 'reportId', report_id);
                            ngp.setRequestState(step.reqId, ecxSessionId, 'reportId', report_id);
                            step.cb(ecxSessionId, null);				
                            ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);                            
                        }
                    }
                });
            }else{
                ngp.logMsg('error', "Report not found error ==> " + JSON.stringify(getReportResObj.rO));
                var failResult = {
                    statusCode: 500,
                    response: {
                        response: getReportResObj.rO
                    },
                };
                ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
            }
        }
    });
}

/**
 * Deletes a report schedule policy
 * 
 * @param step 
 * @param funcStream 
 * @param ecxSessionId 
 * @param oFromNg 
 * @param cb 
 */
function deleteReportSchedulePolicy(step, funcStream, ecxSessionId, oFromNg, cb){
    ngp.ngpCoreHelper.getPolicies(ecxSessionId, function(getPoliciesResObj){
        if (getPoliciesResObj.e || getPoliciesResObj.response.statusCode != 200) {
            ngp.logMsg('error', "Get policies error  ==> " + JSON.stringify(getPoliciesResObj.rO));
            var failResult = {
                statusCode: 500,
                response: {
                    response: getPoliciesResObj.rO
                },
            };
            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);                                
        }else{
            var reportPolicyFound = false;
            var reportPolicyIdToDelete = null;
            //@#@ var policy_name = 'Report' + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR + ngp.getEndpointState(ecxSessionId, 'reportId') + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR + step.userId;
            var policy_name = 'Report' + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR + ngp.getRequestState(step.reqId, ecxSessionId, 'reportId');
            ngp.logMsg('info', 'Report Policy Name To Delete(' + policy_name + ')');
            var arrayOfPolicies = getPoliciesResObj.rO['policies'];
            for (var i = 0; i < arrayOfPolicies.length; i++){
                if (arrayOfPolicies[i]['name'] == policy_name){
                    reportPolicyIdToDelete = arrayOfPolicies[i]['id'];
                    ngp.logMsg('info', 'Found Report Policy(' + policy_name + ')');
                    ngp.logMsg('info', 'Report Policy Id To Delete (' + reportPolicyIdToDelete + ')');
                    reportPolicyFound = true;
                    break;
                }
            }

            if(reportPolicyFound){
                ngp.ngpCoreHelper.deletePolicy(ecxSessionId, reportPolicyIdToDelete, function(deletePolicyResObj){
                    if (deletePolicyResObj.e || ( deletePolicyResObj.response.statusCode != 204 && deletePolicyResObj.response.statusCode != 404 && deletePolicyResObj.response.statusCode != 500)){
                        ngp.logMsg('error', "Delete policy failed because of error ==> " + JSON.stringify(deletePolicyResObj.rO));
                        var failResult = {
                            statusCode: 500,
                            response: {
                                response: deletePolicyResObj.rO
                            }
                        };
                        ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
                    }else{                            
                        step.cb(ecxSessionId, null);
                        ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
                    }
                });
            }else{
                step.cb(ecxSessionId, null);
                ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);                
            }
        }
    });
}

/**
 * Deletes a report schedule trigger
 * 
 * @param step 
 * @param funcStream 
 * @param ecxSessionId 
 * @param oFromNg 
 * @param cb 
 */
function deleteReportScheduleTrigger(step, funcStream, ecxSessionId, oFromNg, cb){
    ngp.ngpCoreHelper.getTriggers(ecxSessionId, function(getTriggersResObj){
        if (getTriggersResObj.e || getTriggersResObj.response.statusCode != 200) {
            ngp.logMsg('error', "Get triggers error  ==> " + JSON.stringify(getTriggersResObj.rO));
            var failResult = {
                statusCode: 500,
                response: {
                    response: getTriggersResObj.rO
                },
            };
            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
        }else{
            var reportTriggerFound = false;
            var reportTriggerIdToDelete = null;
            //@#@ var trigger_name = 'Report' + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR + ngp.getEndpointState(ecxSessionId, 'reportId') + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR + step.userId;
            var trigger_name = 'Report' + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR + ngp.getRequestState(step.reqId, ecxSessionId, 'reportId');
            ngp.logMsg('info', 'Report Trigger Name To Delete(' + trigger_name + ')');
            var arrayOfTriggers = getTriggersResObj.rO['triggers'];
            for (var i = 0; i < arrayOfTriggers.length; i++){
                if (arrayOfTriggers[i]['name'] == trigger_name){
                    reportTriggerIdToDelete = arrayOfTriggers[i]['id']
                    ngp.logMsg('info', 'Found Report Trigger(' + trigger_name + ')');
                    ngp.logMsg('info', 'Report Trigger Id To Delete (' + reportTriggerIdToDelete + ')');
                    reportTriggerFound = true;
                    break;
                }
            }
            if(reportTriggerFound){
                ngp.ngpCoreHelper.deleteTrigger(ecxSessionId, reportTriggerIdToDelete, function(deleteTriggerResObj){
                    if (deleteTriggerResObj.e || ( deleteTriggerResObj.response.statusCode != 204 && deleteTriggerResObj.response.statusCode != 404 && deleteTriggerResObj.response.statusCode != 500)){
                        ngp.logMsg('error', "Delete trigger failed because of error ==> " + JSON.stringify(deleteTriggerResObj.rO));
                        var failResult = {
                                statusCode: 500,
                                response: {
                                    response: deleteTriggerResObj.rO
                                }
                        };
                        ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
                    }else{
                        step.cb(ecxSessionId, null);
                        ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
                    }
                });
            }else{
                step.cb(ecxSessionId, null);
                ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
            }
        }
    });
}

/**
 * Deletes the report schedule report
 * 
 * Note: Last step of report schedule deletion. Sets end point state 
 *  
 * @param step 
 * @param funcStream 
 * @param ecxSessionId 
 * @param oFromNg 
 * @param cb 
 */
function deleteReportScheduleReport(step, funcStream, ecxSessionId, oFromNg, cb){
    ngp.ngpCoreHelper.deleteReport(ecxSessionId, step.reportId, function(deleteReportResObj){
        if (deleteReportResObj.e || ( deleteReportResObj.response.statusCode != 204 && deleteReportResObj.response.statusCode != 404 && deleteReportResObj.response.statusCode != 500)){
            ngp.logMsg('error', "Delete report failed because of error ==> " + JSON.stringify(deleteReportResObj.rO));
            var failResult = {
                statusCode: 500,
                response: {
                    response: deleteReportResObj.rO
                }
            };
            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
        }else{
            ngp.addDoubleSpace();
            ngp.logMsg('info', '==> Done Deleting Report Schedule');
            var successResult = {
                    statusCode: 204,
                    response: deleteReportResObj.rO,
            };
            ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);
        }
    });
}

 /////////////////////////////////////////////////////// END: DELETION OF REPORT SCHEDULE ////////////////////////////////////////////////////////////////////////////////

/**
 * CREATES REPORT SCHEDULE
 * 
 * Function sets up data for the report schedule process
 *
 * @param step
 * @param funcStream
 * @param ecxSessionId
 * @param oFromNg
 * @param cb
 */
function createReportSchedule(step, funcStream, ecxSessionId, oFromNg, cb)
{
    ngp.addDoubleSpace();
    ngp.logMsg('info', 'STARTING CREATION OF REPORT SCHEDULE');
    ngp.addDoubleSpace();

    var bothTriggerNNotificationsEmpty = false;
    var triggerNotEmptyNNotificationsEmpty = false;
    var triggerEmptyNNotificationsNotEmpty = false;
    var bothTriggerNNotificationNotEmpty = false;

    if (ngp.ngpUtils.isObjectEmpty(oFromNg['trigger']) || oFromNg['notification'].length == 0 ){
        if (ngp.ngpUtils.isObjectEmpty(oFromNg['trigger']) && oFromNg['notification'].length == 0 ){
            ngp.logMsg('info', 'Both trigger and notifications are empty');
            bothTriggerNNotificationsEmpty = true;            
        }else if (!ngp.ngpUtils.isObjectEmpty(oFromNg['trigger']) && oFromNg['notification'].length == 0 ){
            ngp.logMsg('info', 'Trigger not empty and notifications is empty');
            triggerNotEmptyNNotificationsEmpty = true;
        }else if (ngp.ngpUtils.isObjectEmpty(oFromNg['trigger']) && oFromNg['notification'].length > 0 ){
            ngp.logMsg('info', 'Trigger is empty and notifications is not empty');
            triggerEmptyNNotificationsNotEmpty = true;
        }
    }else{        
        ngp.logMsg('info', 'Both Trigger and notifications are not empty');
        bothTriggerNNotificationNotEmpty = true;
    }
    
    ngp.addDoubleSpace();

    //HANDLES TWO TYPES OF REQUESTS
    // 1. TRIGGER IS EMPTY; NOTFICATIONS IS EMPTY
    // 2. TRIGGER IS EMPTY; NOTFICATIONS IS NOT EMPTY
    if (bothTriggerNNotificationsEmpty || triggerEmptyNNotificationsNotEmpty)  {        
        ngp.addDoubleSpace();
        var arrayOfPolicies = null;
        var arrayOfJobs = null;
        var arrayOfTriggers = null;
        ngp.ngpCoreHelper.getPolicies(ecxSessionId, function(getPoliciesResObj){
            if (getPoliciesResObj.e || getPoliciesResObj.response.statusCode != 200) {
                ngp.logMsg('error', "   Get policies error  ==> " + JSON.stringify(getPoliciesResObj.rO));
                var failResult = ngp.getGenericEcxErrorResponse(getPoliciesResObj, ecxSessionId, step);
                cb(failResult);
            }else{
                arrayOfPolicies = getPoliciesResObj.rO['policies'];
                ngp.ngpCoreHelper.getJobs(ecxSessionId, function(getJobsResObj){
                    if (getJobsResObj.e || getJobsResObj.response.statusCode != 200) {                        
                        ngp.logMsg('error', "Get jobs error  ==> " + JSON.stringify(getJobsResObj.rO));
                        var failResult = ngp.getGenericEcxErrorResponse(getJobsResObj, ecxSessionId, step);
                        cb(failResult);                      
                    }else{
                        arrayOfJobs = getJobsResObj.rO['jobs'];
                        ngp.ngpCoreHelper.getTriggers(ecxSessionId, function(getTriggersResObj){
                            if (getTriggersResObj.e || getTriggersResObj.response.statusCode != 200) { 
                                ngp.logMsg('error', "Get triggers error  ==> " + JSON.stringify(getTriggersResObj.rO));
                                var failResult = ngp.getGenericEcxErrorResponse(getTriggersResObj, ecxSessionId, step);
                                cb(failResult); 
                            }else{
                                arrayOfTriggers = getTriggersResObj.rO['triggers'];                                
                                var crSummaryResults = [];
                                if (bothTriggerNNotificationsEmpty){
                                    handleEmptyRptScheduleRequest(step, funcStream, ecxSessionId, oFromNg, arrayOfPolicies, arrayOfJobs, arrayOfTriggers, oFromNg['userId'], crSummaryResults, function(failResult){
                                        if ( failResult )
                                        {
                                            ngp.logMsg('error', 'Failure in handling empty report schedule request');
                                            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
                                        }else{
                                            ngp.addDoubleSpace();
                                            ngp.logMsg('info', '==> Done handling empty report schedule request');    
                                            //@#@ ngp.setEndpointState(ecxSessionId, 'reportScheduleCreationResults', crSummaryResults);
                                            //@#@ ngp.setEndpointState(ecxSessionId, 'finishedPjEntries', finishedPjEntries);
                                            
                                            ngp.setRequestState(step.reqId, ecxSessionId, 'reportScheduleCreationResults', crSummaryResults);
                                            var successResult = {
                                                statusCode: 201,
                                                response: crSummaryResults,
                                            };

                                            ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);	
                                        }
                                    });
                                }else if (triggerEmptyNNotificationsNotEmpty){
                                    //FOR NOTFICATIONS NEED TO CHECK IF SMTP SERVER IS CONFIGURED
                                    ngp.ngpCoreHelper.getSmtps(ecxSessionId, function(getSmtpsResObj){
                                        if ( getSmtpsResObj.e || getSmtpsResObj.response.statusCode != 200 )
                                        {
                                                ngp.logMsg('error', "   Get smtps error  ==> " + JSON.stringify(getSmtpsResObj.rO));
                                                var failResult = {
                                                    statusCode: 500,                                  
                                                    response: getSmtpsResObj.rO,
                                                };

                                                ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
                                        }else{
                                            var smtpHref = null;
                                            var arrayOfSmtps = getSmtpsResObj.rO['smtps'];
                                            if (arrayOfSmtps.length > 0){
                                                smtpHref = arrayOfSmtps[0]['links']['self']['href'];
                                                ngp.logMsg('info', 'SMTP server href(' + smtpHref + ')');
                                                ngp.setRequestState(step.reqId, ecxSessionId, 'smtpHref', smtpHref);
                                                //**NOTE: HANDLES ONLY ONE REPORT AT THIS TIME**
                                                //FIRST CHECK IF REPORT POLICY IS CREATED, IF YES UPDATE NOTFICATION ARRAY. IF NOT CREATE REPORT POLICY WITH NOTFICATION ARRAY
                                                //SECOND CHECK IF REPORT JOB IS CREATED, IF YES UPDATE TRIGGERIDS TO EMPTY. IF NOT CREATE REPORT JOB WITH EMPTY TRIGGERIDS
                                                //THIRD DELETE THE TRIGGER
                                                var resourceName = 'Report' + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR  + oFromNg['resources'][0]['name'];
                                                var resourceDisplayName = 'Report' + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR  + oFromNg['resources'][0]['displayName'];
                                                var resourceNameWithId = 'Report' + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR  + oFromNg['resources'][0]['id'] + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR  + oFromNg['userId'];
                                                var reportPolicyFound = false;
                                                var policyName = null;
                                                var policyId = null;
                                                var policyPostBody = null;
                                                for (var i = 0; i < arrayOfPolicies.length; i++){
                                                    if ((arrayOfPolicies[i]['name'] == resourceName) || (arrayOfPolicies[i]['name'] == resourceDisplayName) || (arrayOfPolicies[i]['name'] == resourceNameWithId)){
                                                        reportPolicyFound = true;
                                                        policyName = arrayOfPolicies[i]['name'];                                           
                                                        policyId = arrayOfPolicies[i]['id'];
                                                        policyPostBody = createPolicyPostBody(oFromNg['resources'][0], oFromNg);
                                                        policyPostBody.spec.notification[0].href = smtpHref;
                                                        break;
                                                    }
                                                }
                                                if (reportPolicyFound){                                        
                                                    ngp.ngpCoreHelper.updatePolicy(ecxSessionId, policyId, policyPostBody, function (policyUpdateResObj){
                                                        if ( policyUpdateResObj.e || policyUpdateResObj.response.statusCode != 200)
                                                        {
                                                            var failResult = ngp.getGenericEcxErrorResponse(policyUpdateResObj, ecxSessionId, step);
                                                            cb(failResult);
                                                        }else{
                                                            crSummaryResults.push(policyUpdateResObj.rO);
                                                            handleEmptyTriggerNonEmptyNotifications(step, funcStream, ecxSessionId, oFromNg, resourceName, resourceDisplayName, resourceNameWithId,  arrayOfJobs, 
                                                                arrayOfTriggers, policyId, policyName, crSummaryResults, function(failResult){
                                                                if ( failResult )
                                                                {
                                                                    ngp.logMsg('error', 'Failure in handling empty trigger and non-empty notfications schedule request');
                                                                    ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
                                                                }else{
                                                                    ngp.addDoubleSpace();
                                                                    ngp.logMsg('info', '==> Done handling empty trigger and non-empty notfications schedule request');    
                                                                    //@#@ ngp.setEndpointState(ecxSessionId, 'reportScheduleCreationResults', crSummaryResults);
                                                                    //@#@ ngp.setEndpointState(ecxSessionId, 'finishedPjEntries', finishedPjEntries);
                                                                    
                                                                    ngp.setRequestState(step.reqId, ecxSessionId, 'reportScheduleCreationResults', crSummaryResults);
                                                                    var successResult = {
                                                                        statusCode: 201,
                                                                        response: crSummaryResults,
                                                                    };

                                                                    ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);	
                                                                }
                                                            });
                                                        }
                                                    });
                                                }else{
                                                    policyPostBody = createPolicyPostBody(oFromNg['resources'][0], oFromNg);
                                                    policyPostBody.spec.notification[0].href = smtpHref;
                                                    ngp.ngpCoreHelper.createPolicy(ecxSessionId, policyPostBody, function (policyCreateResObj){
                                                        if ( policyCreateResObj.e || policyCreateResObj.response.statusCode != 200)
                                                        {
                                                            var failResult = ngp.getGenericEcxErrorResponse(policyCreateResObj, ecxSessionId, step);
                                                            cb(failResult);
                                                        }else{
                                                            crSummaryResults.push(policyCreateResObj.rO);
                                                            handleEmptyTriggerNonEmptyNotifications(step, funcStream, ecxSessionId, oFromNg, resourceName, resourceDisplayName, resourceNameWithId, arrayOfJobs, 
                                                                arrayOfTriggers, policyCreateResObj.rO.id, policyCreateResObj.rO.name, crSummaryResults, function(failResult){
                                                                if ( failResult )
                                                                {
                                                                    ngp.logMsg('error', 'Failure in handling empty trigger and non-empty notfications schedule request');
                                                                    ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
                                                                }else{
                                                                    ngp.addDoubleSpace();
                                                                    ngp.logMsg('info', '==> Done handling empty trigger and non-empty notfications schedule request');    
                                                                    //@#@ ngp.setEndpointState(ecxSessionId, 'reportScheduleCreationResults', crSummaryResults);
                                                                    //@#@ ngp.setEndpointState(ecxSessionId, 'finishedPjEntries', finishedPjEntries);
                                                                    
                                                                    ngp.setRequestState(step.reqId, ecxSessionId, 'reportScheduleCreationResults', crSummaryResults);
                                                                    var successResult = {
                                                                        statusCode: 201,
                                                                        response: crSummaryResults,
                                                                    };

                                                                    ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);
                                                                }
                                                            });
                                                        }
                                                    });
                                                }                                          
                                            }else{
                                                //Return error back to UI as a bad request if SMTP servers are not defined
                                                ngp.logMsg('error', "   Notifications are set but SMTP servers are not defined");
                                                var failResult = {
                                                    statusCode: 400,
                                                    response: {rc: 'SMTP server needs to be configured for email notifications.'},
                                                };

                                                ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
                                            }
                                        }
                                    });
                                }
                            }
                        });  
                    }
                });
            }
        });
     //OTHER REQUESTS TYPES ARE HANDLED HERE
    }else{        
        var pjEntries = [];                                                              

        oFromNg.resources.forEach(function(resource){

            var triggerPostBody = createTriggerPostBody(resource, oFromNg);
            var policyPostBody = createPolicyPostBody(resource, oFromNg);

            var pjEntry = {

                //Entries for trigger
                triggerPostBody: triggerPostBody,
                triggerCreated: false,
                triggerId: null,   //Populated downstream
                triggerName: null,

                //Entries for policy
                policyPostBody: policyPostBody,
                policyCreated: false,
                policyId: null,   //Populated downstream
                policyName: null,

                //Entries for job
                jobPostBody: null,  ////Populated downstream
                jobCreated: false,
                jobId: null,
                jobName: null,

                //Entries related to rollback
                triggerRollbackAttempted: false,
                triggerRollbackSuccessful: false,
                policyRollbackAttempted: false,
                policyRollbackSuccessful: false,
                jobRollbackAttempted: false,
                jobRollbackSuccessful: false,
                
                reportRbacPath: resource['rbacPath'],
            };
            
            pjEntries.push(pjEntry);                                                                     
        });

        // IF NOTIFICATIONS ARE SET IN THE REQUEST WE NEED TO CHECK IF SMTP SERVER IS CONFIGURED IN THE ENTERPRISE
        if (oFromNg['notification']){
            if (Array.isArray(oFromNg['notification']) && oFromNg['notification'].length > 0){
                ngp.logMsg('info', 'createReportSchedule ==> notifications need to be set');
                ngp.ngpCoreHelper.getSmtps(ecxSessionId, function(getSmtpsResObj){
                    if ( getSmtpsResObj.e || getSmtpsResObj.response.statusCode != 200 )
                    {
                            ngp.logMsg('error', "   Get smtps error  ==> " + JSON.stringify(getSmtpsResObj.rO));
                            var failResult = {
                                statusCode: 500,                                  
                                response: getSmtpsResObj.rO,
                            };

                            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
                    }else{
                        var smtpHref = null;
                        var arrayOfSmtps = getSmtpsResObj.rO['smtps'];
                        if (arrayOfSmtps.length > 0){
                            smtpHref = arrayOfSmtps[0]['links']['self']['href'];
                            ngp.logMsg('info', 'SMTP server href(' + smtpHref + ')');
                            pjEntries.forEach(function(pjEntry){
                                pjEntry.policyPostBody.spec.notification[0].href = smtpHref;
                            });
                            ngp.setRequestState(step.reqId, ecxSessionId, 'smtpHref', smtpHref);
                            runReportSchedule(step, funcStream, ecxSessionId, oFromNg, cb, pjEntries)
                        }else{
                            //Return error back to UI as a bad request if SMTP servers are not defined
                            ngp.logMsg('error', "   Notifications are set but SMTP servers are not defined");
                            var failResult = {
                                statusCode: 400,
                                response: {rc: 'SMTP server needs to be configured for email notifications.'},
                            };

                            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
                        }
                    }
                });
            }else{
                ngp.logMsg('info', 'createReportSchedule ==> notification property is empty');
                runReportSchedule(step, funcStream, ecxSessionId, oFromNg, cb, pjEntries);            
            }
        }else{
            ngp.logMsg('info', 'createReportSchedule ==> notification property not found in the request');
            runReportSchedule(step, funcStream, ecxSessionId, oFromNg, cb, pjEntries);
        }
    }
}

/**
 * 
 * @param step 
 * @param funcStream 
 * @param ecxSessionId 
 * @param oFromNg 
 * @param arrayOfJobs 
 * @param policyId 
 * @param cb 
 */
function handleEmptyTriggerNonEmptyNotifications(step, funcStream, ecxSessionId, oFromNg, resourceName, resourceDisplayName, resourceNameWithId, arrayOfJobs, arrayOfTriggers, 
    policyId, policyName, crSummaryResults, hdleEmptyTrigNEmptyNot){

    var reportJobFound = true;
    var jobId = true;
    var jobPostBody = null;
    var emptyArray = [];

    var reportTriggerFound = true;
    var triggerId = null;

    for (var i = 0; i < arrayOfJobs.length; i++){
        if ((arrayOfJobs[i]['name'] == resourceName) || (arrayOfJobs[i]['name'] == resourceDisplayName) || (arrayOfJobs[i]['name'] == resourceNameWithId)){
            reportJobFound = true;
            jobId = arrayOfJobs[i]['id'];
            jobPostBody = JSON.parse(JSON.stringify(reportJobTpl));
            jobPostBody.name = policyName;
            jobPostBody.description += policyName;
            jobPostBody.policyId = policyId;
            jobPostBody.triggerIds = emptyArray;
            break;
        }
    }

    for (var i = 0; i < arrayOfTriggers.length; i++){
        if ((arrayOfTriggers[i]['name'] == resourceName) || (arrayOfTriggers[i]['name'] == resourceDisplayName) || (arrayOfTriggers[i]['name'] == resourceNameWithId)){
            reportTriggerFound = true;
            triggerId = arrayOfTriggers[i]['id'];
            break;
        }
    }    
    
    if (reportJobFound){
        ngp.ngpCoreHelper.updateJob(ecxSessionId, jobId, jobPostBody, function (jobUpdateResObj){
            if ( jobUpdateResObj.e || jobUpdateResObj.response.statusCode != 200)
            {
                var failResult = ngp.getGenericEcxErrorResponse(jobUpdateResObj, ecxSessionId, step);
                hdleEmptyTrigNEmptyNot(failResult);                                
            }else{
                crSummaryResults.push(jobUpdateResObj.rO);
                if (reportTriggerFound){
                    //DELETE TRIGGER IN THIS USECASE
                    ngp.ngpCoreHelper.deleteTrigger(ecxSessionId, triggerId, function (trigDelResObj){
                        if (trigDelResObj.e || ( trigDelResObj.response.statusCode != 204 && trigDelResObj.response.statusCode != 404 && 
                            trigDelResObj.response.statusCode != 500 )) {
                            var failResult = ngp.getGenericEcxErrorResponse(trigDelResObj, ecxSessionId, step);
                            hdleEmptyTrigNEmptyNot(failResult);
                        }else{
                            crSummaryResults.push(trigDelResObj.rO);
                            hdleEmptyTrigNEmptyNot();
                        }
                    });
                }else{           
                    hdleEmptyTrigNEmptyNot();
                }                
            }
        });
    }else{        
        jobPostBody = JSON.parse(JSON.stringify(reportJobTpl));        
        jobPostBody.name = policyName;
        jobPostBody.description += policyName;
        jobPostBody.policyId = policyId;
        jobPostBody.triggerIds = emptyArray;
        ngp.ngpCoreHelper.createJob(ecxSessionId, jobPostBody, function (jobCreateResObj){
            if ( jobCreateResObj.e || jobCreateResObj.response.statusCode != 200)
            {
                var failResult = ngp.getGenericEcxErrorResponse(jobCreateResObj, ecxSessionId, step);
                hdleEmptyTrigNEmptyNot(failResult);                                
            }else{
                crSummaryResults.push(jobCreateResObj.rO);   
                hdleEmptyTrigNEmptyNot();
            }
        });
    }
}

/**
 * FUNCTION HANDLES REQUESTS WHERE TRIGGER AND NOTFICATIONS ARE EMPTY
 * 
 * @param step 
 * @param funcStream 
 * @param ecxSessionId 
 * @param oFromNg 
 * @param arrayOfPolicies 
 * @param arrayOfJobs 
 * @param arrayOfTriggers 
 * @param userId 
 * @param crSummaryResults 
 * @param callbackEmptyTriggerCb 
 */
function handleEmptyRptScheduleRequest(step, funcStream, ecxSessionId, oFromNg, arrayOfPolicies, arrayOfJobs, arrayOfTriggers, userId, crSummaryResults, emptyrptScheduleCb){
    
    var newoFromNg = JSON.parse(JSON.stringify(oFromNg));

    if ( newoFromNg['resources'].length < 1 )
	{
        emptyrptScheduleCb();
	}
	else
	{
        var policyPostBody = null;
        var reportPolicyFound = false;
        var policyId = null;
        var policyName = null;

        var jobPostBody = null;
        var reportJobFound = false;
        var jobId = null;

        var reportTriggerFound = false;
        var triggerId = null;
        var reportResource = newoFromNg['resources'].shift();

        var resourceName = 'Report' + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR  + reportResource.name;
        var resourceDisplayName = 'Report' + ngp.ngpConstants.REPORT_ASSOCIATION_SEPERATOR  + reportResource.displayName;

        for (var i = 0; i < arrayOfPolicies.length; i++){
            if ((arrayOfPolicies[i]['name'] == resourceName) || (arrayOfPolicies[i]['name'] == resourceDisplayName)) {
                reportPolicyFound = true;
                policyName = arrayOfPolicies[i]['name'];
                policyId = arrayOfPolicies[i]['id'];
                policyPostBody = createPolicyPostBody(reportResource, oFromNg);
                break;
            }
        }        

        for (var i = 0; i < arrayOfTriggers.length; i++){
            if ((arrayOfTriggers[i]['name'] == resourceName) || (arrayOfTriggers[i]['name'] == resourceDisplayName)) {
                reportTriggerFound = true;                
                triggerId = arrayOfTriggers[i]['id'];
                break;
            }
        }         

        for (var i = 0; i < arrayOfJobs.length; i++){
            if ((arrayOfJobs[i]['name'] == resourceName) || (arrayOfJobs[i]['name'] == resourceDisplayName)) {
                reportJobFound = true;
                jobId = arrayOfJobs[i]['id'];
                jobPostBody = JSON.parse(JSON.stringify(reportJobTpl));
                var emptyArray = [];
                jobPostBody.name = policyName;
                jobPostBody.description += policyName;
                jobPostBody.policyId = policyId;
                jobPostBody.triggerIds = emptyArray;
                break;
            }
        }
       
        if(reportPolicyFound){
            //UPDATE POLICY WITH EMPTY NOTFICATIONS ARRAY
            ngp.ngpCoreHelper.updatePolicy(ecxSessionId, policyId, policyPostBody, function (policyUpdateResObj){
                if ( policyUpdateResObj.e || policyUpdateResObj.response.statusCode != 200)
                {
                    var failResult = ngp.getGenericEcxErrorResponse(policyUpdateResObj, ecxSessionId, step);
                    emptyrptScheduleCb(failResult);
                }else{
                    crSummaryResults.push(policyUpdateResObj.rO);
                    if(reportJobFound){
                        //UPDATE JOB WITH EMPTY TRIGGERS
                        ngp.ngpCoreHelper.updateJob(ecxSessionId, jobId, jobPostBody, function (jobUpdateResObj){
                            if ( jobUpdateResObj.e || jobUpdateResObj.response.statusCode != 200)
                            {                                
                                var failResult = ngp.getGenericEcxErrorResponse(jobUpdateResObj, ecxSessionId, step);
                                emptyrptScheduleCb(failResult);
                            }else{                             
                                crSummaryResults.push(jobUpdateResObj.rO);
                                if (reportTriggerFound){
                                    //DELETE TRIGGER IN THIS USECASE
                                    ngp.ngpCoreHelper.deleteTrigger(ecxSessionId, triggerId, function (trigDelResObj){
                                        if (trigDelResObj.e || ( trigDelResObj.response.statusCode != 204 && trigDelResObj.response.statusCode != 404 && trigDelResObj.response.statusCode != 500 )) {
                                            var failResult = ngp.getGenericEcxErrorResponse(trigDelResObj, ecxSessionId, step);
                                            emptyrptScheduleCb(failResult);                                              
                                        }else{
                                            crSummaryResults.push(trigDelResObj.rO);
                                            handleEmptyRptScheduleRequest(step, funcStream, ecxSessionId, newoFromNg, arrayOfPolicies, arrayOfJobs, arrayOfTriggers, oFromNg['userId'], crSummaryResults, emptyrptScheduleCb);
                                        }
                                    });
                                }else{                                  
                                    handleEmptyRptScheduleRequest(step, funcStream, ecxSessionId, newoFromNg, arrayOfPolicies, arrayOfJobs, arrayOfTriggers, oFromNg['userId'], crSummaryResults, emptyrptScheduleCb);
                                }
                            }
                        });
                    }else{
                         handleEmptyRptScheduleRequest(step, funcStream, ecxSessionId, newoFromNg, arrayOfPolicies, arrayOfJobs, arrayOfTriggers, oFromNg['userId'], crSummaryResults, emptyrptScheduleCb);
                    }                     
                }
            });
        }else{
            handleEmptyRptScheduleRequest(step, funcStream, ecxSessionId, newoFromNg, arrayOfPolicies, arrayOfJobs, arrayOfTriggers, oFromNg['userId'], crSummaryResults, emptyrptScheduleCb);
        }
    }
}

/**
 * 
 * @param step 
 * @param funcStream 
 * @param ecxSessionId 
 * @param oFromNg 
 * @param cb 
 * @param pjEntries 
 */
function runReportSchedule(step, funcStream, ecxSessionId, oFromNg, cb, pjEntries){

	var numberOfReportsToSchedule = pjEntries.length;
    ngp.addDoubleSpace();   
    ngp.logMsg('info', 'runReportSchedule ==> Number of Reports to Schedule: ' + numberOfReportsToSchedule);

	var finishedPjEntries = [];
	var crSummaryResults = [];

    /*
       At any point in time pjEntries will contain unfinished (not yet presented to ECX pjtEntries each of which represents a policy/job/trigger triple) and finishedPjEntries
       will contain finished. The entities represented by an element in finishedPjEntries may or may not have been created. Stuff remaining in pjentries never got created and
       thus can't be rolled back.
    */
    createTriggerAndPolicyAndJob(pjEntries, step, ecxSessionId, finishedPjEntries, crSummaryResults, function(failResult){
		if ( failResult )
		{
			ngp.logMsg('error', 'Failure in scheduling REPORT in session ' + ecxSessionId + '; attempting rollback');

            var rolledBackEntriesArray = [];
 
			rollbackReportSchedule(finishedPjEntries, ecxSessionId, rolledBackEntriesArray, failResult, function(failResult){
 
                var failResult = {
                    statusCode: failResult.statusCode,
                    response: {rc: failResult.response.description}                                                                                                                             
                };
 
                ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
			});
		}
		else
		{
            ngp.addDoubleSpace();
            ngp.logMsg('info', '==> Done Creating Report Schedule');    
   		    //@#@ ngp.setEndpointState(ecxSessionId, 'reportScheduleCreationResults', crSummaryResults);
            //@#@ ngp.setEndpointState(ecxSessionId, 'finishedPjEntries', finishedPjEntries);
            
            ngp.setRequestState(step.reqId, ecxSessionId, 'reportScheduleCreationResults', crSummaryResults);
			ngp.setRequestState(step.reqId, ecxSessionId, 'finishedPjEntries', finishedPjEntries);

			var successResult = {
				statusCode: 201,
				response: crSummaryResults,
			};
	
			ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);	
		}
	});
}

/**
 * CREATES TRIGGER, POLICY AND JOB FOR REPORT
 *
 * NOTES: createPolicyAndJob only returns after all entries in pjEntries (each entry represents a policy/job pair that must be created) have been processed; or after a failure
 *
 * @param pjEntries
 * @param driverStep
 * @param ecxSessionId
 * @param finishedPjEntries
 * @param crSummaryResults
 * @param reportScheduleCreatedCb
 */
function createTriggerAndPolicyAndJob(pjEntries, driverStep, ecxSessionId, finishedPjEntries, crSummaryResults, reportScheduleCreatedCb)    // RECURSIVE function
{
	if ( pjEntries.length < 1 )
	{
        reportScheduleCreatedCb();
	}
	else
	{
		var pjEntry = pjEntries.shift();

        finishedPjEntries.push(pjEntry);
        
        ngp.addDoubleSpace();
        ngp.logMsg('info', 'Creating Trigger for Report ==> ');
        ngp.ngpCoreHelper.createTrigger(ecxSessionId, pjEntry.triggerPostBody, function(trigCreateResObj){
            if ( trigCreateResObj.e || trigCreateResObj.response.statusCode != 201 )
            {
                if ( trigCreateResObj.response.statusCode == 409 && trigCreateResObj.rO.id == "XSBDuplicateRecordException"){
                    //IN THIS CASE REPORT POLICY AND REPORT JOB SHOULD EXIST AS WELL. THERE IS NO NEED TO UPDATE REPORT JOB.
                    //UPDATE REPORT POLICY WITH NOTIFICATION RECEIVED IN THE REQUEST AND THEN EXIT
                    ngp.logMsg('error', "    Trigger resource already exists for Report, updating trigger ==> " + JSON.stringify(trigCreateResObj.rO));
                    ngp.ngpCoreHelper.getTriggers(ecxSessionId, function(getTriggersResObj){
                        if (getTriggersResObj.e || getTriggersResObj.response.statusCode != 200) {
                            ngp.logMsg('error', "   Get triggers error  ==> " + JSON.stringify(getTriggersResObj.rO));
                            var failResult = ngp.getGenericEcxErrorResponse(getTriggersResObj, ecxSessionId, driverStep);
                            reportScheduleCreatedCb(failResult);
                        }else{
                            var triggerIdToUpdate = null;
                            var reportTriggerFound = false;
                            var trigger_name = pjEntry.triggerPostBody.name;
                            ngp.logMsg('info', '    Report Trigger Name To Update(' + trigger_name + ')');
                            var arrayOfTriggers = getTriggersResObj.rO['triggers'];
                            for (var i = 0; i < arrayOfTriggers.length; i++){
                                if (arrayOfTriggers[i]['name'] == trigger_name){
                                    triggerIdToUpdate = arrayOfTriggers[i]['id'];
                                    ngp.logMsg('info', '    Found Report Trigger(' + trigger_name + ')');
                                    ngp.logMsg('info', '    Trigger Id to Update(' + triggerIdToUpdate + ')');
                                    reportTriggerFound = true;
                                    break;
                                }
                            }

                            if(reportTriggerFound){
                                ngp.ngpCoreHelper.updateTrigger(ecxSessionId, triggerIdToUpdate, pjEntry.triggerPostBody, function(trigUpdateResObj){
                                    if ( trigUpdateResObj.e || trigUpdateResObj.response.statusCode != 200)
                                    {
                                        var failResult = ngp.getGenericEcxErrorResponse(trigUpdateResObj, ecxSessionId, driverStep);
                                        reportScheduleCreatedCb(failResult);
                                    }else{
                                        crSummaryResults.push(trigUpdateResObj.rO);                                        
                                        ngp.ngpCoreHelper.getPolicies(ecxSessionId, function(getPoliciesResObj){
                                            if (getPoliciesResObj.e || getPoliciesResObj.response.statusCode != 200) {
                                                ngp.logMsg('error', "   Get policies error  ==> " + JSON.stringify(getPoliciesResObj.rO));
                                                var failResult = ngp.getGenericEcxErrorResponse(getPoliciesResObj, ecxSessionId, driverStep);
                                                reportScheduleCreatedCb(failResult);
                                            }else{
                                                var policyIdToUpdate = null;
                                                var reportPolicyFound = false;
                                                var policy_name = pjEntry.policyPostBody.name;
                                                var arrayOfPolicies = getPoliciesResObj.rO['policies'];
                                                for (var i = 0; i < arrayOfPolicies.length; i++){
                                                    if (arrayOfPolicies[i]['name'] == policy_name){
                                                        reportPolicyFound = true;
                                                        policyIdToUpdate = arrayOfPolicies[i]['id'];
                                                        ngp.logMsg('info', '    Found Report Policy(' + policy_name + ')');
                                                        ngp.logMsg('info', '    Policy Id to Update(' + policyIdToUpdate + ')');
                                                        break;
                                                    }
                                                }

                                                if(reportPolicyFound){
                                                    ngp.ngpCoreHelper.updatePolicy(ecxSessionId, policyIdToUpdate, pjEntry.policyPostBody, function(policyUpdateResObj){
                                                        if ( policyUpdateResObj.e || policyUpdateResObj.response.statusCode != 200)
                                                        {
                                                            var failResult = ngp.getGenericEcxErrorResponse(policyUpdateResObj, ecxSessionId, driverStep);
                                                            reportScheduleCreatedCb(failResult);
                                                        }else{
                                                            crSummaryResults.push(policyUpdateResObj.rO);
                                                            reportScheduleCreatedCb();
                                                        }
                                                    });
                                                }else{
                                                    reportScheduleCreatedCb();
                                                }
                                            }
                                        });
                                    }
                                });
                            }else{
                                reportScheduleCreatedCb();
                            }
                        }
                    });
                }else{
                    var failResult = ngp.getGenericEcxErrorResponse(trigCreateResObj, ecxSessionId, driverStep);
                    reportScheduleCreatedCb(failResult);
                }
            }else{
                crSummaryResults.push(trigCreateResObj.rO);
                pjEntry.triggerId = trigCreateResObj.rO.id;
                pjEntry.triggerCreated = true;
                pjEntry.triggerName = trigCreateResObj.rO.name;

                ngp.logMsg('info', "    Trigger creation response ==> " + JSON.stringify(trigCreateResObj.rO));
                ngp.addDoubleSpace();
                ngp.logMsg('info', 'Creating Policy for Report ==> ');
                ngp.ngpCoreHelper.createPolicy(ecxSessionId, pjEntry.policyPostBody, function(policyCreateResObj){
                    if ( policyCreateResObj.e || policyCreateResObj.response.statusCode != 201 )
                    {
                        if ( policyCreateResObj.response.statusCode == 400 && policyCreateResObj.rO.id == "PolicyNameAlreadyExistException"){
                            //WE CAN ASSUME REPORT JOB EXISTS FOR THIS PATH, REPORT JOB NEEDS TO BE UPDATED WITH NEW TRIGGERID
                            //FIRST UPDATE REPORT POLICY WITH NOTIFICATION RECEIVED IN THE REQUEST
                            //SECOND UPDATE REPORT JOB WITH NEW TRIGGERID
                            ngp.logMsg('error', "    Policy resource already exists for Report, updating policy ==> " + JSON.stringify(policyCreateResObj.rO));                            
                            ngp.ngpCoreHelper.getPolicies(ecxSessionId, function(getPoliciesResObj){
                                if (getPoliciesResObj.e || getPoliciesResObj.response.statusCode != 200) {
                                    ngp.logMsg('error', "   Get policies error  ==> " + JSON.stringify(getPoliciesResObj.rO));
                                    var failResult = ngp.getGenericEcxErrorResponse(getPoliciesResObj, ecxSessionId, driverStep);
                                    reportScheduleCreatedCb(failResult);
                                }else{
                                    var policyIdToUpdate = null;
                                    var reportPolicyFound = false;
                                    var policy_name = pjEntry.policyPostBody.name;
                                    var arrayOfPolicies = getPoliciesResObj.rO['policies'];
                                    for (var i = 0; i < arrayOfPolicies.length; i++){
                                        if (arrayOfPolicies[i]['name'] == policy_name){
                                            reportPolicyFound = true;
                                            policyIdToUpdate = arrayOfPolicies[i]['id'];
                                            ngp.logMsg('info', '    Found Report Policy(' + policy_name + ')');
                                            ngp.logMsg('info', '    Policy Id to Update(' + policyIdToUpdate + ')');
                                            break;
                                        }
                                    }

                                    if(reportPolicyFound){
                                        ngp.ngpCoreHelper.updatePolicy(ecxSessionId, policyIdToUpdate, pjEntry.policyPostBody, function(policyUpdateResObj){
                                            if ( policyUpdateResObj.e || policyUpdateResObj.response.statusCode != 200)
                                            {
                                                var failResult = ngp.getGenericEcxErrorResponse(policyUpdateResObj, ecxSessionId, driverStep);
                                                reportScheduleCreatedCb(failResult);
                                            }else{
                                                crSummaryResults.push(policyUpdateResObj.rO);
                                                var arrayOfJobs = null;
                                                ngp.ngpCoreHelper.getJobs(ecxSessionId, function(getJobsResObj){
                                                    if (getJobsResObj.e || getJobsResObj.response.statusCode != 200) {                        
                                                        ngp.logMsg('error', "Get jobs error  ==> " + JSON.stringify(getJobsResObj.rO));
                                                        var failResult = ngp.getGenericEcxErrorResponse(getJobsResObj, ecxSessionId, driverStep);
                                                            reportScheduleCreatedCb(failResult);                     
                                                    }else{
                                                        arrayOfJobs = getJobsResObj.rO['jobs'];
                                                        var reportJobFound = false;
                                                        var jobIdToUpdate = null;
                                                        for (var i = 0; i < arrayOfJobs.length; i++){
                                                            //POLICY NAME AND JOB NAME ARE THE SAME, SO WE CAN USE IT
                                                            if (arrayOfJobs[i]['name'] == policy_name){
                                                                reportJobFound = true;
                                                                jobIdToUpdate = arrayOfJobs[i]['id'];
                                                                ngp.logMsg('info', '    Found Report Job(' + arrayOfJobs[i]['name'] + ')');
                                                                ngp.logMsg('info', '    Job Id to Update(' + jobIdToUpdate + ')');
                                                                break;
                                                            }
                                                        }
                                                        if (reportJobFound){
                                                            //UPDATE JOB WITH NEW TRIGGER ID
                                                            var jobPostBody = JSON.parse(JSON.stringify(reportJobTpl));
                                                            var emptyArray = [];
                                                            jobPostBody.name = policy_name;
                                                            jobPostBody.description += policy_name;
                                                            jobPostBody.policyId = policyIdToUpdate;
                                                            jobPostBody.triggerIds[0] =  pjEntry.triggerId;                                                            
                                                            ngp.ngpCoreHelper.updateJob(ecxSessionId, jobIdToUpdate, jobPostBody, function (jobUpdateResObj){
                                                                if ( jobUpdateResObj.e || jobUpdateResObj.response.statusCode != 200)
                                                                {
                                                                    var failResult = ngp.getGenericEcxErrorResponse(jobUpdateResObj, ecxSessionId, driverStep);
                                                                    reportScheduleCreatedCb(failResult);                                
                                                                }else{
                                                                    crSummaryResults.push(jobUpdateResObj.rO);
                                                                    reportScheduleCreatedCb();
                                                                }
                                                            });   
                                                        }else{
                                                            reportScheduleCreatedCb();
                                                        }
                                                    }
                                                });                                                                                                                                                
                                            }
                                        });
                                    }else{
                                        reportScheduleCreatedCb();
                                    }
                                }
                            });                            
                        }else{
                            var failResult = ngp.getGenericEcxErrorResponse(policyCreateResObj, ecxSessionId, driverStep);
                            reportScheduleCreatedCb(failResult);
                        }
                    }
                    else
                    {
                        crSummaryResults.push(policyCreateResObj.rO);

                        //Update pjEntry for policy
                        pjEntry.policyId = policyCreateResObj.rO.id;
                        pjEntry.policyCreated = true;
                        pjEntry.policyName =  policyCreateResObj.rO.name;

                        ngp.logMsg('warn', "    Policy creation response ==> " + JSON.stringify(policyCreateResObj.rO));

                        pjEntry.jobPostBody = createJobPostBody(pjEntry);

                        ngp.addDoubleSpace();
                        ngp.logMsg('info', 'Creating Job for Report ==> ');
                        ngp.ngpCoreHelper.createJob(ecxSessionId, pjEntry.jobPostBody, function(jobCreateResObj){
                            if ( jobCreateResObj.e || jobCreateResObj.response.statusCode != 201 )
                            {
                                var failResult = ngp.getGenericEcxErrorResponse(jobCreateResObj, ecxSessionId, driverStep);
                                reportScheduleCreatedCb(failResult);
                            }
                            else
                            {
                                crSummaryResults.push(jobCreateResObj.rO);

                                pjEntry.jobId = jobCreateResObj.rO.id;
                                pjEntry.jobCreated = true;

                                ngp.logMsg('warn', "    Job creation response ==> " + JSON.stringify(jobCreateResObj.rO));

                                //RECURSE
                                createTriggerAndPolicyAndJob(pjEntries, driverStep, ecxSessionId, finishedPjEntries, crSummaryResults, reportScheduleCreatedCb);
                            }
                        });
                    }
                });
            }
        });
	}
}

/**
 * ROLLBACK REPORT SCHEDULE
 *
 * NOTES:
 *        Attempt to rollback all pjEntries (each represents a trigger/policy/job pair) regardless of whether or not they show successful creation. If entity ID is null because, then you
 *        will likely get a 500 from ECX.
 *
 * @param finishedPjEntries
 * @param ecxSessionId
 * @param rolledBackEntriesArray
 * @param policiesRolledbackCb
 */
function rollbackReportSchedule(finishedPjEntries, ecxSessionId, rolledBackEntriesArray, failResult, reportScheduleRollbackCb)
{
	if ( finishedPjEntries.length < 1 )
	{
        reportScheduleRollbackCb(failResult);
	}
	else
	{
		var pjEntry = finishedPjEntries.shift();

        /**
         * Job needs to be deleted first
         */
        if (pjEntry.jobId != null){            
            ngp.addDoubleSpace();
            ngp.logMsg('info', 'Rolling Back Report Job ==> ');
            ngp.logMsg('info', '    jobId(' + pjEntry.jobId + ')');
            ngp.ngpCoreHelper.deleteJob(ecxSessionId, pjEntry.jobId, function(jobDelResObj){
                    pjEntry.jobRollbackAttempted = true;

                    if (jobDelResObj.e || ( jobDelResObj.response.statusCode != 204 && jobDelResObj.response.statusCode != 404 && jobDelResObj.response.statusCode != 500 )) {
                        ngp.logMsg('warn', "    Rollback deletion of report job " + pjEntry.jobName + " failed because of error=" + JSON.stringify(jobDelResObj.rO));
                    }
                    else {
                        pjEntry.jobRollbackSuccessful = true;                        
                    }
            });
        }

        /**
         * Policy needs to be deleted next if it was created
         */
        if (pjEntry.policyId != null){
            ngp.addDoubleSpace();
            ngp.logMsg('info', 'Rolling Back Report Policy ==> ');
            ngp.logMsg('info', '    policyId(' + pjEntry.policyId + ')');
            ngp.ngpCoreHelper.deletePolicy(ecxSessionId, pjEntry.policyId, function(policyDelResObj){
                    pjEntry.policyRollbackAttempted = true;

                    if (policyDelResObj.e || ( policyDelResObj.response.statusCode != 204 && policyDelResObj.response.statusCode != 404 && policyDelResObj.response.statusCode != 500 )) {
                        ngp.logMsg('warn', "    Rollback deletion of report policy " + pjEntry.policyName + " failed because of error=" + JSON.stringify(policyDelResObj.rO));
                    }
                    else {
                        pjEntry.policyRollbackSuccessful = true;
                    }
            });
        }
        
        /**
         * Trigger needs be deleted in the end
         */
        if (pjEntry.triggerId != null){
            ngp.addDoubleSpace();
            ngp.logMsg('info', 'Rolling Back Report Trigger ==> ');
            ngp.logMsg('info', '    triggerId(' + pjEntry.triggerId + ')');
            ngp.ngpCoreHelper.deleteTrigger(ecxSessionId, pjEntry.triggerId, function(trigDelResObj){
                pjEntry.triggerRollbackAttempted = true;

                if (trigDelResObj.e || ( trigDelResObj.response.statusCode != 204 && trigDelResObj.response.statusCode != 404 && trigDelResObj.response.statusCode != 500 )) {
                    ngp.logMsg('warn', "Rollback deletion of report trigger " + pjEntry.triggerName + " failed because of error=" + JSON.stringify(trigDelResObj.rO));
                } else {
                    pjEntry.triggerRollbackSuccessful = true;
                }
            });
        }        

        //Add pjEntry to rolledBackEntriesArray and recurse
        rolledBackEntriesArray.push(pjEntry);
        rollbackReportSchedule(finishedPjEntries, ecxSessionId, rolledBackEntriesArray, failResult,reportScheduleRollbackCb);
	}
}