if (!('contains' in String.prototype)) {
	  String.prototype.contains = function(str, startIndex) {
	    return ''.indexOf.call(this, str, startIndex) !== -1;
	  };
	}

if (typeof String.prototype.startsWith != 'function') {
    String.prototype.startsWith = function(prefix) {
        return this.slice(0, prefix.length) == prefix;
    };
}
		 
if (typeof String.prototype.endsWith != 'function') {
    String.prototype.endsWith = function(suffix) {
        return this.slice(-suffix.length) == suffix;
    };
}

var ngp = require('./ecxngp');     
const urlParse = require('url-parse');                                                                      

var funcSteps = [                                                                                  // generalized user handling + provider registration
	{                                                                                             
		continue: 'next',         
		name: 'reg_site',                                                                      
		verb: 'POST',    
		path: null,                                                                                
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
    {                                                                                              // NB: nothing in this block will change
		continue: 'next',                                                                          // first constituent always continues
		name: 'upd_isprovider',
		verb: 'GET',
		path: null,                                                                                // further qualified by init: /api/hypervisor, /api/storage, /api/application
		ip: null,                                                                                  //!! set by ecxngp. but maybe you don't really need per step ip and ort                                                     
		port: null, 
		type: null,
		func: null,                                                                                // see initialization below 
		cb: function(res, data){ logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              
		continue: 'next',                                                                          // veuser is->reg_usermode, else ->reg_vecreateuser
		name: 'upd_veusermode',
		verb: 'GET',
		path: '/api/identity/user',
		ip: null,                                                                                                                                       
		port: null, 
		type: null,
		func: null,                                                                                // see init 
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              
		continue: 'next',                                                                          // set by previous step [ next | skip | fail ]     
		name: 'upd_vecreateuser',                                                                  // -> reg_usermode
		verb: 'POST',
		path: '/api/identity/user',                                                            
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	},	
	{                                                                                              // NB: nothing in this block will change
		continue: 'next',                                                                          // first constituent always continues
		name: 'upd_usermode',
		verb: 'GET',
		path: '/api/identity/user',
		ip: null,                                                                                  //!! set by ecxngp. but maybe you don't really need per step ip and ort                                                     
		port: null, 
		type: null,
		func: null,                                                                                // see initialization in regProviderCommon 
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              // NB: nothing in this block will change
		continue: 'next',                                                                          // set by previous step [ next | skip | fail ]     
		name: 'upd_createuser',   
		verb: 'POST',
		path: '/api/identity/user',                                                            
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              
		continue: 'next',                                                                          // set by previous step [ next | skip | fail ]     
		name: 'upd_createuser_Instances',                                                                    // -> reg_provider
		verb: 'POST',
		path: '/api/identity/user',                                                            
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              
		continue: 'next',                                                                          // set by previous step [ next | skip | fail ]     
		name: 'upd_office365key',                                                                    // -> reg_provider
		verb: 'PUT',
		path: '/api/identity/key',                                                            
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              // NB: this block is changed before registration to conform to entity being registered
		continue: 'next',         
		name: 'upd_provider', 
		verb: 'PUT',    
		path: null,                                                                                // NB: depends on type of entity we are registering, further qualified by init
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              // NB: this block is changed before registration to conform to entity being registered
		continue: 'next',                                                                          // set to skip if reg_register succeeds, but if driven it will delete identity only if we created one  
		name: 'upd_deleteuser',                                                                    
		verb: 'DELETE',    
		path: null,   
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 	
];

var HashMap = require('hashmap');

const updProviderStepFuncs = {
	reg_site: function(step, funcStream, ecxSessionId, oFromNg, cb){                                 //@@ Needs to know to set opProperties in the POST body
		
		regSite(step, funcStream, ecxSessionId, oFromNg, cb);		
	},
	upd_isprovider: function(step, funcStream, ecxSessionId, oFromNg, cb){ 
		isProvider(step, funcStream, ecxSessionId, oFromNg, cb);
	},  
	upd_veusermode: function(step, funcStream, ecxSessionId, oFromNg, cb){
		if ( ngp.hasPropertyPath(oFromNg, 'opProperties.veServerInfo'.split('.') ) )                             // present only on hypervisor registration
		{
// It is incumbent upon the called routine to reset oFromNg.ve because these calls are asynchronous; thus we might unset it here before the call finishes or before .ve has been interrogated.			
			oFromNg.ve = true;                                                                           // tell common routine that we're interested in veuser for this call                                                         
			ngp.userMode(step, funcStream, ecxSessionId, oFromNg, cb);                                   //@@!! NB: before control ever returns here, the next step will be driven - LISTEN
		}
		else                                                                                             // simply recurse to next step reg_vecreateuser who will either create VE user or recurse to hypv user mode
		{
			step.cb(ecxSessionId, null);                                                                 
			ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	
		}
	},
	upd_vecreateuser: function(step, funcStream, ecxSessionId, oFromNg, cb){ 
		if ( ngp.hasPropertyPath(oFromNg, 'opProperties.veServerInfo'.split('.') ) )                     // present only on hypervisor registration
		{
			oFromNg.ve = true; 
			ngp.createUser(step, funcStream, ecxSessionId, oFromNg, cb);
		}
		else
		{
			step.cb(ecxSessionId, null);                                                                 
			ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	
		}
	},	  
	upd_usermode: function(step, funcStream, ecxSessionId, oFromNg, cb){ 
		ngp.userMode(step, funcStream, ecxSessionId, oFromNg, cb);
	},
	upd_createuser: function(step, funcStream, ecxSessionId, oFromNg, cb){ 
		ngp.createUser(step, funcStream, ecxSessionId, oFromNg, cb); 		
	},
	upd_createuser_Instances: function(step, funcStream, ecxSessionId, oFromNg, cb) { 
		var instances = oFromNg.instances;
		var appCredentialsFromPostBody = [];
		 
		for (var i=0; i<instances.length; i++) {
			if( instances[i].user.username) {
					appCredentialsFromPostBody.push(instances[i]);
			}
		}
		   
		for (var i=0; i<appCredentialsFromPostBody.length; i++) {
			if(appCredentialsFromPostBody[i].user.username) {
				var idenityPostBody = {
					name: require('querystring').unescape(appCredentialsFromPostBody[i].user.username + '_' + oFromNg.hostAddress + '_' + appCredentialsFromPostBody[i].name),    // ****NB: can't create standard username_resourcename here****
					password: appCredentialsFromPostBody[i].user.password,
					username: require('querystring').unescape(appCredentialsFromPostBody[i].user.username),
				};
				ngp.createUserInstances(step, funcStream, ecxSessionId, oFromNg, idenityPostBody, i, appCredentialsFromPostBody.length, cb);   
			}
		}                                  
	},
	upd_office365key: function(step, funcStream, ecxSessionId, oFromNg, cb){                                 //@@ Needs to know to set opProperties in the POST body
		addOrUpdateOffice365Key(step, funcStream, ecxSessionId, oFromNg, cb);		
	},
	upd_provider: function(step, funcStream, ecxSessionId, oFromNg, cb){                                 //@@ Needs to know to set opProperties in the POST body
		if ( ngp.hasPropertyPath(oFromNg, 'opProperties.veServerInfo'.split('.') ) )                     // present only on hypervisor registration
			oFromNg.ve = true;

		updateProvider(step, funcStream, ecxSessionId, oFromNg, cb);		
	},
	upd_deleteuser: function(step, funcStream, ecxSessionId, oFromNg, cb){ 
		ngp.regDeleteUser(step, funcStream, ecxSessionId, oFromNg, cb);
	},			
};


exports.init = init;


function init(req, res, providerType)                                    
{
	ngp.logMsg('warn', 'Using new provider update');

	var ngState = ngp.getSessionState(req, res);

	if ( ngState.stale )                                                                                    // FTI we got bounced while client remained up. AND response already sent.
		return;                                                                                             // !!! R U CERTAIN THAT RESPONSE SENT IN ALL CASES HERE?

	var ecxSid = ngState.ecxSid;
	res.ecxSid = ecxSid;

	var typology = ngp.getRegistrationTypology(req, providerType);                                        

	var stepsStream = JSON.parse(JSON.stringify(funcSteps));                                    

	var updStep = null;
    var isProviderStep = null;

	stepsStream.forEach(function(step) {
		step.ip = ngp.localIp;
		step.port = ngp.ecxExternal;

		step.resourceType = typology.resourceType;
		step.type = typology.type;                                                                          // NB: we need to be told the API path for type we are registering

		if ( step.name == 'upd_createuser_Instances' ) {
			if(req.body.applicationType == 'office365') {
				step.continue = 'skip';
			}
			else {
				if(!req.body.instances) {
					step.continue = 'skip';
				}
				else {
					var instances = req.body.instances;
					var appCredentialsFromPostBody = [];
			 
					for (var i=0; i<instances.length; i++) {
						if( instances[i].user.username) {
							appCredentialsFromPostBody.push(instances[i]);
						}
					}
	
					if(appCredentialsFromPostBody.length == 0)
					step.continue = 'skip';
				}
			}		
		}

		if(step.name == 'upd_office365key' ) {
			if(req.body.applicationType != 'office365')
			step.continue = 'skip'
		}

    	if ( step.name == 'upd_isprovider' || step.name == 'upd_provider' )
		step.path = typology.path + '/' + req.params.id;	

		step.func = updProviderStepFuncs[step.name];                                                        // hook up the step's real workhorse
		
		step.cb = function(ecxSid, data) { 
			//@#@ ngp.logMsg('info', "Finished step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'");
			ngp.logMsg('info', "Finished step '" + ngp.getCurrentFunc(step.reqId, ecxSid) + '(' + this.name + ")' in session ID '" + ecxSid + "'");

			if ( data ) {
				ngp.logMsg('debug', "Data returned by step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'=" + JSON.stringify(data));
			}
		};
	});

	ngp.createRequestStore(stepsStream, ecxSid, req.reqId, 'update_' + typology.resourceType + '_' + typology.type);    // add reqId to steps, create store for the req & set currentFunc to genericFuncName
	
	ngp.setRequestState(req.reqId, ecxSid, 'crudelUrl', req.url);                                       // used by delete_userrpool steps
    
	ngState.currentFunc = 'update_' + typology.resourceType + '_' + typology.type;                        // e.g., register_storage_netapp
	return stepsStream;
}

/*
  Despite this being PUT logic, we still GET the object before updating it to capture the current (perhaps soon to be previous) user associated with the provider. Later, we may use try to 
  delete this previous user.

  MAYBE LOOK FOR BOTH regular provider user identity and ve identity (hypervisor only) and SES accordingly. Then use .ve flag in updateProvider
*/

function isProvider(step, funcStream, ecxSessionId, oFromNg, cb) {                       //!! BECUZ used by serveral submodules, keep in ecxngp   
	var scrubbedNgInput = ngp.scrubPostBody(oFromNg, true);
	ngp.logMsg('warn', "Input from client for provider existence in session ID='" + ecxSessionId + "'=" + JSON.stringify(scrubbedNgInput));

	ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path, ecxSessionId, null, function(provObj){   
		if ( provObj.e || provObj.response.statusCode != 200 ) {
			var failResult = ngp.getGenericEcxErrorResponse(provObj, ecxSessionId, step);
			ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult);
		}
		else {
			var ngState = ngp.ecxSidHash[ecxSessionId];    

			var prov = provObj.rO;                                                             // no longer compare name property; the URL :id gave us what it did

			var effUserProp = 'user';

			if ( (step.path.contains('appserver')) || ( step.path.contains('scriptserver') ) ) {
				if(!prov.osuser) 
				effUserProp = 'oskey';
				else
				effUserProp = 'osuser';
			}

			//@#@ if ( ngState.currentFunc.startsWith('update') )                                    //!! ALWAYS SO, AT LEAST UNTIL i MOVE THIS CODE TO ecxngp and use w/other funcs
			if ( ngp.getCurrentFunc(step.reqId, ecxSessionId).startsWith('update') )      
			{
				ngp.logMsg('debug', 'Extant provider=' + JSON.stringify(prov));
				
				//@#@ngp.setEndpointState(ecxSessionId, 'prevUserHref', prov[effUserProp].href);      // we will try to delete this if update has diff user		
				ngp.setRequestState(step.reqId, ecxSessionId, 'prevUserHref', prov[effUserProp].href);      // we will try to delete this if update has diff user						
				ngp.logMsg('warn', 'State for ' + ngState.currentFunc + '=' + JSON.stringify(ngState.endPoints[step.reqId]));

				if ( step.resourceType == 'hypervisor' && ngp.hasPropertyPath(prov, 'opProperties.veServerInfo.osuser'.split('.') ) )					
				{
					//ngp.logMsg('warn', 'BINGO provider=' + JSON.stringify(prov));
					//@#@ngp.setEndpointState(ecxSessionId, 'prevVeUserHref', prov.opProperties.veServerInfo.osuser);  
					ngp.setRequestState(step.reqId, ecxSessionId, 'prevVeUserHref', prov.opProperties.veServerInfo.osuser);     
				}

				step.cb(ecxSessionId, provObj.rO);                                                   
				ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	   
			}
		}
	});	
};

var addOrUpdateOffice365Key = function(step, funcStream, ecxSessionId, oFromNg, cb) {
	var instances = oFromNg.instances;
	var postBody = instances[0];
	
	delete postBody.keytype;
	var crudelUrl = ngp.getRequestState(step.reqId, ecxSessionId, 'crudelUrl');  
	var crudeUrlFixed = crudelUrl.replace('ngp', 'api');
	ngp.logMsg('warn','Sahil: ' + crudeUrlFixed);
	
	ngp.sendRequestToEcx('GET', ngp.ecxExternal, crudeUrlFixed, ecxSessionId, null, function(appServerInfo) {
		if ( appServerInfo.e || ( appServerInfo.response.statusCode != 200 ) ) {
            ngp.logMsg('warn', 'Failed to get the app server Info because of error=' + JSON.stringify(appServerInfo.rO)); 
            var failResult = ngp.getGenericEcxErrorResponse(appServerInfo, ecxSessionId);          
            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
        } else {
			var appServers = appServerInfo.rO['appCredentials'][0]['appuser']['href'];
				
			var path = urlParse(appServers).pathname;
               
			ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, path, ecxSessionId, postBody, function(regO) {
				if ( regO.e || regO.response.statusCode != 200 ) {
					var failResult = ngp.getGenericEcxErrorResponse(regO, ecxSessionId, step); 
					ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);    
				} else {
					var successResult = {
						statusCode: regO.response.statusCode,
						response: regO.rO,
					};
			
					ngp.setSerialStreamResult(ecxSessionId, successResult);

					step.cb(ecxSessionId, regO.rO);                                                       // From ECX callback now drive the recursive driver to move to next constituent function, if any                                    
					ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 					
				}	   
			});
        }
	});
}

var updateProvider = function(step, funcStream, ecxSessionId, oFromNg, cb) 
{
	var scrubbedNgInput = ngp.scrubPostBody(oFromNg, true);
	ngp.logMsg('warn', "Input from client for provider update in session ID='" + ecxSessionId + "'=" + JSON.stringify(scrubbedNgInput));

	var ngState = ngp.ecxSidHash[ecxSessionId];

	var postBody = getProviderUpdatePutBody(step, ngState, oFromNg);  
	
	ngp.ngpCoreHelper.getSites(ecxSessionId, function(getSitesResObj) {
		ngp.ngpCoreHelper.getIdentities(ecxSessionId, function(getIdentitiesResObj) {
			ngp.ngpCoreHelper.getKeys(ecxSessionId, function(getKeysResObj) {
				if (oFromNg.siteName) {
					var arrayOfSites = getSitesResObj.rO['sites'];
					for (var i = 0; i < arrayOfSites.length; i++) {
						if (arrayOfSites[i]['name'] == oFromNg.siteName) {
							postBody.siteId = arrayOfSites[i]['id']
							break;
						}
		    		}
				}	

				if(oFromNg.applicationType != 'office365') {
					if((step.resourceType == "appserver") && (oFromNg.instances)) {
						var arrayOfIdentities = getIdentitiesResObj.rO['users'];
						var mapIdentitesNameId = new HashMap();
				
						for (var i = 0; i < arrayOfIdentities.length; i++) {
							mapIdentitesNameId.set(arrayOfIdentities[i]['name'],arrayOfIdentities[i]['links']['self']['href']);
						}

						var instances = oFromNg.instances;
						var appCredentialsFromPostBody = [];

						for (var i=0; i<instances.length; i++) {
							if( ! instances[i].user.username) {
								appCredentialsFromPostBody.push({
									appuser : {
										href:instances[i].user
									},
									useForAllInstances:false,
									instanceName: instances[i].name
								});
							}
							else {
								appCredentialsFromPostBody.push({
									appuser : {
										href: mapIdentitesNameId.get(require('querystring').unescape(instances[i].user.username + '_' + oFromNg.hostAddress + '_' + instances[i].name))
									},
									useForAllInstances:false,
									instanceName: instances[i].name	
								});
							}
						}
	
						postBody.appCredentials = appCredentialsFromPostBody;
					}
				} else {
					var arrayOfKeys = getKeysResObj.rO['keys'];
					var mapKeysNameId = new HashMap();
						
					for (var i = 0; i < arrayOfKeys.length; i++) {
						mapKeysNameId.set(arrayOfKeys[i]['name'],arrayOfKeys[i]['links']['self']['href']);
					}
			
					var instances = oFromNg.instances;	
					var appCredentialsFromPostBody = [];

					for (var i=0; i<instances.length; i++) {
						appCredentialsFromPostBody.push({
							appuser : {
								href: mapKeysNameId.get(instances[i].name)
							},
							useForAllInstances:false,
							instanceName: instances[i].tenant
						});
					}
			
					postBody.appCredentials = appCredentialsFromPostBody;
				}

		    	ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path, ecxSessionId, postBody, function(regO) {
					if ( regO.e || regO.response.statusCode != 200 ) {
						var failResult = ngp.getGenericEcxErrorResponse(regO, ecxSessionId, step); 
      					var createdIdentities = ngp.getRequestState(step.reqId, ecxSessionId, 'createdIdentities');
		   
        				if (oFromNg.siteName) {
							ngp.sendRequestToEcx('DELETE', ngp.ecxExternal, ngp.ngpConstants.SITE_API + '/' + postBody.siteId , ecxSessionId, null, function(deleteSiteResponse) {
								if(deleteSiteResponse.response.statusCode != 204) {
									var failResult = {
										statusCode: deleteSiteResponse.response.statusCode,
				            	    	response: {rc: 'Failed to delete Site ' + oFromNg.siteName}																
			                		};
										
			                		ngp.logMsg('warn', 'Failed to delete Site ' + oFromNg.siteName);
		
				                	if ( ! createdIdentities || createdIdentities.length < 1 ) {
										if ( funcStream.length > 0 )                                               
					                	funcStream[0].continue = 'skip'; 	
					                	ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, true);       				
			    	            	}
			        	        	else {
										ngp.setSerialStreamResult(ecxSessionId, failResult);
				            	    	step.cb(ecxSessionId, regO.rO);                                                  
				                		ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	      
			                		}	  			   
		                		} else {
									var failResult = ngp.getGenericEcxErrorResponse(regO, ecxSessionId, step); 
			                		if ( ( ! createdIdentities || createdIdentities.length < 1 ) && funcStream.length > 0 ) {
										funcStream[0].continue = 'skip'; 
                 						ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, true);    // boolean arg causes ssfr not to set 'fail' (we've already set 'skip')
		                    		}
		                    		else {
										ngp.logMsg('warn', 'Detected PUT failure in session ' + ecxSessionId + '; attempting rollback of created identities=' + JSON.stringify(createdIdentities)); 			
			                    		ngp.setSerialStreamResult(ecxSessionId, failResult);                                        // final status
			                    		step.cb(ecxSessionId, regO.rO);                                                  
			                    		ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	                                // let run rollback of created users
		                    		} 
		                		}
	                		});
                		}
                		else {
							if ( ( ! createdIdentities || createdIdentities.length < 1 ) && funcStream.length > 0 ) {
								funcStream[0].continue = 'skip'; 
		                		ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, true);    // boolean arg causes ssfr not to set 'fail' (we've already set 'skip')
                   			}
                    		else {
								ngp.logMsg('warn', 'Detected PUT failure in session ' + ecxSessionId + '; attempting rollback of created identities=' + JSON.stringify(createdIdentities)); 			
	                    		ngp.setSerialStreamResult(ecxSessionId, failResult);                                        // final status
	                    		step.cb(ecxSessionId, regO.rO);                                                  
	                    		ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	                                // let run rollback of created users
                    		}
                		}
		   	 		}
		    		else {
						var prov = regO.rO;
			    		ngp.logMsg('debug', "Update OK response=" + JSON.stringify(prov));  
     					var oldUserHref = ngp.getRequestState(step.reqId, ecxSessionId, 'prevUserHref').split(/\:\d{4}/)[1];  // always set	
	    				var oldVeUserHref = ngp.getRequestState(step.reqId, ecxSessionId, 'prevVeUserHref');          // not always set
			
			    		var perhapsNewVeUserHref = null;

				    	if ( oldVeUserHref ) {
							oldVeUserHref = oldVeUserHref.href.split(/\:\d{4}/)[1];
				   			ngp.logMsg('warn', 'PREVVEUSER=' + oldVeUserHref);
				
					    	if ( ngp.hasPropertyPath(postBody, 'opProperties.veServerInfo'.split('.') ) )                             // present only on hypervisor registration
					    	perhapsNewVeUserHref = prov.opProperties.veServerInfo.osuser.href.split(/\:\d{4}/)[1];;	
				 
						    ngp.logMsg('warn', 'MAYBENEWPREVVEUSER=' + perhapsNewVeUserHref);				
			    		}

			    		var effUserProp = 'user';

						if ( (step.path.contains('appserver')) || ( step.path.contains('scriptserver') ) ) {
							if(!prov.osuser) 
							effUserProp = 'oskey';
							else
							effUserProp = 'osuser';
						}
					
			    		var perhapsNewUserHref = prov[effUserProp].href.split(/\:\d{4}/)[1];					

   			    		if ( oldUserHref && oldUserHref.contains('/2000') ) {
						   ngp.logMsg('warn', "Will not delete local vSnap identity");
			    		}
						else if ( oldUserHref != perhapsNewUserHref ) {
							ngp.sendRequestToEcx('DELETE', ngp.ecxExternal, oldUserHref, ecxSessionId, null, function(delO) {
								if ( delO.response && delO.response.statusCode == 204 )
					   		 	ngp.logMsg('warn', "Deleted provider's previous user association=" + oldUserHref);

					    		if ( oldVeUserHref != perhapsNewVeUserHref ) {
									ngp.logMsg('warn', "Deleting old VE user=" + JSON.stringify(oldVeUserHref));                         
						    		ngp.sendRequestToEcx('DELETE', ngp.ecxExternal, oldVeUserHref, ecxSessionId, null, function(delOve) {
										if ( delOve.response && delOve.response.statusCode == 204 )
							    		ngp.logMsg('warn', "Deleted provider's previous VE user association=" + oldUserHref);
						    		});
					    		}
				    		});
			    		}

			    		var successResult = {
							statusCode: regO.response.statusCode,
				   		 	response: regO.rO,
			    		};

				   		ngp.setSerialStreamResult(ecxSessionId, successResult);                              // NB: let caller see result of successful update

				    	if ( funcStream.length > 0 ) {
							funcStream[0].continue = 'skip';                                                 // cause skip of next serial step (deletion of identity)                        
				    	}

				    	step.cb(ecxSessionId, regO.rO);                                                      // From ECX callback now drive the recursive driver to move to next constituent function, if any                                    
			    		ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 					
		    		}
	    		});
			});
		});
	});
};

var regSite = function(step, funcStream, ecxSessionId, oFromNg, cb)
{
	if(oFromNg.siteName)
	{
    	var siteCreationPostBody = null;
	
	    siteCreationPostBody = {  
	      name: oFromNg.siteName,              
	      description: '',                                                             
	      defaultSite: false, 
	    };			   

	    ngp.sendRequestToEcx('POST', ngp.ecxExternal, ngp.ngpConstants.SITE_API, ecxSessionId, siteCreationPostBody, function(regO){ 
		    if ( regO.e || regO.response.statusCode != 201 )   
		    {
			  ngp.logMsg('info', "Site Creation Failed response=" + JSON.stringify(regO.rO));
			  var failResult = ngp.getGenericEcxErrorResponse(regO, ecxSessionId, step); 
			  ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false); 
		    }
		    else                                                                                   
		    {
			  ngp.logMsg('info', "Site successfully created OK response=" + JSON.stringify(regO.rO));  
			  step.cb(ecxSessionId, regO.rO);                                                  
			  ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 
		    }		
	    });	
	}
	else
	ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 
	
};

function getProviderUpdatePutBody(step, ngState, oFromNg)                                   //MAKEMEAWARE OF .VE as in regprovider's getProviderRegistrationPostBody    
{
	var postBody = null;

	postBody = {  
		addToCatJob: true,              
		hostAddress: oFromNg.hostAddress,
		maxStreams: oFromNg.maxStreams,                                                             
		portNumber: oFromNg.portNumber,                                                        // use whatever GUI gives us, if anything; else backend decides
		sslConnection: oFromNg.sslConnection,                                                  // ditto
        comment: oFromNg.comment,                                                              //!! backend seems to exclude now; too bad, useful to test update
		user: {
			//@#@href: ngp.getEndpointState(ngState.ecxSid, 'userHref'),	                           //!! NOT NECESSARILY EVIDENCE THAT WE CREATED THE USER!						
			href: ngp.getRequestState(step.reqId, ngState.ecxSid, 'userHref'),	                           //!! NOT NECESSARILY EVIDENCE THAT WE CREATED THE USER!	
		},
	};

	if ( oFromNg.ve )                                                                          // FTI must also provide href for opProperties.veServerInfo.username (extant | created)
	{
		oFromNg.opProperties.veServerInfo.osuser = {			
			//@#@href: ngp.getEndpointState(ngState.ecxSid, 'veUserHref'),                            //@@ EXTANT OR CREATED
			href: ngp.getRequestState(step.reqId, ngState.ecxSid, 'veUserHref'),                            //@@ EXTANT OR CREATED
		};

		delete oFromNg.opProperties.veServerInfo.username;
		delete oFromNg.opProperties.veServerInfo.password; 		
		postBody.opProperties = oFromNg.opProperties;
		delete oFromNg.ve;                                                                     // ESSENTIAL so as not to screw up remaining steps
	}

// Handle DIFFERENTIAE here. Note that funcO.resourceType is assigned by regProviderCommon based on the typology. It does NOT come from GUI.

	if ( step.resourceType == 'hypervisor' )
	{
		delete postBody.type;
		postBody.type = oFromNg.type;
		postBody.name = oFromNg.hostAddress; 
		postBody.opProperties = oFromNg.opProperties;                                          // NB should be safe even if already set by .ve case in if block above
	}

if ( step.resourceType == 'storage' )
{
	var hrefUrl = postBody.user.href;

	postBody.user = {href: hrefUrl};                                                       
	postBody.type = oFromNg.type;
	postBody.siteId = oFromNg.siteId;
	postBody.name = oFromNg.hostAddress; 
}

//!! only HANA requires appCredentials: oFromNg.appCredentials,
if (step.resourceType == 'appserver' )
{
		var hrefPathName = urlParse(postBody.user.href).pathname;
		if(hrefPathName.includes('/api/identity/key')) 
		postBody.oskey = postBody.user;
		else
		postBody.osuser = postBody.user;

		delete postBody.user;

		postBody.osType = oFromNg.osType;
		postBody.applicationType = oFromNg.applicationType;
		postBody.name = (oFromNg.applicationType === 'k8s' || oFromNg.applicationType === 'openshift') ? oFromNg.name : oFromNg.hostAddress;  
		postBody.useKeyAuthentication = oFromNg.useKeyAuthentication;
		postBody.opProperties = oFromNg.opProperties;

		if(!oFromNg.portNumber)
		delete postBody.portNumber;
		
		delete postBody.sslConnection;
}

if (step.resourceType == 'scriptserver' )
{
		postBody.osuser = postBody.user;
		
		delete postBody.user;

		postBody.osType = oFromNg.osType;
		postBody.name = oFromNg.hostAddress;  
		postBody.useKeyAuthentication = oFromNg.useKeyAuthentication;

		delete postBody.portNumber;
		delete postBody.sslConnection;
		delete postBody.addToCatJob;
}

return postBody;
}
