if (!('contains' in String.prototype)) {
	  String.prototype.contains = function(str, startIndex) {
	    return ''.indexOf.call(this, str, startIndex) !== -1;
	  };
	}

if (typeof String.prototype.startsWith != 'function') {
    String.prototype.startsWith = function(prefix) {
        return this.slice(0, prefix.length) == prefix;
    };
}
		 
if (typeof String.prototype.endsWith != 'function') {
    String.prototype.endsWith = function(suffix) {
        return this.slice(-suffix.length) == suffix;
    };
}


var ngp = require('./ecxngp');                                                                
var HashMap = require('hashmap');


   
var funcSteps = [                                                                                  // generalized user handling + provider registration
	{                                                                                              // NB: nothing in this block will change
		continue: 'next',                                                                          // first constituent always continues
		name: 'upd_get_sprof',
		verb: 'GET',                                                                               
		path: '/api/spec/storageprofile', 
		ip: null,                                                                                                                                       
		port: null, 
		type: null,
		func: null,                                                                                 
		cb: function(res, data){ logMsg('info', 'Step=' + this.name + ' finished');}
	}, 	
	{                                                                                              // NB: nothing in this block will change
		continue: 'next',                                                                          // first constituent always continues
		name: 'upd_sprof',
		verb: 'PUT',                                                                               
		path: '/api/spec/storageprofile',                    
		ip: null,                                                                                  //!! set by ecxngp. but maybe you don't really need per step ip and ort                                                     
		port: null, 
		type: null,
		func: null,                                                                                 
		cb: function(res, data){ logMsg('info', 'Step=' + this.name + ' finished');}
	}, 	
	{                                                                                              // NB: nothing in this block will change
		continue: 'next',                                                                          // first constituent always continues
		name: 'upd_sprof_jobs',
		verb: 'GET',                                                                               
		path: '/api/endeavour/job?filter=[{"property":"name","value":"*qqVVqq"}]',               
		ip: null,                                                                                  //!! set by ecxngp. but maybe you don't really need per step ip and ort                                                     
		port: null, 
		type: null,
		func: null,                                                                                 
		cb: function(res, data){ logMsg('info', 'Step=' + this.name + ' finished');}
	},	
];

//https://172.20.2.162:8443/api/spec/storageprofile?filter=[{"property":"name","value":"Bronze"}]

const updSprofStepFuncs = {
		upd_get_sprof: function(step, funcStream, ecxSessionId, oFromNg, cb){                              // if we're updating sprof then it must exist
			getSprof(step, funcStream, ecxSessionId, oFromNg, cb); 		
		},	
		upd_sprof: function(step, funcStream, ecxSessionId, oFromNg, cb){                              // if we're updating sprof then it must exist
			updateSprof(step, funcStream, ecxSessionId, oFromNg, cb); 		
		},
		upd_sprof_jobs: function(step, funcStream, ecxSessionId, oFromNg, cb){                         // if we're updating sprof then it must exist
			updateSprofJobs(step, funcStream, ecxSessionId, oFromNg, cb); 		
	},			
};


exports.init = init;


function init(req, res, providerType)                                    
{
		var ngState = ngp.getSessionState(req, res);

		if ( ngState.stale )                                                                           // FTI we got bounced while client remained up. AND response already sent.
			return;                                                                                    

		var ecxSid = ngState.ecxSid;
		res.ecxSid = ecxSid;

		var stepsStream = JSON.parse(JSON.stringify(funcSteps));                                    

		stepsStream.forEach(function(step){
			step.ip = ngp.localIp;
			step.port = ngp.ecxExternal;

			//!!if ( step.name == 'upd_sprof' )
			if ( step.name == 'upd_sprof' || step.name == 'upd_get_sprof' )
				step.path += '/' + req.params.id;	                                                       // PUT URL must contain :id

			step.func = updSprofStepFuncs[step.name];                                                  // hook up the step's real workhorse
			
			step.cb = function(ecxSid, data){ 
			//@#@ ngp.logMsg('info', "Finished step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'");
					ngp.logMsg('info', "Finished step '" + ngp.getCurrentFunc(step.reqId, ecxSid) + '(' + this.name + ")' in session ID '" + ecxSid + "'");

					if ( data )
					{
							ngp.logMsg('debug', "Data returned by step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'=" + JSON.stringify(data));
					}
			};
		});

		ngp.createRequestStore(stepsStream, ecxSid, req.reqId, 'update_storageprofile');  
		ngState.currentFunc = 'update_storageprofile';                                               // as opposed to, say, register_storageprofile

		return stepsStream;
}


function getSprof(step, funcStream, ecxSessionId, oFromNg, cb)                                   // ECC-11991
{
		ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path, ecxSessionId, null, function(sprofObj){ 				
				if ( sprofObj.e || sprofObj.response.statusCode != 200 )                         
				{
						var failResult = ngp.getGenericEcxErrorResponse(sprofObj, ecxSessionId, step);			
						ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult);	
				}	
				else
				{
					if(sprofObj.rO.spec.subpolicy[1])
					{
						var updateSprofInfo = {
								subpolName: sprofObj.rO.spec.subpolicy[0].name,
								subpolDesc: sprofObj.rO.spec.subpolicy[0].description,
								subpolIDBackup: sprofObj.rO.spec.subpolicy[1].id,
								subpolIDReplication: sprofObj.rO.spec.subpolicy[0].id,
								sprofDesc: sprofObj.rO.description,
						};
					}
					else
					{
						var updateSprofInfo = {
							subpolName: sprofObj.rO.spec.subpolicy[0].name,
							subpolDesc: sprofObj.rO.spec.subpolicy[0].description,
							subpolIDBackup: sprofObj.rO.spec.subpolicy[0].id,
							sprofDesc: sprofObj.rO.description,
					};
					}
					
						//var subpolName = sprofObj.rO.spec.subpolicy[0].name;
						ngp.setRequestState(step.reqId, ecxSessionId, 'updateSprofInfo', updateSprofInfo);           // save for next step

						step.cb(ecxSessionId, sprofObj.rO);                                                  
						ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	
				}					
			});
}


// Vide note 116: if the POST body to register OR UPDATE sprof  specifies "offload to Spectrum Protect" but lacks a subpolicy definition for REPLICATION w/software: true, that's an error.

var updateSprof = function(step, funcStream, ecxSessionId, oFromNg, cb) { 
	var ngState = ngp.ecxSidHash[ecxSessionId];   
		
	//!!var targetSubpol = ngp.getRequestState(step.reqId, ecxSessionId, 'targetSubpol');        
	var updateSprofInfo = ngp.getRequestState(step.reqId, ecxSessionId, 'updateSprofInfo');   

	var sprofTplClone = JSON.parse(JSON.stringify(ngp.sprofTpl));
	ngp.convertReplSubpolicies(sprofTplClone, oFromNg, ecxSessionId, updateSprofInfo);  
		
	ngp.ngpCoreHelper.getSites(ecxSessionId, function(getSitesResObj) {
		var arrayOfSites = getSitesResObj.rO['sites'];
		var mapSitesNameId = new HashMap();
		var mapSitesNameLinks = new HashMap();
						
		for (var i = 0; i < arrayOfSites.length; i++) {
			mapSitesNameId.set(arrayOfSites[i]['name'],arrayOfSites[i]['id']);
			mapSitesNameLinks.set(arrayOfSites[i]['name'],arrayOfSites[i].links.self.href);
		}

		sprofTplClone.spec.subpolicy.forEach(function(subpolicy) {
			
			if ( subpolicy.type == 'REPLICATION' ) {
				subpolicy.spec.destination.target.href = mapSitesNameLinks.get(subpolicy.spec.destination.target.href);
				subpolicy.id = updateSprofInfo.subpolIDReplication;
				subpolicy.spec.destination.target.id = mapSitesNameId.get(subpolicy.spec.destination.target.id);			
			}
	
			if ( subpolicy.type == 'BACKUP'  ) {
				subpolicy.spec.destination.target.href = mapSitesNameLinks.get(subpolicy.spec.destination.target.href);
				subpolicy.id = updateSprofInfo.subpolIDBackup;
				subpolicy.spec.destination.target.id = mapSitesNameId.get(subpolicy.spec.destination.target.id);			
			
			}	
		});
	   
	    ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path, ecxSessionId, sprofTplClone, function(sprofObj) { 				
			if ( sprofObj.e || sprofObj.response.statusCode != 200 ) {
				var failResult = ngp.getGenericEcxErrorResponse(sprofObj, ecxSessionId, step);			
				ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult);	
			}	
			else {
				ngp.setRequestState(step.reqId, ecxSessionId, 'sprofName', sprofObj.rO.name);        // save for next step

				var successResult = {                       
					statusCode: sprofObj.response.statusCode,
				    response: sprofObj.rO,
			    };

				ngp.setSerialStreamResult(ecxSessionId, successResult);                              // perhaps replaced by next step
				step.cb(ecxSessionId, sprofObj.rO);                                                  
				ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	
			}					
		});
	});
};


function updateSprofJobs(step, funcStream, ecxSessionId, oFromNg, cb)
{
		var sprofName = ngp.getRequestState(step.reqId, ecxSessionId, 'sprofName');                  // set by previous step
		step.path = step.path.replace('qqVVqq', require('querystring').escape(sprofName));
		
		var currentStepPath = step.path;
		var currentStepPathSplitForFilter = currentStepPath.split('=');
		var currentStepPathFilter = currentStepPathSplitForFilter[1];
		
		var fixedFilterOption = ngp.escape(currentStepPathFilter);
		var fixedStepPath = currentStepPathSplitForFilter[0] + '=' + fixedFilterOption;

		ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, fixedStepPath, ecxSessionId, null, function(jobO){ 				
				if ( jobO.e || jobO.response.statusCode != 200 )                         
				{
						var failResult = ngp.getGenericEcxErrorResponse(jobO, ecxSessionId, step);			
						ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult);	
				}	
				else
				{
					if ( jobO.rO.jobs.length < 1 )
					{
							step.cb(ecxSessionId, null);                                                  
							ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);                         // let rule the result set by the update step
							return;
					}

					driveJobEdit(jobO.rO.jobs, ecxSessionId, function(rc, result){
							step.cb(ecxSessionId, null);                                                  
							ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);                         // ALL DONE 
					});
				}					
		});		
}


function driveJobEdit(jobs, ecxSessionId, cb)                                                    // RECURSIVE function
{
		if ( jobs.length < 1 )
		{
				cb('ok');
				return;
		}

		var job = jobs.shift();

		var jobTplClone = JSON.parse(JSON.stringify(ngp.jobTpl));
	
		jobTplClone.name = job.name;
		jobTplClone.description = "";
		jobTplClone.policyId = job.policyId;

		ngp.sendRequestToEcx('PUT', ngp.ecxExternal, job.links.edit.href, ecxSessionId, jobTplClone, function(jobO){ 				
			if ( jobO.e || jobO.response.statusCode != 200 )                         
			{
					var failResult = ngp.getGenericEcxErrorResponse(jobO, ecxSessionId);			
					cb('ng', failResult);
			}	
			else
			{
					driveJobEdit(jobs, ecxSessionId, cb);                                                    // RECURSE
			}
		});
}